/*
 * mp_mpi.h
 *
 * MPI (Message Passing Interface) protocol configuration for pr_mp.c
 * 
 * rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/mp_mpi.h,v 1.9 1997/07/22 18:05:34 ludban Exp $"
 */

#include "mpi.h"

#define NEXUS_MP_PROTO_MPI
#define NEXUS_PROTO_TYPE_MP NEXUS_PROTO_TYPE_MPI
#define NEXUS_PROTO_NAME_MP "mpi"
#define MP_PROTOCOL_INFO _nx_pr_mpi_info
#define MP_STARTUP_INFO _nx_st_mpi_info

/*
 * MP_PROTO_IS_THREAD_SAFE
 *
 * It is advantageous if an MP module be thread safe, since this
 * allows you to simply create a handler thread which blocks on a
 * receive, and automatically wakes up and handles the message when
 * one arrives.
 *
 * In order for an MP module to be considered "thread safe", the
 * following must be true:
 *   1) One thread can be doing a blocking receive, while other threads
 *	are simultaneously doing sends.
 *   2) A thread can send a message to its own node, which will
 *	be received by the blocking receive in the handler thread.
 *
 * If these conditions are true, then #define MP_PROTO_IS_THREAD_SAFE
 */
/*
#define MP_PROTO_IS_THREAD_SAFE
*/

/*
 * mp_destination_t
 *
 * This encapsulates the information about a node to which we
 * may send messages.
 */
typedef int mp_destination_t;

/* Desribe contents of mp_destination_t for nexus_mi_proto_t manipulation */
#define MI_INT_COUNT 1
#define MI_STR_COUNT 0

/*
 * mp_send_status_t
 * mp_receive_status_t
 */
typedef MPI_Request mp_send_status_t;
typedef MPI_Request mp_receive_status_t;

#define MPI_MSG_TYPE 42

/*
 * MP_DEFAULT_STORAGE_SIZE
 *
 * This is the default buffer size that will be used on sends and
 * receives to avoid extraneous mallocs and locks.
 */
#define MP_DEFAULT_STORAGE_SIZE 32736

/*
 * MP_BIG_MESSAGE_MIN_SIZE
 *
 * Any message >= to this size will be sent using a two message
 * protocol.  A first, small message is sent with the size,
 * followed by a second message with the actual data.
 */
#define MP_BIG_MESSAGE_MIN_SIZE 1500

/*
 * MP_INITIALIZE()
 *
 * Do any initialization required before calling any other MP_ functions.
 */
#define MP_INITIALIZE() \
{ \
    int argc; \
    char *argv[1]; \
    argc = 0; \
    argv[0] = (char*)NULL; \
    MPI_Init(&argc, (char ***)&argv );\
}


/*
 * MP_INIT_NODE_INFO(mp_destination_t Mynode, int N_nodes)
 *
 * Set Mynode to the mp_destionation_t info for my node.
 * Set N_nodes the the number of nodes.
 */
#define MP_INIT_NODE_INFO(My_node, N_nodes) \
{ \
     MPI_Comm_rank(MPI_COMM_WORLD, &(My_node)); \
     MPI_Comm_size(MPI_COMM_WORLD, &(N_nodes)); \
}

/*
 * MP_IS_MASTER_NODE(mp_destination_t dest, nexus_bool_t result)
 */
#define MP_IS_MASTER_NODE(dest, result) \
{ \
    if ((dest) == 0) \
	result = NEXUS_TRUE; \
    else \
	result = NEXUS_FALSE; \
}

/*
 * MP_SET_DESTINATION(mp_destination_t dest, int node_num)
 */
#define MP_SET_DESTINATION(dest, node_num) \
    (dest) = (node_num)

/*
 * MP_POST_RECEIVE(Func,
 *		   nexus_byte_t *Buf,
 *		   unsigned long Buf_size,
 *		   mp_receive_status_t Status)
 *
 * Post a receive into Buf with a maximum size of Buf_size.
 * MP_RECEIVE_STATUS() and MP_RECEIVE_WAIT() should be used
 * with Status to determine completion of the receive.
 */
#define MP_POST_RECEIVE(Func, Buf, Buf_size, Status) \
{ \
    if (MPI_Irecv((char *) (Buf), \
		  (int) (Buf_size), \
		  MPI_BYTE, \
		  MPI_ANY_SOURCE, \
		  MPI_MSG_TYPE, \
		  MPI_COMM_WORLD, \
		  &(Status)) != MPI_SUCCESS) \
    { \
        nexus_fatal("MP_POST_RECEIVE(): MPI_Irecv() failed\n"); \
    } \
}

/*
 * MP_RECEIVE_STATUS(Func,
 *                   mp_receive_status_t Status,
 *                   nexus_bool_t Done)
 *
 * See if the receive described by Status has completed.  If so, then
 * set Done=NEXUS_TRUE. else set Done=NEXUS_FALSE.
 */
#define MP_RECEIVE_STATUS(Func, Status, Done) \
{ \
    int mpi_done; \
    MPI_Status mpi_status; \
    if ((MPI_Test(&(Status), \
		  &mpi_done, \
		  &mpi_status)) == MPI_SUCCESS) \
    { \
	(Done) = (mpi_done ? NEXUS_TRUE : NEXUS_FALSE); \
    } \
    else \
    { \
        nexus_fatal("MP_RECEIVE_STATUS(): MPI_Test() failed\n"); \
    } \
}


/*
 * MP_RECEIVE_WAIT(Func,
 *                 mp_receive_status_t Status,
 *                 nexus_bool_t Done)
 *
 * Wait for the receive described by Status to complete.
 * If it completes successfully, then
 * set Done=NEXUS_TRUE. else set Done=NEXUS_FALSE.
 */
#define MP_RECEIVE_WAIT(Func, Status, Done) \
{ \
    MPI_Status mpi_status; \
    if (MPI_Wait(&(Status), &mpi_status) == MPI_SUCCESS) \
    { \
	(Done) = NEXUS_TRUE; \
    } \
    else \
    { \
        nexus_fatal("MP_RECEIVE_WAIT(): MPI_Wait() failed\n"); \
    } \
}

/*
 * MP_RECEIVE_CANCEL(mp_receive_status_t Status)
 *
 * Cancel a previously posted receive.
 */
#define MP_RECEIVE_CANCEL(Status) \
    MPI_Cancel(&Status)

/*
 * MP_SEND(mp_destination_t Dest,
 *         nexus_byte_t *Buf,
 *         unsigned long Buf_size,
 *         mp_send_status_t Status)
 *
 * Send Buf, of size Buf_size, to Dest.
 * The completion of the send will be checked by calling
 * MP_SEND_STATUS() using the Status argument.
 */
#define MP_SEND(Dest, Buf, Buf_size, Status) \
{ \
    if (MPI_Isend((char *) (Buf), \
		  (int) (Buf_size), \
		  MPI_BYTE, \
		  (Dest), \
		  MPI_MSG_TYPE, \
		  MPI_COMM_WORLD, \
		  &(Status)) != MPI_SUCCESS) \
    { \
        nexus_fatal("MP_SEND(): MPI_Isend() failed\n"); \
    } \
}


/*
 * MP_SEND_STATUS(mp_send_status_t Status, nexus_bool_t Done)
 *
 * See if the send described by Status has completed.
 * If so, set Done=NEXUS_TRUE, else set Done=NEXUS_FALSE.
 */
#define MP_SEND_STATUS(Status, Done) \
{ \
    int mpi_done; \
    MPI_Status mpi_status; \
    if (MPI_Test(&(Status), \
		 &mpi_done, \
		 &mpi_status) == MPI_SUCCESS) \
    { \
	(Done) = (mpi_done ? NEXUS_TRUE : NEXUS_FALSE); \
    } \
    else \
    { \
        nexus_fatal("MP_RECEIVE_STATUS(): MPI_Test() failed\n"); \
    } \
}


/*
 * MP_COMPARE_DESTINATIONS(mp_destination_t D1, mp_destination_t D2,
 *			  nexus_bool_t Result)
 *
 * Set Result==NEXUS_TRUE if mp_destination_t's D1 and D2 are the same.
 */
#define MP_COMPARE_DESTINATIONS(D1, D2, Result) \
    Result = (((D1) == (D2)) ? NEXUS_TRUE : NEXUS_FALSE)


/*
 * MP_COPY_DESTINATION(mp_destination_t To, mp_destination_t From)
 *
 * Copy mp_destination_t 'From' to 'To'.
 */
#define MP_COPY_DESTINATION(To, From) \
    (To) = (From)


/*
 * MP_FREE_DESTINATION(mp_destination_t Dest)
 *
 * Free up any memory in 'Dest' which was malloced.
 */
#define MP_FREE_DESTINATION(Dest)


/*
 * MP_HASH_DESTINATION(mp_destination_t Dest, int Hash)
 *
 * Hash the destination information into a value
 * between 0 and PROTO_TABLE_SIZE.
 */
#define MP_HASH_DESTINATION(Dest, Hash) \
    (Hash) = ((Dest) % PROTO_TABLE_SIZE)


/*
 * MP_INIT_DESTINATION(mp_destination_t Dest)
 *
 * Initialize the destination.
 */
#define MP_INIT_DESTINATION(Dest)


/*
 * MP_NODE_EXIT()
 *
 * If a terminating process needs to do something to map itself out
 * of the set of processes, this macro should do it.
 */
#define MP_NODE_EXIT() \
{ \
     MPI_Finalize(); \
}


/*
 * MP_ABORT()
 *
 * Special statements to abort all processes.
 */
#define MP_ABORT() \
{ \
    MPI_Abort(MPI_COMM_WORLD,1); \
}


/*
 * MP_GET_MY_MI_PROTO_SIZE(int Size)
 *
 * Fillin Size with the number of bytes I need to store my_node
 * into the mi_proto byte array.
 */
#define MP_GET_MY_MI_PROTO_SIZE(Size) (Size) = sizeof(int)


/*
 * MP_GET_MY_MI_PROTO(nexus_byte_t *Array)
 *
 * Fillin my_node into the mi_proto byte array
 */
#define MP_GET_MY_MI_PROTO(Array) \
{ \
    memcpy((Array), &my_node, sizeof(int)); \
}


/* 
 * MP_CONSTRUCT_FROM_MI_PROTO(mp_destination_t Dest,
 *			      nexus_mi_proto_t *Mi_proto,
 *			      nexus_byte_t *   Array)
 *
 * Copy the needed elements from Array to Dest.
 */
#define MP_CONSTRUCT_FROM_MI_PROTO(Dest, Mi_proto, Array) \
{ \
    memcpy(&(Dest), (Array), sizeof(int)); \
}


/* 
 * MP_BROADCAST_COMMAND(int Command)
 *
 * Send the command out to everybody.
 */
#define MP_BROADCAST_COMMAND(Command) \
{ \
    nexus_byte_t mpi_buf[1]; \
    MPI_Request mpi_request; \
    int mpi_done; \
    MPI_Status mpi_status; \
    int mpi_i; \
    mpi_buf[0] = (nexus_byte_t) (Command); \
    for (mpi_i = 0; mpi_i < n_nodes; mpi_i++) \
    { \
	if (mpi_i != my_node) \
	{ \
	    MPI_Isend((char *) mpi_buf, \
		      1, \
		      MPI_BYTE, \
		      mpi_i, \
		      MPI_MSG_TYPE, \
		      MPI_COMM_WORLD, \
		      &mpi_request); \
	    MPI_Test(&mpi_request, &mpi_done, &mpi_status); \
	    while (!mpi_done) \
	    { \
		if (!using_handler_thread) \
		{ \
		    receive_messages(NON_BLOCKING); \
		} \
                MPI_Test(&mpi_request, &mpi_done, &mpi_status); \
	    } \
	} \
    } \
}

