/*
 * Nexus
 * Authors:     James Stepanek
 *              Aerospace
 *
 * overseer_qos.c	- Overseer QoS module
 */

static char *rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/overseer_qos.c,v 1.1 1997/09/02 21:25:35 stepanek Exp $";

#include "internal.h"

#ifdef BUILD_RESOURCE

#include <sys/socket.h>
#include <netinet/in.h>

#include "nexus_resource.h"

#include "qoslib.h"
 
/*
 * Only one thread is allowed to be in the QosLib code (and thus
 * mucking with data structures) at a time.  
 */

#ifdef BUILD_LITE

#define qos_enter()
#define qos_exit()

#else  /* BUILD_LITE */

static nexus_mutex_t	qos_mutex;
#define qos_enter()	nexus_debug_printf(1,("QoS_Enter\n"));nexus_mutex_lock(&qos_mutex);
#define qos_exit()	nexus_mutex_unlock(&qos_mutex);nexus_debug_printf(1,("QoS_Exit\n"));

#endif /* BUILD_LITE */


/*
 * Qoslib <-> Qosman 
 */

#ifndef URL_FILE
#define URL_FILE "/tmp/qosbert_url"
#endif

static nexus_endpoint_t     qoslib_ep;
static nexus_endpointattr_t qoslib_ep_attr;
static unsigned long qoslib_id;

/*********************************/
/*                               */
/* Nexus Handler Functions Table */
/*                               */
/*********************************/

static void attach_reply_handler( nexus_endpoint_t *, nexus_buffer_t *, 
				  nexus_bool_t);

static void qoslib_op_handler( nexus_endpoint_t *, nexus_buffer_t *, 
			       nexus_bool_t  );

#define ATTACH_REPLY_HANDLER_ID        0
#define QOSLIB_OP_HANDLER_ID           1


static nexus_handler_t qoslib_handlers[] = 
{ 
    {NEXUS_HANDLER_TYPE_NON_THREADED, attach_reply_handler},
    {NEXUS_HANDLER_TYPE_THREADED, qoslib_op_handler}
};

static nexus_startpoint_t   qosman_sp;
static int qosman_fd = -1;

/* 
 * Qosman attachment
 */

static char qosman_attach_url[256];

/* Monitor: a mutex, a condition variable and a simple     */
/* boolean to denote when the desired condition is true    */

typedef struct
{
    nexus_mutex_t mutex;
    nexus_cond_t cond;
    nexus_bool_t done;
} monitor_t;

static monitor_t 		qosman_attach_monitor;


/*
 * List of objects which may be associated with qos at some point
 * doubly-linked; null-terminated
 */

static enum {
	QOS_BOUND_RUSAGE	= 0x01,
	QOS_BOUND_REQUEST	= 0x02,
	QOS_BOUND_MASK		= 0x03
};

#define QOS_FULLY_BOUND(s)   \
	(((s) & QOS_BOUND_MASK) == (QOS_BOUND_RUSAGE | QOS_BOUND_REQUEST))


typedef struct _qos_cell {
	const nexus_resource_name_t	*key;
	struct _qos_cell		*next, *prev;
	int 				binding_state;
	nexus_rusage_info		rusage_info;
	qos_request_info		request_info;
} qos_cell;

static struct {
	struct _qos_cell		*head, *tail;
} qosres_list;

static qos_cell		*qosres_freelist = NULL;


#define QosCellAlloc(Caller, NewCell) \
	if (qosres_freelist) { \
		(NewCell) = qosres_freelist; \
		qosres_freelist = qosres_freelist->next; \
	} else { \
		NexusMalloc(Caller, (NewCell), qos_cell *, \
			sizeof(qos_cell)); \
	} \
	memset(NewCell, 0, sizeof(qos_cell)); 

#define QosCellFree(Cell) \
	(Cell)->next = qosres_freelist; \
	qosres_freelist = (Cell);

#define QosCellFind(Cell, Value) \
	for((Cell) = qosres_list.head; (Cell); (Cell) = (Cell)->next) \
		if ((nexus_resource_name_cmp((Cell)->key, (Value))) == 0) break;


typedef struct _qos_buf_hanger {
  nexus_buffer_t *bufp;
  struct _qos_buf_hanger *next, *prev;
}  qos_buf_hanger;

static struct {
  struct _qos_buf_hanger  *head, *tail;
}  qos_buf_hanger_list;

static qos_buf_hanger *qos_buf_hanger_freelist = NULL;


#define QosBufHangerAlloc(Caller, NewHanger) \
	if (qos_buf_hanger_freelist) { \
		(NewHanger) = qos_buf_hanger_freelist; \
		qos_buf_hanger_freelist = qos_buf_hanger_freelist->next; \
	} else { \
		NexusMalloc(Caller, (NewHanger), qos_buf_hanger *, \
			sizeof(qos_buf_hanger)); \
	} \
	memset(NewHanger, 0, sizeof(qos_buf_hanger)); 

#define QosBufHangerFree(Hanger) \
	(Hanger)->next = qos_buf_hanger_freelist; \
	qos_buf_hanger_freelist = (Hanger);


/*  These macros work for either kind of freelist as long as the cells
 *  have ->next and ->prev and the list head has ->head and ->tail
 */
#define ListPrepend(List, Cell) \
	(Cell)->next = (List)->head; \
	if ((List)->head) (List)->head->prev = (Cell); \
	(List)->head = (Cell); \
	(Cell)->prev = NULL; 

#define ListRemove(List, Cell) \
	if ((Cell)->prev) (Cell)->prev->next = (Cell)->next; \
	else (List)->head = (Cell)->next; \
	if ((Cell)->next) (Cell)->next->prev = (Cell)->prev; \
	else (List)->tail = (Cell)->prev; \
	(Cell)->next = (Cell)->prev = NULL; 


/*
 * Flags for qos module
 */
static nexus_bool_t	qos_done;
static nexus_bool_t     qos_aborting;

static nexus_bool_t	qos_enabled;


#define qos_fatal qos_exit(); nexus_fatal


/********************************************/
/*                                          */
/* Forward declarations for Local Functions */
/*                                          */
/********************************************/

static nexus_bool_t read_attach_file( char *url, char *file );
static void attach_to_qosman( char *attach_url );
static void post_to_qosman( void * arg, int fd );
static void qos_detach( void );

static void	qos_init(int *argc, char ***argv);
static void	qos_shutdown(nexus_bool_t shutdown_others);
static void	qos_abort(void);
static int 	qos_rusage(nexus_rusage_obj*);

static int 	_rusage_socket(nexus_rusage_obj*, qos_cell*);
static int 	_rusage_process(nexus_rusage_obj*, qos_cell*);
static int 	_rusage_thread(nexus_rusage_obj*, qos_cell*);
static int 	_rusage_memory(nexus_rusage_obj*, qos_cell*);

static int 	rusage_dump(char * msg, nexus_rusage_obj * obj);

static int 	_request_socket(qos_request_obj *, qos_cell*);
static int 	_request_process(qos_request_obj *, qos_cell*);
static int 	_request_thread(qos_request_obj *, qos_cell*);
static int 	_request_memory(qos_request_obj *, qos_cell*);

/*
 * Rusage function table
 */
static int (*qos_rusage_funcs[NEXUS_RESOURCE_NTYPES])
  (nexus_rusage_obj*, qos_cell*) = {
    _rusage_socket,
    _rusage_process,
    _rusage_thread,
    _rusage_memory
  };

/*
 * Request function table
 */
static int (*qos_request_funcs[NEXUS_RESOURCE_NTYPES])
  (qos_request_obj*, qos_cell*) = {
    _request_socket,
    _request_process,
    _request_thread,
    _request_memory
  };

/*
 * Overseer interface for quality of service, a.k.a. rusage
 */

static nexus_overseer_funcs_t qos_funcs = {
  qos_init,
  qos_shutdown,
  qos_abort,
  qos_rusage,
};


/*
 * nexus_overseer_qos_info()
 *
 * Return the nexus_overseer_funcs_t function table for this overseer module.
 */
void *nexus_overseer_qos_info(void)    {

  return((void *) (&qos_funcs));

} /* nexus_overseer_qos_info() */

/*
 * qos_init()
 *
 * Initialize the QoS overseer
 */
static void qos_init(int *argc, char ***argv)  {

  int i;

#ifndef BUILD_LITE
  nexus_mutex_init(&qos_mutex, (nexus_mutexattr_t *) NULL);
#endif /* BUILD_LITE */

  qosres_list.head = qosres_list.tail = NULL;
  qos_buf_hanger_list.head = qos_buf_hanger_list.tail = NULL;

  /*
   * Try to get attachment url
   */
  if  ( !read_attach_file( qosman_attach_url, URL_FILE ) )
    {
      nexus_stdio_lock();

      fprintf(stderr, 
	      "ERROR: Could not read attach_url from file: %s\n", 
	      URL_FILE);
      fprintf(stderr, 
	      "qosman may not be running on this file system.\n");
      nexus_stdio_unlock();
    }

  /*
   * Initialize qosman qoslib endpoint
   */
  nexus_endpointattr_init( &qoslib_ep_attr );
  nexus_endpointattr_set_handler_table( &qoslib_ep_attr, qoslib_handlers,
					sizeof(qoslib_handlers)/sizeof(nexus_handler_t) );
  nexus_endpoint_init( &qoslib_ep, &qoslib_ep_attr );

  /*
   * Initialize attachment monitor
   */
  nexus_mutex_init( &(qosman_attach_monitor.mutex), (nexus_mutexattr_t *)NULL );
  nexus_cond_init( &(qosman_attach_monitor.cond), (nexus_condattr_t *)NULL );

  /* Enter the Monitor and clear the flag */

  nexus_mutex_lock( &(qosman_attach_monitor.mutex) );
  qosman_attach_monitor.done = NEXUS_FALSE;
  nexus_mutex_unlock( &(qosman_attach_monitor.mutex) );

  /* Nexus should be able to do RSRs by now */
  
  attach_to_qosman( qosman_attach_url );
  
} /* qos_init() */


/*
 * qos_shutdown()
 *
 * This routine is called during normal shutdown of a process.
 */
static void qos_shutdown(nexus_bool_t shutdown_others)   {

  nexus_printf("shutting down\n");

  qos_enter();

  qos_detach();
  qos_done = NEXUS_TRUE;

  qos_exit();

  nexus_debug_printf(1, ("qos_shutdown(): done\n"));
    
} /* qos_shutdown() */


/*
 * qos_abort()
 *
 * This routine is called during the _abnormal_ shutdown of a process.
 *
 */
static void qos_abort()   {

  nexus_printf("aborting\b");
  qos_detach();
  qos_done     = NEXUS_TRUE;
  qos_aborting = NEXUS_TRUE;

} /* qos_abort() */


/*
 * qos_detach()
 *
 * This detaches the current qoslib from qosman implying that qosman
 * releases all reservations made from this qoslib.
 *
 */
static void qos_detach( void )
{
  nexus_buffer_t buffer;
  int rsr_op = QOS_DETACH_OP;

  /* Prevent any further rusage/request calls by logically "detaching" */

  nexus_mutex_lock( &(qosman_attach_monitor.mutex) );
  qosman_attach_monitor.done = NEXUS_FALSE;
  nexus_mutex_unlock( &(qosman_attach_monitor.mutex) );

  /* Detach for real */

  nexus_buffer_init( &buffer, nexus_sizeof_u_long(1)+nexus_sizeof_int(1), 0 );
  nexus_put_u_long( &buffer, &qoslib_id, 1 );
  nexus_put_int( &buffer, &rsr_op, 1 );
  nexus_send_rsr( &buffer, 
		  &qosman_sp,
		  QOSLIB_OP_HANDLER_ID, 
		  NEXUS_TRUE, 
		  NEXUS_FALSE );
  /*  Should we put in a handshake for the detach???  */
}


/*
 * read_attach_file()
 */

static nexus_bool_t read_attach_file( char *attach_url, char *file )
{
    FILE *fp;
    nexus_bool_t rc;

    /* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */

    nexus_stdio_lock();

    printf("opening attach url file ...");

    if ((fp = fopen(file, "r")) == (FILE *) NULL) {

	rc = NEXUS_FALSE;
	printf(" FAILED\n");

    }  else  {

	rc = NEXUS_TRUE;

	printf("scanning ...");
	fscanf(fp, "%s", attach_url);
	printf(" done\n");

	fclose(fp);
    } /* endif */

    nexus_stdio_unlock();

    return rc;

} /* end read_attach_file() */


/***********************************************************************
 * RUSAGE Interface
 **********************************************************************/


/*
 * qos_rusage()
 *
 * Notify the QoS overseer about resource usage
 *
 */
static int qos_rusage(nexus_rusage_obj* rusage_obj) {

  int rc = QOS_SUCCESS;
  qos_cell * cell;

  qos_enter();

  QosCellFind(cell, rusage_obj->who);

  /*
   * If we have never seen this resource, add it to our list; otherwize
   * modify the existing list entry with the passed rusage object.  
   */

  if (cell == NULL) {
    QosCellAlloc(qos_rusage(), cell);
    cell->key = rusage_obj->who;
    ListPrepend(&qosres_list, cell);
  }

  cell->binding_state |= QOS_BOUND_RUSAGE;

  if (qos_rusage_funcs[rusage_obj->who->type]) {

    rc = (qos_rusage_funcs[rusage_obj->who->type])(rusage_obj, cell);

  }

  if (QOS_FULLY_BOUND(cell->binding_state)) {

    nexus_fd_register_for_select(qosman_fd, 0, post_to_qosman, 0, (void*)cell);

  }

/*
 * On a destroy, remove the entry from the list if present.  
 */

  switch(rusage_obj->op) {
    case RUSAGE_DESTROY:
      ListRemove(&qosres_list, cell);
      QosCellFree(cell);
      break;
  }

  qos_exit();

  return (rc);

} /* qos_rusage() */


/*
 * _rusage_socket()
 *
 * Notify the QoS overseer about resource usage
 *
 */
static int _rusage_socket(nexus_rusage_obj* rusage_obj, qos_cell * cell) {

  memcpy(&cell->rusage_info.socket, rusage_obj, sizeof(nexus_rusage_socket));

  return (rusage_dump("_rusage_socket()!", rusage_obj));
} /* qos_rusage_socket */


/*
 * _rusage_process()
 *
 * Notify the QoS overseer about resource usage
 *
 */
static int _rusage_process(nexus_rusage_obj* rusage_obj, qos_cell * cell)
{
  nexus_buffer_t buffer;

  memcpy(&cell->rusage_info.process, rusage_obj, sizeof(nexus_rusage_process));

  return (rusage_dump("_rusage_process()!", rusage_obj));
} /* _rusage_process */


/*
 * _rusage_thread()
 *
 * Notify the QoS overseer about resource usage
 *
 */
static int _rusage_thread(nexus_rusage_obj* rusage_obj, qos_cell * cell) {

  memcpy(&cell->rusage_info.thread, rusage_obj, sizeof(nexus_rusage_thread));

  return (rusage_dump("_rusage_thread()!", rusage_obj));
} /* _rusage_thread */


/*
 * _rusage_memory()
 *
 * Notify the QoS overseer about resource usage
 *
 */
static int _rusage_memory(nexus_rusage_obj* rusage_obj, qos_cell * cell) {

  memcpy(&cell->rusage_info.memory, rusage_obj, sizeof(nexus_rusage_memory));

  return (rusage_dump("_rusage_memory()!", rusage_obj));
} /* _rusage_memory */



static int rusage_dump(char * msg, nexus_rusage_obj * obj) {
  char * type, * op;

  switch(obj->who->type) {
    case RESOURCE_SOCKET:
      type = "SOCKET";
      break;
    case RESOURCE_PROCESS:
      type = "PROCESS";
      break;
    case RESOURCE_THREAD:
      type = "THREAD";
      break;
    case RESOURCE_MEMORY:
      type = "MEMORY";
      break;
  }

  switch(obj->op) {
    case RUSAGE_CREATE:
      op = "CREATE";
      break;
    case RUSAGE_DESTROY:
      op = "DESTROY";
      break;
    case RUSAGE_MODIFY:
      op = "MODIFY";
      break;
  }

  nexus_printf("%s: resource: %d  type: %s  op: %s\n", 
	       msg, obj->who->id, type, op);

  return ( QOS_SUCCESS );
}



/***********************************************************************
 * REQUEST Interface
 **********************************************************************/

/*
 * qos_request()
 *
 * Submit request for quality of service 
 */

int qos_request(qos_request_obj *request_obj) {

  int rc = QOS_SUCCESS;
  qos_cell * cell;

  if  ( !qosman_attach_monitor.done )  return QOS_NOT_ATTACHED;

  qos_enter();

  QosCellFind(cell, request_obj->who);

  if (cell == NULL) {
    QosCellAlloc(qos_request(), cell);
    cell->key = request_obj->who;
    ListPrepend(&qosres_list, cell);
  }

  cell->binding_state |= QOS_BOUND_REQUEST;

  if (qos_request_funcs[request_obj->who->type]) {

    rc = (qos_request_funcs[request_obj->who->type])(request_obj, cell);

  }

  if (QOS_FULLY_BOUND(cell->binding_state)) {

    nexus_fd_register_for_select(qosman_fd, 0, post_to_qosman, 0, (void*)cell);
  }

  qos_exit();

  return (rc);
}


/*
 * _request_socket()
 *
 * Submit request for quality of service 
 */
static int _request_socket( qos_request_obj *request_obj, qos_cell* cell )
{
  memcpy( &(cell->request_info.socket), request_obj, sizeof(qos_request_socket) );

  return QOS_SUCCESS;
} /* _request_socket() */


/*
 * _request_process()
 *
 * Submit request for quality of service 
 */
static int _request_process( qos_request_obj *request_obj, qos_cell* cell )
{
  memcpy( &(cell->request_info.process), request_obj, sizeof(qos_request_process) );

  return QOS_SUCCESS;
} /* _request_process() */


/*
 * _request_thread()
 *
 * Submit request for quality of service 
 */
static int _request_thread( qos_request_obj *request_obj, qos_cell* cell )
{
  memcpy( &(cell->request_info.thread), request_obj, sizeof(qos_request_thread) );

  return QOS_SUCCESS;
} /* _request_thread() */


/*
 * _request_memory()
 *
 * Submit request for quality of service 
 */
static int _request_memory( qos_request_obj *request_obj, qos_cell* cell )
{
  memcpy( &(cell->request_info.memory), request_obj, sizeof(qos_request_memory) );

  return QOS_SUCCESS;
} /* _request_memory() */


/*
 *  Name access routines
 */

nexus_resource_name_t *nexus_context_name( void )
{
  nexus_context_t *contextp;

  _nx_context( &contextp );
  return  &(contextp->resource_name);
}


nexus_resource_name_t *nexus_thread_name( void )
{
  _nx_thread_t *tp;

  _nx_thread_self( &tp );
  return  &(tp->resource_name);
}


nexus_resource_name_t *nexus_memory_name( void )
{
  nexus_resource_name_t *new_mem_name;

  new_mem_name = nexus_malloc( sizeof(nexus_resource_name_t) );
  nexus_resource_name( new_mem_name, RESOURCE_MEMORY );
  return new_mem_name;
}


nexus_resource_name_t *nexus_socket_name( nexus_startpoint_t *sp )
{
  return sp->mi_proto->proto->funcs->get_resource_name( sp->mi_proto->proto );
}

/***********************************************************************
 *  RSR handling routines:  send an RSR
***********************************************************************/

/*
 * post_to_qosman()
 *
 * Send resource rusuage and request pair to QosMan
 *
 */

static void post_to_qosman( void * arg, int fd )
{
  nexus_buffer_t *buffer;
  int rsr_op = QOS_RESOURCE_OP;
  qos_buf_hanger *hanger;

  qos_cell * cell = (qos_cell*) arg;

  /*
   *  The callback func pointer is sent to QosMan for possible later use on the
   *  standard QosLib callback EP
   */

  NexusMalloc( post_to_qosman(), buffer, nexus_buffer_t *, sizeof(nexus_buffer_t) );

  nexus_buffer_init( buffer,

		     nexus_sizeof_int(2) +
		     nexus_resource_sizeof_name(1) + 
		     nexus_sizeof_rusage( (nexus_rusage_obj *)&(cell->rusage_info), 1) + 
		     qos_sizeof_request( (qos_request_obj*)&(cell->request_info), 1) ,

		     0
		   );

  nexus_put_u_long( buffer, &qoslib_id, 1 );
  nexus_put_int( buffer, &rsr_op, 1 );

  nexus_resource_put_name( buffer, cell->key );  /* The resource name contains the type */

  nexus_rusage_put( buffer, (nexus_rusage_obj *)&(cell->rusage_info) );

  qos_request_put( buffer, (qos_request_obj *)&(cell->request_info) );

  nexus_send_rsr( buffer, &qosman_sp, QOSLIB_OP_HANDLER_ID, NEXUS_TRUE, NEXUS_FALSE);
  NexusFree( buffer );

}


/***********************************************************************
 * attach_to_qosman()
 *
 * Attachment in Nexus is done by making a request to attach to another
 * process designated by a URL.  If the request is granted, we get a 0
 * return code and a startpoint that points back to the granting process.
 *
 * NOTE: We get a startpoint to the process we just attached to, they
 *       do not get a startpoint back to us.  If we need them to have
 *       a startpoint back to us, we need to send them one using the
 *       startpoint we just received ... the one that points back to them.
 *
 * If the request is denied, we get a non-zero return code back.
 *
 ***********************************************************************/

static void attach_to_qosman( char *attach_url )
{
    int rc;
    nexus_startpoint_t sp_to_my_ep;
    nexus_buffer_t buffer;

    nexus_resource_name_t *qosman_who;
    qos_cell *cell;


    nexus_printf("attempting attach\n");
    rc = nexus_attach( attach_url, &qosman_sp );
    nexus_printf("attach successful\n");

    /* if things went OK, we now have a startpoint to qosman */

    if  ( rc )
    {
        nexus_printf("ERROR: Could not attach to qosman using url: %s rc %d\n", 
		     attach_url, rc);
        nexus_abort();
    } /* endif */

    /*********************************/
    /* send qosman a startpoint      */
    /* to myself and wait for ACK    */
    /*********************************/

    nexus_printf("sending my startpoint to qosman\n");

    nexus_startpoint_bind( &sp_to_my_ep, &qoslib_ep );

    nexus_buffer_init( &buffer, nexus_sizeof_startpoint(&sp_to_my_ep, 1), 0 );
    nexus_put_startpoint_transfer( &buffer, &sp_to_my_ep, 1 );

    /* sending startpoint to qosman and invoking attach_handler() there */

    nexus_send_rsr( &buffer,
                    &qosman_sp,
                    ATTACH_REPLY_HANDLER_ID,
                    NEXUS_TRUE,
                    NEXUS_FALSE );

    nexus_printf("startpoint successfully sent\n");

    /* waiting for ACK */

    /* Monitors can get a false wakeup so we have to test for the desired  */
    /* condition on each wakeup                                            */

    nexus_printf("awaiting attachment callback rsr\n");

    nexus_mutex_lock( &(qosman_attach_monitor.mutex) );
    while  ( !(qosman_attach_monitor.done) )  {
       nexus_cond_wait( &(qosman_attach_monitor.cond), &(qosman_attach_monitor.mutex) );
    } /* endwhile */


    /* 
     * Initial fd on qosman's socket
     */

    qosman_who = nexus_socket_name(&qosman_sp);
    QosCellFind(cell, qosman_who);
    if(cell) {
      qosman_fd = cell->rusage_info.socket.fd;
    } else {
      nexus_printf("ERROR: Could not find qosman's file descriptor.\n");
      nexus_abort();
    } /* endif */

    nexus_mutex_unlock( &(qosman_attach_monitor.mutex) );

    nexus_printf("attachment callback rsr received\n");

} /* end attach_to_qosman() */


/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

/*
 * attach_reply_handler()
 */
static void attach_reply_handler( nexus_endpoint_t *endpoint, 
                                  nexus_buffer_t *buffer,
                                  nexus_bool_t is_non_threaded_handler )
{
  /*  This is where qoslib receives its id from QoSMan  */

  nexus_get_u_long( buffer, &qoslib_id, 1 );
  nexus_printf( "QosLib recvd client id [%x]\n", qoslib_id );

  /* clean-up */

  nexus_buffer_destroy(buffer);

  nexus_mutex_lock( &(qosman_attach_monitor.mutex) );
    qosman_attach_monitor.done = NEXUS_TRUE;
    nexus_cond_signal( &(qosman_attach_monitor.cond) );
  nexus_mutex_unlock( &(qosman_attach_monitor.mutex) );   

} /* end attach_reply_handler() */


/*
 * qoslib_op_handler()
 * This is where qoslib will execute operations on behalf of qosman.
 * The current body here is just for debugging and development.
 */

static void qoslib_op_handler( nexus_endpoint_t *endpoint, 
                               nexus_buffer_t *buffer, 
                               nexus_bool_t is_non_threaded_handler )
{
    qos_callback cb_func;
    qos_status_obj arg;
    nexus_resource_name_t dummy_name;

    nexus_get_u_long( buffer, (unsigned long *)&(cb_func), 1 );
    nexus_get_int( buffer, (int *)&arg, 1 );

    nexus_printf( "Client recvd arg [%d]\n", arg );

    (cb_func)( &dummy_name, &arg );  /* Call the App Callback */

    nexus_buffer_destroy(buffer);    /* clean-up */

} /* end qoslib_op_handler() */


#endif /* BUILD_RESOURCE */


/***************************/
/*                         */
/*  Development Code       */
/*                         */
/***************************/

void qos_list_qosman( void )
{
  nexus_buffer_t buffer;
  int rsr_op = QOS_LIST_OP;

  nexus_buffer_init( &buffer, nexus_sizeof_u_long(1)+nexus_sizeof_int(1), 0 );
  nexus_put_u_long( &buffer, &qoslib_id, 1 );
  nexus_put_int( &buffer, &rsr_op, 1 );
  nexus_send_rsr( &buffer, &qosman_sp, QOSLIB_OP_HANDLER_ID, 
		  NEXUS_TRUE, NEXUS_FALSE);
}
