
/*
 * pr_shm.c	- Protocol module for shared memory.
 */


static char *rcsid = "$Header: /nfs/globus1/src/master/nexus_4.1/pr_shm.c,v 1.11 1997/09/07 06:52:10 toonen Exp $";

/* todo :

     1) Add process shared mutex locks and conditions when 
        _POSIX_THREAD_PROCESS SHARED is set.

     2) Handle messages larger than segment size.

     3) Make user specified segment size pass on to nodes.
*/

#include "internal.h"

#include <netdb.h> 
#include <limits.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>


/*
 * Include protocol and shared memory specific setup for type
 * of shared memory used.
 */
#define NEXUS_SHMEM_INCLUDE "pr_shm_ipc.h"
#include NEXUS_SHMEM_INCLUDE


/*
 * Hash table size for the proto table
 */
#define PROTO_TABLE_SIZE 1021

/*
 * Thread is handler?
 *
 * Thread specific storage is used to keep track if the current
 * thread is a handler thread or not.
 */
#ifdef BUILD_LITE
#define _nx_set_i_am_shm_handler_thread() /* nop */
#define _nx_i_am_shm_handler_thread(Result) *(Result) = NEXUS_TRUE
#else
static nexus_thread_key_t i_am_shm_handler_thread_key;
#define _nx_set_i_am_shm_handler_thread() \
    nexus_thread_setspecific(i_am_shm_handler_thread_key, (void *) 1)
#define _nx_i_am_shm_handler_thread(Result) \
    *(Result) = (nexus_bool_t)nexus_thread_getspecific(i_am_shm_handler_thread_key)
#endif /* BUILD_LITE */


/*
 * Only one thread is allowed to be in the shm code (and thus
 * mucking with data structures) at a time.
 */
static nexus_mutex_t		shm_mutex;
static nexus_bool_t		shm_done;
static nexus_bool_t		handle_in_progress;
static nexus_bool_t		using_handler_thread;
static nexus_bool_t		handler_thread_done;
static nexus_mutex_t		handler_thread_done_mutex;
static nexus_cond_t		handler_thread_done_cond;


#define shm_enter() nexus_mutex_lock(&shm_mutex);
#define shm_exit()  nexus_mutex_unlock(&shm_mutex);
#define shm_fatal  shm_exit(); nexus_fatal

/*
 * The mi_proto for this protocol module carries a host name
 * string with it.  This allows one node to completely distinguish
 * itself from another.
 */
static char 	_nx_hostname_string[MAXHOSTNAMELEN];
static int	_nx_hostname_string_length;
    
/*
 * Other useful defines
 */
#define BLOCKING			NEXUS_TRUE
#define NON_BLOCKING			NEXUS_FALSE

#define CLOSE_SHUTDOWN_FLAG		NEXUS_DC_FORMAT_LAST + 1
#define CLOSE_HANDLER_FLAG		NEXUS_DC_FORMAT_LAST + 2


/*
 * Some forward typedef declarations...
 */
typedef struct _shm_buffer_t	shm_buffer_t;
typedef struct _shm_proto_t	shm_proto_t;


/*
 * Some useful queue macros
 */
#define Enqueue(Qhead, Qtail, Item) \
{ \
    if (Qhead) \
    { \
	(Qtail)->next = (Item); \
	(Qtail) = (Item); \
    } \
    else \
    { \
	(Qhead) = (Qtail) = (Item); \
    } \
}

#define Dequeue(Qhead, Qtail, Item) \
{ \
    (Item) = (Qhead); \
    (Qhead) = (Qhead)->next; \
}

#define QueueNotEmpty(Qhead)	(Qhead)


/******************************************************************
 * shared memory specific setup 		 
 ******************************************************************/

/*
 * SHMEM_DEFAULT_SEGMENT_SIZE
 *
 * This is the default segment size for the shared memory segment
 * that each process will use when they create.  The segment created
 * by each process acts as an in box for sends by other nodes.  A
 * node will receive out of it's own segmnent, and send by putting
 * messages in the destination's segment.
 */
#define SHMEM_DEFAULT_SEGMENT_SIZE 1024 * 256
#define SHMEM_MIN_SEGMENT_SIZE     1024 * 4


/*
 * SHMEM_LOCK_PROBE_LIMIT
 *
 * When process shared mutex conditions are unavailable, shared memory
 * is locked via semaphores.  Since a full semaphore lock will block
 * all threads, a flag, or "soft lock" is utilized to minimize full
 * process blocking.  To avoid starvation, we'll eventually have to
 * request the full semaphore lock, but only after a set number of tries
 * on the soft lock.  The soft lock probe limit defined here sets the 
 * number of times a process will check the soft lock value before 
 * queueing up for the full semaphore lock request. 
 */
#define SHMEM_LOCK_PROBE_LIMIT  500000


/*
 * Memory requests in a destination segment do not fail immediately.
 * Storage that is not immediately available will become free when
 * the destination receives it's pending messages.  For this reason
 * mallocs will retry their requests, waiting for shared memory to
 * become available.  This limit puts a bound on how many times we
 * should try before giving up.
 */
#define SHMEM_MALLOC_MAX_TRIES  500000


/*
 * Lock values and operations
 */

#define SHMEM_UNLOCKED_VALUE  1
#define SHMEM_LOCKED_VALUE    0
#define SHMEM_LOCK_OP        -1
#define SHMEM_UNLOCK_OP       1


/*
 *  Define semaphore control request union for systems without it. 
 */
#ifndef HAVE_UNION_SEMUN
union semun
{
  int val;
  struct semid_ds *buf;
  ushort *array;
};
#endif


/*
 * As pointer addresses will differ between processes, we
 * can not store pointers within the shared memory segment.
 * Instead, we store offsets from top of segment in their 
 * place.
 */  
typedef long shm_offset_t;


/*
 * Misc shared memory defines
 */
#define SHMEM_FAILURE -1
#define NULL_OFFSET   -1
#define SHMEM_CREATE_PERMS  (IPC_CREAT | IPC_EXCL | 0600)
#define SHMEM_CONNECT_PERMS (0600)


/*
 * Forward declare shared memory control record types.
 */
typedef struct shm_segment_struct_t shm_segment_t;
typedef struct shm_heap_struct_t    shm_heap_t;
typedef struct shm_cell_struct_t    shm_cell_t;


/*
 * Shared memory segment is split into two heaps, top and bottom.
 * Only one heap is considered active at any time.
 * The inactive heap is switched to during garbage collection.
 */   
typedef struct shm_segment_struct_t
{
  shm_offset_t   cur_heap;
  shm_offset_t   top_heap;
  shm_offset_t   bot_heap;

  nexus_mutex_t  thread_lock;
  int            soft_lock;
 
  int            heap_size;
  int            alloc_count;
  nexus_bool_t   reclaiming;
  nexus_bool_t   receiving;

} shm_segment_struct_t;


/*
 * Each shared memory heap begins with a shared memory heap header.
 * Allocated heap cells are maintained as a linked list.
 */

typedef struct shm_heap_struct_t                    
{ 
  shm_offset_t  alloc_cell_head;           
  shm_offset_t  alloc_cell_tail;

  shm_offset_t  free_stg_loc;         
  int           free_stg_amt;
  int           unref_stg_amt; 
     
} shm_heap_struct_t;



/*
 * Heap cells form a two way linked list of allocated shared memory.
 * On allocation, they are marked with a reference count of two.
 * There will be two reference to this cell, one by sender and one
 * by recoeiver.  This reference count indicates the current status
 * of the shared memory heap cell. It can be received on reference
 * count of 1.  It may be reclaimed on reference count of 0.
 */

typedef struct shm_cell_struct_t                
{
  shm_offset_t  prev_cell;
  shm_offset_t  next_cell;
  shm_offset_t  this_cell;
  int           ref_count;
  int           cell_size;
  int           data_size;
  nexus_bool_t  reclaimed;
  nexus_byte_t  cell_data;

} shm_cell_struct_t;


/*
 * Heap reference values 
 */
#define SHMEM_SENDING      3
#define SHMEM_RECEIVING    2
#define SHMEM_RECV_WAIT    1
#define SHMEM_UNREFERENCED 0


/*
 * Shared memory address generation from offsets based
 * on top of segment.
 */
#define SHMEM_SEGMENT_ADDR(Segment, Offset, Addr) \
  if ((Offset) == NULL_OFFSET) \
    (Addr) = (shm_segment_t *) NULL; \
  else \
    (Addr) = (shm_segment_t *) (((char *) (Segment)) + (Offset));


#define SHMEM_HEAP_ADDR(Segment, Offset, Addr) \
  if ((Offset) == NULL_OFFSET) \
    (Addr) = (shm_heap_t *) NULL; \
  else \
    (Addr) = (shm_heap_t *) (((char *) (Segment)) + (Offset));


#define SHMEM_CELL_ADDR(Segment, Offset, Addr) \
  if ((Offset) == NULL_OFFSET) \
    (Addr) = (shm_cell_t *) NULL; \
  else \
    (Addr) = (shm_cell_t *) (((char *) (Segment)) + (Offset));


/*
 *  Thread level shared memory lock operations
 */

#ifndef BUILD_LITE

#define SHMEM_THREAD_LOCK_INIT(Segment) \
    nexus_mutex_init(&Segment->thread_lock, \
                       (nexus_mutexattr_t *) NULL) 

#define SHMEM_THREAD_SET_LOCKED(Segment) \
   nexus_mutex_lock (&Segment->thread_lock)

#define SHMEM_THREAD_SET_UNLOCKED(Segment) \
   nexus_mutex_unlock (&Segment->thread_lock)

#define SHMEM_THREAD_SCHED_YIELD()  nexus_thread_yield()

#else
#define SHMEM_THREAD_LOCK_INIT(Segment)
#define SHMEM_THREAD_SET_LOCKED(Segment)
#define SHMEM_THREAD_SET_UNLOCKED(Segment)
#define SHMEM_THREAD_SCHED_YIELD()
#endif



/******************************************************************
 * Shared memory protocol	       
 ******************************************************************/

/*
 * shm_proto_t
 *
 * This is an overload of nexus_proto_t.  It adds the
 * shm specific information to that structure.
 */
struct _shm_proto_t
{
    nexus_proto_type_t		type;	/* NEXUS_PROTO_TYPE_SHMEM */
    nexus_proto_funcs_t *	funcs;
    unsigned long		direct_custom_min_size;
    unsigned long		direct_custom_max_size;
    unsigned long		direct_pointer_min_size;
    unsigned long		direct_pointer_max_size;
    nexus_bool_t		can_use_iovec;
    unsigned long		reserved_header_size;
    shm_destination_t		destination;
    int				reference_count;
    shm_handle_t                dest_handle;
    shm_segment_t             * dest_segment;
    int                         dest_lock;     
};

/******************************************************************
 * Outgoing message handling
 ******************************************************************/

static struct _nexus_buffer_t *	send_buffer;
static struct _nexus_buffer_t *	send_q_head;
static struct _nexus_buffer_t *	send_q_tail;

/******************************************************************
 * Incoming message handling
 ******************************************************************/

static nexus_byte_t * receive_buffer = (nexus_byte_t *) NULL;
static unsigned long  receive_buffer_size;

/******************************************************************
 * Shared memory handling
 ******************************************************************/

static shm_proto_t  * my_proto = NULL;
static int            shm_segment_size;
static int            shm_alignment;
static nexus_bool_t   sending;

/*****************************************************************/

/*
 * Protocol table stuff
 *
 * The protocol table is hashed on the destination. The table itself is an
 * array of header structures pointing to a linked list of buckets.
 *
 * This table is used to avoid creating multiple shm_proto_t
 * objects to the same context.  Multiple global pointers to the same
 * context share a shm_proto_t.
 */
typedef struct _proto_table_entry_t
{
    shm_proto_t *proto;
    struct _proto_table_entry_t *next;
} proto_table_entry_t;

struct _proto_table_entry_t	proto_table[PROTO_TABLE_SIZE];

static void			proto_table_init(void);
static void			proto_table_insert(shm_proto_t *proto);
static shm_proto_t *		proto_table_lookup(shm_destination_t *dest);


/*
 * Various forward declarations of procedures
 */
static nexus_proto_type_t shm_proto_type(void);
static void		  shm_init(int *argc, char ***argv,
				   nexus_bool_t * add_to_my_mi_proto);
static void		  shm_shutdown(nexus_bool_t shutdown_others);
static void		  shm_abort(void);
static nexus_bool_t	  shm_poll(void);
static void		  shm_blocking_poll(void);

static int		  shm_send_rsr(struct _nexus_buffer_t *buffer);
static void               shm_increment_reference_count(nexus_proto_t *nproto);
static nexus_bool_t	  shm_decrement_reference_count(nexus_proto_t *nproto);
static int		  shm_get_my_mi_proto(nexus_byte_t **array,
	  			              int *size,
					      void *proto_info,
					      nexus_endpoint_t *endpoint);

static nexus_bool_t	  shm_construct_from_mi_proto(nexus_proto_t **proto,
						   nexus_mi_proto_t *mi_proto,
						   nexus_byte_t *array,
						   int size);
static int		  shm_direct_info_size(void);

static void		  shm_start_send();
static void		  shm_check_outstanding_send();

static shm_proto_t *	  construct_proto(shm_destination_t destination);

static nexus_bool_t	  receive_messages(int blocking);


static void               shm_send(shm_proto_t   * proto, 
                                   nexus_byte_t  * send_data, 
                                   int             send_size);

static nexus_bool_t  shm_received_message();
                     

static void shm_receive(void);

static void shm_local_node_connect(void);
static void shm_local_node_disconnect(void);
static void shm_remote_node_connect(shm_proto_t * proto);
static void shm_remote_node_disconnect(shm_proto_t * proto);

static shm_cell_t * shm_malloc(shm_proto_t * proto, int req_amount); 
static void shm_free(shm_proto_t * proto, shm_cell_t * cell);

static nexus_bool_t shm_pending_heap_operation(shm_segment_t * segment); 

static void shm_local_node_setup(void);
static void shm_remote_node_setup(shm_proto_t * proto);

static void shm_reset_heap(shm_segment_t * segment);
static void shm_reclaim_heap(shm_segment_t * segment);  

static void shm_set_unlocked(shm_proto_t * proto);
static void shm_set_locked(shm_proto_t * proto);
static void shm_calc_alignment(void);

static nexus_proto_funcs_t shm_proto_funcs =
{
    shm_proto_type,
    shm_init,
    shm_shutdown,
    shm_abort,
    shm_poll,
    shm_blocking_poll,
    shm_increment_reference_count,
    shm_decrement_reference_count,
    shm_get_my_mi_proto,
    shm_construct_from_mi_proto,
    NULL /* shm_destroy_my_mi_proto */,
    NULL /* shm_test_proto */,
    shm_send_rsr,
    shm_direct_info_size,
    NULL /* shm_direct_get */,
};


/*****************************************************************/

/*
 * _nx_pr_*_info()
 *
 * Return the nexus_proto_funcs_t function table for this protocol module.
 *
 * This procedure is used for bootstrapping the protocol module.
 * The higher level Nexus code needs to call this routine to
 * retrieve the functions it needs to use this protocol module.
 */
void *_nx_pr_shm_info(void)
{
    return((void *) (&shm_proto_funcs));
} /* _nx_pr_shm_info() */


/*
 * shm_proto_type()
 *
 * Return the nexus_proto_type_t for this protocol module.
 */
static nexus_proto_type_t shm_proto_type(void)
{
    return (NEXUS_PROTO_TYPE_SHMEM);
} /* shm_proto_type() */


#ifdef SHMEM_PROTO_IS_THREAD_SAFE    
/*
 * shm_handler_thread()
 *
 * In the multi-threaded version, this is the entry point
 * for the handler thread.
 */
static void *shm_handler_thread(void *arg)
{
    _nx_set_i_am_shm_handler_thread();
    
    shm_enter();
    receive_messages(BLOCKING);
    shm_exit();

    nexus_mutex_lock(&handler_thread_done_mutex);
    handler_thread_done = NEXUS_TRUE;
    nexus_cond_signal(&handler_thread_done_cond);
    nexus_mutex_unlock(&handler_thread_done_mutex);
    
    return ((void *) NULL);
} /* shm_handler_thread() */
#endif /* SHMEM_PROTO_IS_THREAD_SAFE */


/*
 * shm_init()
 *
 * Initialize the SHMEM protocol.
 */
static void shm_init(int *argc, char ***argv,
		     nexus_bool_t * add_to_my_mi_proto)
)
{
    int arg_num;

    /* Get the default segment size */
    /*
     * This doesn't propatgate to other nodes.
     */
    if ((arg_num = nexus_find_argument(argc, argv, "shmsize", 2)) >= 0)
    {
	shm_segment_size = NEXUS_MAX(atoi((*argv)[arg_num + 1]),
				      SHMEM_MIN_SEGMENT_SIZE);
    }
    else
    { 
	shm_segment_size = SHMEM_DEFAULT_SEGMENT_SIZE;
    }
    _nx_md_gethostname(_nx_hostname_string, MAXHOSTNAMELEN);
    _nx_hostname_string_length = strlen(_nx_hostname_string);

    /* Connect this node to a new shared memory segment  */
    shm_local_node_connect();
    shm_local_node_setup(); 

#ifndef BUILD_LITE
    nexus_thread_key_create(&i_am_shm_handler_thread_key, NULL);
#endif
    proto_table_init();
    nexus_mutex_init(&shm_mutex, (nexus_mutexattr_t *) NULL);
    shm_done = NEXUS_FALSE;
    handle_in_progress = NEXUS_FALSE;

    send_buffer = (struct _nexus_buffer_t *) NULL;
    send_q_head = (struct _nexus_buffer_t *) NULL;
    send_q_tail = (struct _nexus_buffer_t *) NULL;

   
#ifdef SHMEM_PROTO_IS_THREAD_SAFE    
    if (nexus_preemptive_threads())
    {
	nexus_thread_t thread;

	using_handler_thread = NEXUS_TRUE;

	/* NULL out the poll and blocking_poll entries in my function table */
	shm_proto_funcs.poll = NULL;
	shm_proto_funcs.poll_blocking = NULL;
	
	/* Create the handler thread */
	handler_thread_done = NEXUS_FALSE;
	nexus_mutex_init(&handler_thread_done_mutex,
			 (nexus_mutexattr_t *) NULL);
	nexus_cond_init(&handler_thread_done_cond,
			(nexus_condattr_t *) NULL);
	nexus_thread_create(&thread,
			    (nexus_thread_attr_t *) NULL,
			    shm_handler_thread,
			    (void *) NULL);
    }
    else
#endif /* SHMEM_PROTO_IS_THREAD_SAFE */
    {
	using_handler_thread = NEXUS_FALSE;
#ifndef _POSIX_THREAD_PROCESS_SHARED
        shm_proto_funcs.poll_blocking = NULL;
#endif
    }

    *add_to_my_mi_proto = NEXUS_TRUE;
} /* shm_init() */


/*
 * shm_shutdown()
 *
 * This routine is called during normal shutdown of a process.
 */
static void shm_shutdown(nexus_bool_t shutdown_others)
{
    nexus_bool_t i_am_shm_handler_thread;
    nexus_byte_t msg;
    int msg_size = 1;

    shm_enter();
    shm_done = NEXUS_TRUE;

  /*
   * Send shutdown message back to self to complete shutdown process.
   * This is normally only necessary when using an shm handler thread.
   * By always sending this message, we handle rare occasions that the
   * broader shutdown doesn't get it's handler shutdown correctly.
   * When not using an shm handler thread, this message is of no 
   * consequence when the general handler shuts down correctly.
   * Should the general handler not shut down, it will continually 
   * poll this protocol after shutdown has been requested. We'll catch 
   * this abnormal condition here shm_poll.
   */
  msg = CLOSE_SHUTDOWN_FLAG;
  shm_send(my_proto, &msg, msg_size);


    if (using_handler_thread)
    {
	_nx_i_am_shm_handler_thread(&i_am_shm_handler_thread);
	if (!i_am_shm_handler_thread)
	{
	    /* Wait for the handler thread to shutdown */
	    shm_exit();
	    nexus_mutex_lock(&handler_thread_done_mutex);
	    while (!handler_thread_done)
	    {
		nexus_cond_wait(&handler_thread_done_cond,
				&handler_thread_done_mutex);
	    }
	    nexus_mutex_unlock(&handler_thread_done_mutex);
	    shm_enter();
	}
	nexus_mutex_destroy(&handler_thread_done_mutex);
	nexus_cond_destroy(&handler_thread_done_cond);
	using_handler_thread = NEXUS_FALSE;
    }

    shm_local_node_disconnect();
    shm_exit();

} /* shm_shutdown() */


/*
 * shm_abort()
 *
 * This routine is called during the _abnormal_ shutdown of a process.
 */
static void shm_abort(void)
{
  shm_local_node_disconnect(); 

} /* shm_abort() */



     
/*
 * shm_increment_reference_count()
 *
 * Increase the reference count on the associated proto and copy the
 * pointer to the nexus_proto_t
 *
 */
static void shm_increment_reference_count(nexus_proto_t *nproto)
{
    shm_proto_t *proto = (shm_proto_t *) nproto;
    shm_enter();
    proto->reference_count++;
    shm_exit();
 
} /* shm_increment_reference_count() */


/*
 * shm_decrement_reference_count()
 *
 * Decrement the reference count for this proto.  
 *
 * Return NEXUS_TRUE if this function frees the proto.
 */
static nexus_bool_t shm_decrement_reference_count(nexus_proto_t *nproto)
{
    shm_proto_t *proto = (shm_proto_t *) nproto;
    shm_enter();
    proto->reference_count--;
    NexusAssert2((proto->reference_count >= 0),
		 ("shm_decrement_reference_count(): Internal error: Reference count < 0\n"));
    shm_exit();

    return(NEXUS_FALSE);
} /* shm_decrement_reference_count() */


/*
 * shm_get_my_mi_proto()
 *
 * Return the machine independent shm protocol information
 * for this protocol.
 */
static int shm_get_my_mi_proto(nexus_byte_t **array,
			       int *size,
			       void *proto_info,
			       nexus_endpoint_t *endpoint)
{
    int my_size; 
    SHMEM_GET_MY_MI_PROTO_SIZE(my_size);
    *size = (_nx_hostname_string_length + 1 + my_size); \
    NexusMalloc(shm_get_my_mi_proto(), \
		*array, \
		nexus_byte_t *, \
		*size); \
    memcpy(*array, _nx_hostname_string, 
                  _nx_hostname_string_length + 1);
    SHMEM_GET_MY_MI_PROTO(((*array)+_nx_hostname_string_length+1),
                          my_proto->destination);

    return(0);
} /* shm_get_my_mi_proto() */


/*
 * shm_construct_from_mi_proto()
 *
 * From the passed machine independent protocol list ('mi_proto'), plus
 * the shm specific entry from that list ('proto_array' and 'size'),
 * see if I can use the information to create a nexus_proto_t object
 * that can be used to connect to the node:
 *	- If I cannot use this protocol to attach to the node, then
 *		return NEXUS_FALSE.  (This option is useful if two nodes
 *		both speak a particular protocol, but they cannot
 *		talk to each other via that protocol.  For example,
 *              nodes on different host machines.)
 *	- If this shm protocol points to myself, then set
 *		*proto=NULL, and return NEXUS_TRUE.
 *	- Otherwise, construct a shm protocol object for this mi_proto
 *		and put it in *proto.  Then return NEXUS_TRUE.
 */
static nexus_bool_t shm_construct_from_mi_proto(nexus_proto_t **proto,
					       nexus_mi_proto_t *mi_proto,
					       nexus_byte_t *array,
					       int size)
{
    shm_destination_t destination;
    nexus_bool_t result;
    /*
     * Compare the hostname string from the array with mine.
     */
    if (strcmp((char *) array, _nx_hostname_string) != 0)
    {
	return(NEXUS_FALSE);
    }

    /*
     * Extract the shm_destination_t from the array
     */
    SHMEM_CONSTRUCT_FROM_MI_PROTO(destination,
		  	          mi_proto,
			          (array + _nx_hostname_string_length + 1));
    
    /*
     * Test to see if this mi_proto points to myself.
     * If it does, then return *proto=NULL.
     */
    SHMEM_COMPARE_DESTINATIONS(destination, my_proto->destination, result);
    if (result)
    {
	*proto = (nexus_proto_t *) NULL;
    }
    else
    {
	shm_enter();
	*proto = (nexus_proto_t *) construct_proto(destination);
	shm_exit();
        shm_remote_node_setup( (shm_proto_t *) *proto);
    }

    return (NEXUS_TRUE);
} /* shm_construct_from_mi_proto() */



/*
 * shm_direct_info_size()
 */
static int shm_direct_info_size(void)
{
    /* TODO: This needs to be filled in */
    return(0);
} /* shm_direct_info_size() */


/*
 * construct_proto()
 *
 * Construct a shm_proto_t for the given destination. Look up in the
 * proto table to see if one already exists. If it does, bump its reference
 * count and return that one. Otherwise create one, insert into the
 * table with a reference count of 1 and return it.
 */
static shm_proto_t *construct_proto(shm_destination_t destination)
{
    shm_proto_t *proto;
    proto = proto_table_lookup(&destination);
    nexus_debug_printf(3,
		       ("construct_proto(): Table lookup returns proto=%x\n",
			proto));
    if (proto == (shm_proto_t *) NULL)
    {
	NexusMalloc(construct_proto(), proto, shm_proto_t *,
		    sizeof(shm_proto_t));

	proto->type = NEXUS_PROTO_TYPE_SHMEM;
	proto->funcs = &shm_proto_funcs;
	proto->direct_custom_min_size = NEXUS_DC_MAX_U_LONG;
	proto->direct_custom_max_size = NEXUS_DC_MAX_U_LONG;
	proto->direct_pointer_min_size = NEXUS_DC_MAX_U_LONG;
	proto->direct_pointer_max_size = NEXUS_DC_MAX_U_LONG;
	proto->can_use_iovec = NEXUS_FALSE;
	proto->reserved_header_size = 0;
	proto->reference_count = 1;

	SHMEM_COPY_DESTINATION(proto->destination, destination);
        proto->dest_handle = SHMEM_FAILURE;
        proto->dest_segment= NULL;

	proto_table_insert(proto);
    }
    else
    {
	proto->reference_count++;
    }
	

    return (proto);
} /* construct_proto() */


/*
 * proto_table_init()
 *
 * Initialize the protocol table.
 */
static void proto_table_init(void)
{
    int i;

    for (i = 0; i < PROTO_TABLE_SIZE; i++)
    {
	proto_table[i].proto = (shm_proto_t *) NULL;
	proto_table[i].next = (proto_table_entry_t *) NULL;
    }
} /* proto_table_init() */


/*
 * proto_table_insert()
 *
 * Insert the given proto into the table, hashing on its destination.
 *
 * We assume that the entry is not present in the table.
 */
static void proto_table_insert(shm_proto_t *proto)
{
    int bucket;
    proto_table_entry_t *new_ent;

    SHMEM_HASH_DESTINATION(proto->destination, bucket);

    if (proto_table[bucket].proto == (shm_proto_t *) NULL)
    {
	/* Drop it into the preallocated table entry */
	proto_table[bucket].proto = proto;
    }
    else
    {
	/*
	 * Need to allocate a new proto_table_entry_t and add it
	 * to the bucket
	 */
	NexusMalloc(proto_table_insert(),
		    new_ent,
		    proto_table_entry_t *,
		    sizeof(struct _proto_table_entry_t));

	new_ent->proto = proto;
	new_ent->next = proto_table[bucket].next;

	proto_table[bucket].next = new_ent;
    }

} /* proto_table_insert() */


/*
 * proto_table_lookup()
 *
 * Look up and return the shm_proto_t for the given destination.
 * Return NULL if none exists.
 */
static shm_proto_t *proto_table_lookup(shm_destination_t *dest)
{
    proto_table_entry_t *ent;
    int bucket;
    nexus_bool_t result;

    SHMEM_HASH_DESTINATION(*dest, bucket);

    for (ent = &(proto_table[bucket]);
	 ent != (proto_table_entry_t *) NULL;
	 ent = ent->next)
    {
	if (ent->proto != (shm_proto_t *) NULL)
	{
	    SHMEM_COMPARE_DESTINATIONS(*dest, ent->proto->destination, result);
	    if (result)
	    {
		return (ent->proto);
	    }
	}
    }
    
    return ((shm_proto_t *) NULL);
} /* proto_table_lookup() */


/*
 * shm_poll()
 *
 * In a version of the shm protocol module that does not
 * use a handler thread (preemptive thread module & thread safe blocking
 * receives), this routine should check to see if there are
 * any messages to receive, and if so then receive them and invoke them.
 */
static nexus_bool_t shm_poll(void)
{
    nexus_bool_t message_handled = NEXUS_FALSE;
    /*
     * This should not be called if a separate handler thread is in use.
     */
    NexusAssert2((!using_handler_thread),
		 ("shm_poll(): Internal error: Should never be called when using a handler thread\n") );
    
    nexus_debug_printf(5, ("shm_poll(): entering\n"));
    
    shm_enter();
    if (send_buffer)
    { 
	shm_check_outstanding_send();
    }
    if (!handle_in_progress)
    { 
	message_handled = receive_messages(NON_BLOCKING);
    }
 
    /*
     * Handle exception of general handler thread not shutdown properly
     * by catching our own shutdown message.  Under normal shutdown, 
     * this message will be "lost" during shutdown, and never received.
     * If the handler thread is still polling to receive after shutdown,
     * we've hit the very infrequent, but reproducable, case of a
     * "runaway" handler.
     */
    else if (shm_done && shm_received_message()) 
    {
	shm_receive();
        if ( (int) *receive_buffer == CLOSE_SHUTDOWN_FLAG);
	{
          nexus_abort();
	}
    }
    shm_exit();

#ifndef BUILD_LITE
    /*
     * Only yield the processor if there was a message handled.
     * That handler may have enabled another thread for execution.
     */
    if (message_handled)
    {
	nexus_thread_yield();
    }
#endif
    
    nexus_debug_printf(5, ("shm_poll(): exiting\n"));


    return (message_handled);
} /* shm_poll() */


/*
 * shm_blocking_poll()
 *
 * In a version of the shm protocol module that does not
 * use a handler thread, this routine should do a blocking
 * message receive and invoke the handler for the message.
 * This is only called if this is the only protocol module
 * requiring polling, and all other threads are suspended.
 *
 * It is ok for this to return without actually handling
 * a message.
 *
 * Note: There is no need to call nexus_thread_yield() from this
 * routine like there is in shm_poll(), because this routine will only
 * be used from an idle or handler thread which will do a subsequent
 * nexus_thread_yield() anyway.
 */
static void shm_blocking_poll(void)
{
    /*
     * This should not be called if a separate handler thread is in use.
     */
    NexusAssert2((!using_handler_thread),
		 ("shm_blocking_poll(): Internal error: Should never be called when using a handler thread\n") );

    nexus_debug_printf(5, ("shm_blocking_poll(): entering\n"));

    if (send_buffer)
    {
	nexus_debug_printf(5, ("shm_blocking_poll(): switching to shm_poll() until all outstanding sends are delivered\n"));
	shm_poll();
    }

    shm_enter();
    if (!handle_in_progress)
    {
	receive_messages(BLOCKING);
    }
    shm_exit();
    nexus_debug_printf(5, ("shm_blocking_poll(): exiting\n"));

} /* shm_blocking_poll() */


/*
 * shm_send_rsr()
 */
static int shm_send_rsr(struct _nexus_buffer_t *buffer)
{
    nexus_debug_printf(2,("shm_send_rsr(): invoked with buffer: %x\n",buffer));

    if (buffer->n_direct > 0)
    { 
	nexus_mutex_init(&(buffer->barrier.mutex),(nexus_mutexattr_t *) NULL);
	nexus_cond_init(&(buffer->barrier.cond), (nexus_condattr_t *) NULL);
	buffer->barrier.count = buffer->n_direct;
	buffer->using_barrier = NEXUS_TRUE;
    }

    shm_enter();
    
    if (send_buffer)
    { 
	/*
	 * There is a send outstanding.
	 * So enqueue this message.
	 * It will be sent either now or later
	 * from shm_check_outstanding_send().
	 */
	Enqueue(send_q_head, send_q_tail, buffer);
	shm_check_outstanding_send();
    }
    else
    {
	/*
	 * There is no send outstanding.
	 * So starting send this buffer.
	 */
	send_buffer = buffer;
	shm_start_send();
	shm_check_outstanding_send();
    }

    shm_exit();
    return(0);
    
} /* shm_send_rsr() */


/*
 * shm_start_send()
 *
 * Start the send for the buffer in the 'send_buffer' global variable.
 *
 * Send the base_segment of the buffer.
 *
 * Set the various 'send_*' global variables so that the send
 * can be continued by shm_check_outstanding_send().
 */
static void shm_start_send()
{
    shm_proto_t *proto;
    nexus_byte_t *send_bytes;
    unsigned long send_size;

    proto = (shm_proto_t *) send_buffer->proto;
    send_size = send_buffer->base_segments->size_used;
    send_bytes = send_buffer->base_segments->current;
    send_buffer->current_base_segment = (nexus_base_segment_t *) NULL;
    
    shm_send(proto,
	     send_bytes,
             send_size);

} /* shm_start_send() */


/*
 * shm_check_outstanding_send()
 *
 * Check the status of the current outstanding send.  If it is
 * complete, then post the next send if there is one.
 * The next send may be either a direct component of send_buffer,
 * or the next message in the send_q.
 *
 * Assumptions:
 *   - send_buffer != NULL
 *   - A send has already been posted for send_buffer
 *   - shm_enter() has already been called
 */
static void shm_check_outstanding_send()
{
    nexus_bool_t done;
    nexus_byte_t *send_bytes;
    unsigned long send_size;
    shm_proto_t *proto;
    nexus_bool_t send_status;

    done = NEXUS_FALSE;
    while (!done)
    {
      if (sending)
        done = NEXUS_TRUE; 
      else
      {

      /*
       * Figure out what the next send is, and post it.
       */

       if ( send_buffer->direct_segments
	     && (send_buffer->direct_segments->n_left > 0))
       {
	   /*
	    * The previous segment (base or direct) has been sent,
	    * and there is at least one more direct segment to send.
	    */
	   nexus_direct_info_t *direct_info;
	
	   nexus_debug_printf(1, ("shm_check_outstanding_send(): begin sending buffer direct segment 0x%lx\n", (unsigned long) send_buffer));

	   direct_info = send_buffer->direct_segments->current;
	   send_buffer->direct_segments->current++;
	   send_buffer->direct_segments->n_left--;

	   proto = (shm_proto_t *) send_buffer->proto;
	   send_bytes = direct_info->data;
	   send_size = direct_info->size;

	   shm_send(proto,
	            send_bytes,
                    send_size); 
       }
       else
       {
	   /*
	    * We are done with this buffer.
	    * So free the buffer, and go on to the next one
	    * in the send_q.
	    */

	   if (send_buffer->using_barrier)
	   {
	       /*
	        * Signal the thread waiting on this buffer.
	        */
	       nexus_mutex_lock(&(send_buffer->barrier.mutex));
	       send_buffer->barrier.count--;
	       nexus_cond_signal(&(send_buffer->barrier.cond));
	       nexus_mutex_unlock(&(send_buffer->barrier.mutex));
	   }
	   else
	   {
	       /*
	        * There is no thread waiting on this buffer.
	        * So destroy the buffer.
	        */
	       nexus_buffer_destroy(&send_buffer);
	    }

	   if (QueueNotEmpty(send_q_head))
	   {
	       /* Grab the next buffer and start sending it */
	       Dequeue(send_q_head,
	   	       send_q_tail,
		       send_buffer);
	       shm_start_send();
	   }
	   else
	   {
	       /* Nothing else waiting to be sent */
	       send_buffer = (struct _nexus_buffer_t *) NULL;
	       done = NEXUS_TRUE;
	   }
		
	 } 
       }

    }

} /* shm_check_outstanding_send() */



/*
 * receive_messages();
 *
 * Receive all pending messages to this node.
 *
 * If blocking==BLOCKING, use a blocking receive.
 * If blocking==NON_BLOCKING, use a non-blocking receive.  Do not
 * return until the non-blocking receive fails.
 *
 * Note: If using_handler_thread==NEXUS_TRUE,
 *	 this routine will never be called by anyone except the
 *	 handler thread.  (Though it may be called recursively by the
 *	 the handler thread, if a send_rsr is done from within a handler.)
 *
 * Return: NEXUS_TRUE if a message is handled, otherwise NEXUS_FALSE
 */
static nexus_bool_t receive_messages(int blocking)
{
    nexus_bool_t done = NEXUS_FALSE;
    nexus_bool_t message_handled = NEXUS_FALSE;
    nexus_bool_t need_to_relock;
    nexus_byte_t *b;
    int format;
    struct _nexus_buffer_t *buffer;
    nexus_bool_t message_received;

    nexus_debug_printf(5, ("receive_messages(): entering\n"));
    
    while (!done)
    {
	if (blocking)
	{ 
	    if (using_handler_thread)
	    {
		/*
		 * This is the handler thread.  Since it is thread safe,
		 * we should do an shm_exit() so that other threads can
		 * do sends while this thread is blocked in the receive.
		 */
		need_to_relock = NEXUS_TRUE;
		shm_exit();
	    }
	    else
	    {
		need_to_relock = NEXUS_FALSE;
		done = NEXUS_TRUE; /* Do the blocking receive just once */
	    }

            /* "blocking" for receive - blocking is presently defeated */
	    /* todo : create "true" blocking via process shared mutex
                      condition.  This spin-blocking is not used without
                      process shared thread constructs by nulling out the
                      blocking in shm_init 
            */
            message_received = NEXUS_FALSE;
            while(shm_done == NEXUS_FALSE && message_received == NEXUS_FALSE) 
	    { 
               message_received = shm_received_message();
               if (message_received== NEXUS_FALSE)
	       {
                  nexus_thread_yield();
	       }
	    }


	    if (need_to_relock)
	    {
		shm_enter();
	    }
	} 
        else
	{ 
            message_received = shm_received_message(); 
	}


	if (shm_done)
	    return (message_handled);


	/* message received and ready to process */
	
	if (message_received)
	{ 
            shm_receive();

	    /*
	     * We received a message into 'receive buffer'.
	     */
	    b = receive_buffer;
	    format = (int) *b++;

	    if (format < NEXUS_DC_FORMAT_LAST)
	    {  
		handle_in_progress = NEXUS_TRUE;

		shm_exit();

		_nx_buffer_create_from_raw(receive_buffer,
					   receive_buffer_size,
					   0,
					   receive_buffer_size,
                                           NULL,
                                           &buffer);

		nexus_debug_printf(2,("pr_shm.c:receive_messages(): dispatching message\n"));
		_nx_buffer_dispatch(buffer);
		nexus_debug_printf(2,("pr_shm.c:receive_messages(): message dispatch complete\n"));

		shm_enter();
		handle_in_progress = NEXUS_FALSE;
		message_handled = NEXUS_TRUE;
	    }

            else if (format == CLOSE_SHUTDOWN_FLAG)
	    { 
               _nx_exit_transient_process(0);
	    }

	    else
	    {
		nexus_fatal("receive_messages(): Got unknown control message\n");
	    }
	}
	else
	{
	    /* The (non-blocking) receive didn't receive anything, */
	    done = NEXUS_TRUE;
	}

    }

    nexus_debug_printf(5,
		       ("receive_messages(): returning message_handled=%d\n",
			message_handled) );

    return (message_handled);
    
} /* receive_messages() */



/******************************************************************
 * Shared memory communication 
 ******************************************************************/

/* 
 * Send message by allocating shared memory in reciever's segment
 * and copying the data into it.  To avoid going over the maximum
 * number of segments that can be attached, we'll attach and detach
 * the target's segment for the actual "send".  Shared memory 
 * allocation will be retried if unavailable to give the receiver
 * a chance to free up some space.
 */
static void shm_send(shm_proto_t   * proto,
                    nexus_byte_t  * send_data, 
                    int             send_size)  
{  
    shm_cell_t * send_cell = NULL;
    int tries = 0; 

    shm_remote_node_connect(proto);

    /* allocated space in receivers segment, retrying if necessary
     * until either receiver frees up storage, or we decide the 
     * storage will never be available.
     */
    for (; tries < SHMEM_MALLOC_MAX_TRIES && send_cell == NULL; tries++)
    {
        send_cell = shm_malloc(proto, send_size);  
        if (send_cell == NULL) 
        { 
           SHMEM_THREAD_SCHED_YIELD();
        }
     }

    /* "send" the actual data by copying into the receiver's segment */ 
    if (send_cell != NULL)  
    {  
        memcpy(&send_cell->cell_data, send_data, send_size);  
        send_cell->ref_count = SHMEM_RECV_WAIT;  
        sending = NEXUS_FALSE;  
    }  
    else  
    {  
        nexus_fatal("shm_send() Couldn't allocate send buffer\n");
    }  

    shm_remote_node_disconnect(proto);


} /* shm_send() */


 
/* 
 * See if someone's sent us a message in our segment
 * by traversing list of active messages to see if 
 * there's one ready to receive.
 *
 * Note that it is possible that we'll check for a
 * new receive while the heap is being reclaimed.
 * Since the "old" heap is left untouched, we can
 * still use it to determine if there is a message
 * to actually receive.  The receive operation itself
 * will lock the heap and find the actual location of
 * the new message we found here.
 */
static nexus_bool_t shm_received_message(void) 
{  
    nexus_bool_t message_received = NEXUS_FALSE;
    shm_segment_t * segment;
    shm_heap_t    * cur_heap;  
    shm_cell_t    * cur_cell;  

    segment = my_proto->dest_segment;
    SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
    SHMEM_CELL_ADDR(segment, cur_heap->alloc_cell_head, cur_cell);  
   
    while(message_received == NEXUS_FALSE && cur_cell != NULL)
    {  
        if (cur_cell->ref_count == SHMEM_RECV_WAIT)  
        {  
            message_received = NEXUS_TRUE;  
	}  
        else  
        {  
            SHMEM_CELL_ADDR(segment, cur_cell->next_cell, cur_cell);  
	}  
    }  

  return(message_received);

} /* shm_received_message */


/* 
 * Receive message placed in our shared memory segment by locating
 * it in our list of active messages and copying it to local heap
 * space.  While this operation does not require that we lock the
 * heap until we actually free, we'll try to avoid receiving during
 * heap reclaimation.  While it's not absolutly essential, both
 * receive and heap reclaimatin operations will try to avoid each
 * other.  The few receives that actually do slip through and occur
 * during reclaimation will simply copy their data out of the "old"
 * heap.
 */
static void shm_receive(void)
{  
    shm_segment_t * segment;
    shm_heap_t    * cur_heap;  
    shm_cell_t    * recv_cell;  
   
    receive_buffer = NULL;  
    receive_buffer_size = 0; 
   
    segment = my_proto->dest_segment;
    while (segment->reclaiming)  
    {  
        SHMEM_THREAD_SCHED_YIELD();  
    }  

    segment->receiving = NEXUS_TRUE;

    /* find message to receive */  
    SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
    SHMEM_CELL_ADDR(segment, cur_heap->alloc_cell_head, recv_cell);  

    while(recv_cell != NULL && recv_cell->ref_count != SHMEM_RECV_WAIT)  
    {  
       SHMEM_CELL_ADDR(segment, recv_cell->next_cell, recv_cell);  
    }  
   
    /* move message to receive from shared space to local nexus space */
    if (recv_cell != NULL)  
    { 
       recv_cell->ref_count = SHMEM_RECEIVING;
       receive_buffer_size = recv_cell->data_size;  
       NexusMalloc(shm_receive(), 
                   receive_buffer, 
                   nexus_byte_t *, receive_buffer_size);  

       memcpy(receive_buffer, 
              &recv_cell->cell_data, 
              receive_buffer_size);

       segment->receiving = NEXUS_FALSE;     
       shm_free(my_proto, recv_cell);  
    }  
    else
    {
       segment->receiving = NEXUS_FALSE;     
    }


} /* shm_receive() */



/*
 * Prepare the process for communication via shared memory.
 * The unique destination identifier of this node will be
 * established here as the common ipc key used for shared
 * memory and the inter-process semaphore lock for it.
 * By allocating these system resources in this fashion,
 * we'll enable any process to access them via this destination
 * id, which we'll load into our proto connection array.
 */
static void shm_local_node_connect(void)
{  
    shm_destination_t  new_node = 0;
    shm_destination_t  my_dest;
    union semun        sem_info;   
    int                actual_size; 

    /*
     * the actual amount of shared memory we'll need must
     * allow us to maintain two heap spaces for reclaimation,
     * room for the shared memory segment control header,
     * as well as be sized on an even boundary of this 
     * machine's memory alignment.
     */
    shm_calc_alignment();
    if (shm_alignment != 0)
    {
      shm_segment_size += shm_alignment;
      shm_segment_size -= (shm_segment_size % shm_alignment);
    } 
    actual_size = 2 * shm_segment_size + sizeof(shm_segment_t);  

    /*
     * find a unique, unused set of shared memory and semaphore resources
     * keyed on the same id.
     */
    my_proto = construct_proto(new_node);
    my_proto->dest_handle = SHMEM_FAILURE;  
 
    for(new_node=0;  
        my_proto->dest_handle == SHMEM_FAILURE && new_node <= UCHAR_MAX;  
        new_node++)  
    {  
       my_dest = (new_node == 0) ? _nx_md_getpid() 
                                 : ftok(_nx_executable_name(), new_node);
 
       SHMEM_CREATE(my_proto->dest_handle, my_dest, 
                    actual_size, SHMEM_CREATE_PERMS);

       if (my_proto->dest_handle != SHMEM_FAILURE)  
       {  
           my_proto->dest_lock = semget(my_dest, 1, SHMEM_CREATE_PERMS);  
           if (my_proto->dest_lock == SHMEM_FAILURE)  
	   {  
               SHMEM_DESTROY(my_proto->dest_handle);
               my_proto->dest_handle = SHMEM_FAILURE;  
	   }  
       }  
    }  

    /*
     * with resources allocated, try to attach to the shared 
     * memory segment acquired.
     */
    SHMEM_COPY_DESTINATION(my_proto->destination, my_dest);
  
    if (my_proto->dest_handle == SHMEM_FAILURE)  
    {  
        nexus_fatal("shm_initial_node_connect(): Couldn't get segment \n");  
    }  
    else  
    {  
        SHMEM_ATTACH(my_proto->dest_segment, my_proto->dest_handle);
        if (my_proto->dest_segment == (shm_segment_t *) SHMEM_FAILURE)  
	{  
            SHMEM_DESTROY(my_proto->dest_handle);
            sem_info.val = 0;
            semctl(my_proto->dest_lock, 0, IPC_RMID, sem_info);  
            nexus_fatal("shm_initial_node_connect(): Couldn't attach segment\n");  
	}  
    }  

} /* shm_local_node_connect() */


/*
 * "Connect" to another process by attaching
 *  to the other's shared memory segment.
 */
static void shm_remote_node_connect(shm_proto_t * proto)
{  
   if ( proto->dest_segment == NULL) {

      SHMEM_CREATE(proto->dest_handle, proto->destination, 
                  0, SHMEM_CONNECT_PERMS);

      if (proto->dest_handle == SHMEM_FAILURE)  
      {  
        nexus_fatal("shm_connect(): Couldn't get shared memory segment\n");  
      }  

      SHMEM_ATTACH(proto->dest_segment, proto->dest_handle);
      if (proto->dest_segment == (shm_segment_t *) SHMEM_FAILURE)  
      {  
        proto->dest_segment = NULL; 
        nexus_fatal("shm_connect(): Couldn't attach shared memory\n");  
      }  
   }
} /* shm_remote_node_connect() */


/*
 * "Disconnect" this process by releasing it's shared memory 
 * resources.  This process will no longer communicate via shm. 
 */
static void shm_local_node_disconnect(void)
{  
    if (my_proto->dest_segment != NULL)  
    { 
       semctl(my_proto->dest_lock, 0, IPC_RMID, 0);
       SHMEM_DESTROY(my_proto->dest_handle);
    }

} /* shm_local_node_disconnect() */


/*
 * "Disconnect" from the specified process
 * by detaching it's shared memory segment.
 */
static void shm_remote_node_disconnect(shm_proto_t * proto)
{  
    if (proto->dest_segment != NULL && proto != my_proto)  
    {  
       shmdt((char*) proto->dest_segment);
       proto->dest_handle  = SHMEM_FAILURE;
       proto->dest_segment = NULL;
    } 

} /* shm_remote_node_disconnect() */


/******************************************************************
 * Shared memory storage management
 ******************************************************************/

/* 
 * Fill storage request for shared memory from a specified node.  
 * Note that requests for more storage than the shared memory was
 * allocated for are ignored. todo: handle this case.
 */
static shm_cell_t * shm_malloc(shm_proto_t * proto, int req_amount) 
{  
    shm_segment_t * segment;
    shm_heap_t    * cur_heap;  
    shm_cell_t    * new_cell = NULL;
    shm_cell_t    * old_tail;  
    int actual_req_amount, pad_amt;  
   
    /*
     * actual memory request accounts for memory management 
     * control space as well as padding for this machine's
     * memory alignment requirements.
     */
    actual_req_amount = req_amount + sizeof(shm_cell_t) - 1;  
    if (shm_alignment != 0)
    {
       pad_amt = shm_alignment - (actual_req_amount % shm_alignment);  
       actual_req_amount += pad_amt;  
    }

    segment = proto->dest_segment;
    shm_set_locked(proto);  

    SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
   
    /*
     * reclaim/garbage collect/cleanup heap if it
     * will allow us to fill this request.
     */
    if (actual_req_amount > cur_heap->free_stg_amt
        && actual_req_amount <= cur_heap->unref_stg_amt)  
    {  
      shm_reclaim_heap(segment);  
      SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
    }  
   
    /*
     * allocate new heap cell for storage request by adding
     * a new link into our heap list of active cells.
     * newly allocated cells are marked with a reference
     * status of "sending" and will be considered active 
     * until they have been received and set to "unreferenced".
     */
    if (actual_req_amount <= cur_heap->free_stg_amt)  
    { 
        SHMEM_CELL_ADDR(segment, cur_heap->free_stg_loc, new_cell);  
        new_cell->prev_cell = cur_heap->alloc_cell_tail;  
        new_cell->next_cell = NULL_OFFSET;  
        new_cell->this_cell = cur_heap->free_stg_loc;  
        new_cell->ref_count = SHMEM_SENDING;  
        new_cell->cell_size = actual_req_amount;  
        new_cell->data_size = req_amount;  
        new_cell->cell_data = ' ';  
        new_cell->reclaimed = NEXUS_FALSE;  
  
        if (cur_heap->alloc_cell_tail != NULL_OFFSET)  
        {  
           SHMEM_CELL_ADDR(segment, cur_heap->alloc_cell_tail, old_tail);
           old_tail->next_cell = cur_heap->free_stg_loc;
	}     

        if (cur_heap->alloc_cell_head == NULL_OFFSET)  
	{  
           cur_heap->alloc_cell_head = cur_heap->free_stg_loc;  
	}
        cur_heap->alloc_cell_tail = cur_heap->free_stg_loc;  
        segment->alloc_count++;
    
        cur_heap->free_stg_loc += actual_req_amount;  
        cur_heap->free_stg_amt -= actual_req_amount;  
    }
   
    sending = NEXUS_TRUE;  
    shm_set_unlocked(proto);   

    return(new_cell);

} /* shm_malloc() */



/*
 * Remove specified shared memory heap cell from actively referenced
 * list of cells.
 */
static void shm_free(shm_proto_t * proto, shm_cell_t * free_cell)  
{  
    shm_segment_t * segment;
    shm_heap_t    * cur_heap;  
    shm_cell_t    * prev_cell, * next_cell;  

    free_cell->ref_count = SHMEM_UNREFERENCED;
    segment = proto->dest_segment;
    shm_set_locked(proto); 

    /* if cell wasn't reclaimed during receive copying, free it */
    if (free_cell->reclaimed == NEXUS_FALSE)
    {
      /* this was last active cell, clear heap */
      if (segment->alloc_count == 1) 
      {  
         shm_reset_heap(segment);  
      }  

      /* remove cell from heap list of active cells */
      else  
      { 
         segment->alloc_count--;
 
         SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);
         SHMEM_CELL_ADDR(segment, free_cell->prev_cell, prev_cell);
         SHMEM_CELL_ADDR(segment, free_cell->next_cell, next_cell);

         cur_heap->unref_stg_amt += free_cell->cell_size;  

         /* cell not at head of list, adjust prior cell's next */
         if (free_cell->prev_cell != NULL_OFFSET)  
         {  
            prev_cell->next_cell = free_cell->next_cell;  
         }  
         else  /* cell was head, reset allocated cell list head */ 
	 { 
            cur_heap->alloc_cell_head = free_cell->next_cell;  
	 }  

   
         /* cell not at tail of list, adjust next cell's prior */
         if (free_cell->next_cell != NULL_OFFSET)  
         {  
	    next_cell->prev_cell = free_cell->prev_cell;  
         }  
         else /* cell was tail, adjust list tail and try to reclaim */  
         {
	    cur_heap->alloc_cell_tail = free_cell->prev_cell;  

	   /* reclaim up to end of previous cell's allocated storage */
            if (free_cell->prev_cell != NULL_OFFSET)   
	    {  
                cur_heap->free_stg_loc = prev_cell->this_cell +  
                                         prev_cell->cell_size ;  
                cur_heap->free_stg_amt = segment->heap_size -  
	                                 cur_heap->free_stg_loc;  
	    }
         }
      }  
    }  

    shm_set_unlocked(proto);  

} /* shm_free() */


/* 
 * Determine if shared memory is actively referenced, 
 * and can not presently be reclaimed.
 */
static nexus_bool_t shm_pending_heap_operation(shm_segment_t * segment) 
{  
   shm_heap_t * cur_heap;  
   shm_cell_t * heap_cell;  
   nexus_bool_t is_active = NEXUS_FALSE;   

   SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
   SHMEM_CELL_ADDR(segment, cur_heap->alloc_cell_head, heap_cell);  
   
   while (is_active == NEXUS_FALSE && heap_cell != NULL)  
   { 
       if (segment->receiving || heap_cell->ref_count >= SHMEM_RECEIVING)  
       {  
           is_active = NEXUS_TRUE;  
       }  
       else  
       {  
           SHMEM_CELL_ADDR(segment, heap_cell->next_cell, heap_cell);  
       }  
   }  

  return(is_active);

} /* shm_pending_heap_operations() */



/* 
 * Reclaim current heap by copying into inactive heap and making it active.
 * While we may be under heap lock, preventing further mallocs and sends,
 * we've only "discouraged" current receives with a soft lock.
 * Some "receives" may start copying their messages at the same time that
 * we begin copying the heap to reclaim.  As these few concurrent recieves
 * will copy out of the old heap, we'll skip reclaiming them and mark 
 * their old heap cells as reclaimed.  This instructs the receive
 * operation's free that there is no action required for this reclaimed
 * cell.
 */
static void shm_reclaim_heap(shm_segment_t * segment)  
{  
    shm_heap_t * cur_heap, * next_heap;  
    shm_cell_t * cur_cell, * prev_cell, * new_cell;  
    shm_offset_t next_heap_offset, new_cell_offset, prev_cell_offset;  
    int new_cell_size, new_pad_amt;  
 
    segment->reclaiming = NEXUS_TRUE;  

    /* stall while recieves finish up */
    while( shm_pending_heap_operation(segment) ) 
    { 
       SHMEM_THREAD_SCHED_YIELD();
    }  
     
    /* setup next heap, well flip to this when reclaimation is done */
    if (segment->cur_heap == segment->top_heap)  
    {  
        next_heap_offset = segment->bot_heap;  
        SHMEM_HEAP_ADDR(segment, segment->bot_heap, next_heap);  
    }  
    else  
    {  
        next_heap_offset = segment->top_heap;  
        SHMEM_HEAP_ADDR(segment, segment->top_heap, next_heap);  
    }  
   
    /* clear next heaps control header */
    next_heap->free_stg_loc  = next_heap_offset + sizeof(shm_heap_t);  
    next_heap->free_stg_amt  = segment->heap_size - sizeof(shm_heap_t);  
    next_heap->unref_stg_amt = 0;
    next_heap->alloc_cell_head = NULL_OFFSET;  
    next_heap->alloc_cell_tail = NULL_OFFSET;  

    /* traverse and reclaim current heap's active cells */
    SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
    SHMEM_CELL_ADDR(segment, cur_heap->alloc_cell_head, cur_cell);  
    prev_cell_offset = NULL_OFFSET;  
    segment->alloc_count = 0;
    while(cur_cell != NULL)  
    {  
      cur_cell->reclaimed = NEXUS_TRUE;
      if (cur_cell->ref_count == SHMEM_RECV_WAIT)
	{
            segment->alloc_count++;
            new_cell_offset = next_heap->free_stg_loc;  
            SHMEM_CELL_ADDR(segment, new_cell_offset, new_cell);  

	    /* we must realign reclaimed cells */
            new_cell_size = cur_cell->data_size + sizeof(shm_cell_t);
            if (shm_alignment != 0)
            {
              new_pad_amt   = shm_alignment - (new_cell_size % shm_alignment);
              new_cell_size += new_pad_amt;
	    }
            memcpy(new_cell, cur_cell, new_cell_size);

            new_cell->cell_size = new_cell_size;   
            new_cell->next_cell = NULL_OFFSET;  
            new_cell->prev_cell = prev_cell_offset;  
            new_cell->this_cell = new_cell_offset;
            new_cell->reclaimed = NEXUS_FALSE;
            if (prev_cell_offset != NULL_OFFSET)  
	    {  
                SHMEM_CELL_ADDR(segment, prev_cell_offset, prev_cell);  
                prev_cell->next_cell = new_cell_offset;  
     	    }  
            prev_cell_offset = new_cell_offset;  
   
            next_heap->free_stg_loc   += cur_cell->cell_size;  
            next_heap->free_stg_amt   -= cur_cell->cell_size;  
            next_heap->alloc_cell_tail = new_cell_offset;   
            if (next_heap->alloc_cell_head == NULL_OFFSET)  
	    {  
                next_heap->alloc_cell_head = new_cell_offset;  
	    }  
	}
            SHMEM_CELL_ADDR(segment, cur_cell->next_cell, cur_cell);  
   }  
   
    segment->cur_heap = next_heap_offset;  
    segment->reclaiming = NEXUS_FALSE;  

} /* shm_reclaim_heap() */


/*  
 * Clear current shared memory heap by resetting control values.
 */
static void shm_reset_heap(shm_segment_t * segment)  
{  
   shm_heap_t * cur_heap;  
   segment->alloc_count = 0;  
   SHMEM_HEAP_ADDR(segment, segment->cur_heap, cur_heap);  
   cur_heap->free_stg_loc  = segment->cur_heap  + sizeof(shm_heap_t);  
   cur_heap->free_stg_amt  = segment->heap_size - sizeof(shm_heap_t);  
   cur_heap->unref_stg_amt = 0;
   cur_heap->alloc_cell_head = NULL_OFFSET;  
   cur_heap->alloc_cell_tail = NULL_OFFSET;  

} /* shm_reset_heap() */


/* 
 * One time setup of this local node's shared memory communication resources.
 */
static void shm_local_node_setup(void)
{
    union semun sem_info; 
    shm_segment_t * segment;
    segment = my_proto->dest_segment;
    segment->soft_lock = SHMEM_UNLOCKED_VALUE;
  
    sem_info.val = SHMEM_UNLOCKED_VALUE;  
    semctl(my_proto->dest_lock, 0, SETVAL, sem_info);
    SHMEM_THREAD_LOCK_INIT(segment);  
  
    segment->heap_size   = shm_segment_size;  
    segment->reclaiming  = NEXUS_FALSE;  
    segment->receiving   = NEXUS_FALSE;  
    segment->top_heap    = sizeof(shm_segment_t);  
    segment->bot_heap    = segment->top_heap + shm_segment_size;  
   
    segment->cur_heap = segment->bot_heap;  
    shm_reset_heap(segment);  
   
    segment->cur_heap = segment->top_heap;  
    shm_reset_heap(segment);  

} /* shm_local_node_setup() */


/*
 * One time setup of remote node's communication linkage.
 */
static void shm_remote_node_setup(shm_proto_t * proto)
{
  proto->dest_lock = semget(proto->destination, 0, 0);
  if (proto->dest_lock == SHMEM_FAILURE)
  {
    nexus_fatal("shm_remote_node_setup() - couldn't get semaphore\n");
  }
} /* shm_remote_node_setup() */




/******************************************************************
 * Shared memory locking support
 ******************************************************************/

/* 
 * SHMEM_SET_LOCKED(shm_segment_t * Segment)
 *
 * Lock shared memory segment heap control header.
 */
static void shm_set_locked(shm_proto_t * proto)  
{  
   int soft_lock_probes;  
   struct sembuf full_lock_req;  
   shm_segment_t * segment;
   
   segment = proto->dest_segment;
   SHMEM_THREAD_SET_LOCKED(my_proto->dest_segment);   
  
   for (soft_lock_probes=0;  
        segment->soft_lock == SHMEM_LOCKED_VALUE  
          && soft_lock_probes < SHMEM_LOCK_PROBE_LIMIT;  
        soft_lock_probes++)  
   {  
       SHMEM_THREAD_SCHED_YIELD();  
   }  

   segment->soft_lock = SHMEM_LOCKED_VALUE;  
   full_lock_req.sem_num = 0;  
   full_lock_req.sem_op  = SHMEM_LOCK_OP;  
   full_lock_req.sem_flg = 0;
   semop(proto->dest_lock, &full_lock_req, 1); 

} /* shm_set_locked() */



/* 
 * Release shared memory segment heap control header.
 */
static void shm_set_unlocked(shm_proto_t * proto)  
{  
   struct sembuf full_lock_req;  
   full_lock_req.sem_num = 0;  
   full_lock_req.sem_op  = SHMEM_UNLOCK_OP;  
   full_lock_req.sem_flg = 0;
   semop(proto->dest_lock, &full_lock_req, 1);  
   proto->dest_segment->soft_lock = SHMEM_UNLOCKED_VALUE;  
   SHMEM_THREAD_SET_UNLOCKED(my_proto->dest_segment);   

} /* shm_set_unlocked() */



/*
 *  Determine run time memory alignment to use for shared memory management.
 */
static void shm_calc_alignment(void)
{
  struct aligned_struct
  {
     char shortest_field;
      
     union longest_field
     {
       long   long_field;
       long  *ptr_field;
     } longest_field;
  };

  shm_alignment = sizeof(struct aligned_struct) - sizeof(union longest_field);

} /* shm_calc_alignment */




