/******************************************************************************
Description:

  UDP/IP protocol module

CVS Information:

  $Source: /nfs/globus1/src/master/nexus_4.1/pr_udp.c,v $
  $Date: 1997/11/04 16:01:13 $
  $Revision: 1.57 $
  $State: Exp $
  $Author: tuecke $
******************************************************************************/

/******************************************************************************
		  Define pre-include configuration parameters
******************************************************************************/
/*
#define USE_TRACEPOINT
*/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "internal.h"
  
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <fcntl.h>

#ifdef USE_TRACEPOINT
#   include "tracepoint.h"
#else
#   define tracepoint(str)
#endif /* USE_TRACEPOINT */


/******************************************************************************
		       Define module specific constants
******************************************************************************/

/*
 * NEXUS_UDP_DIRECT_CUSTOM_MAX_SIZE
 */
#ifndef NEXUS_UDP_DIRECT_CUSTOM_MIN_SIZE
#define NEXUS_UDP_DIRECT_CUSTOM_MIN_SIZE 2001
#endif


/*
 * Only one thread is allowed to be in the udp code (and thus
 * mucking with data structures) at a time.
 */
#ifdef BUILD_LITE
#define udp_enter()
#define udp_exit()
#else  /* BUILD_LITE */
static nexus_mutex_t	udp_mutex;
#define udp_enter()	nexus_mutex_lock(&udp_mutex);
#define udp_exit()	nexus_mutex_unlock(&udp_mutex);
#endif /* BUILD_LITE */

static nexus_bool_t	udp_done;
static nexus_bool_t     udp_aborting;

#define udp_fatal udp_exit(); nexus_fatal

/*
 * Other useful defines
 */

#define UDP_INCOMING_DEFAULT_SIZE 8192
#define UDP_HDR_SIZE             4

/*
 * TODO: Figure out if there is a more dynamic way to determine
 * what this should be.  For now keep it at 1024
 */
#define UDP_MAX_TRANSMISSION_UNIT	1024

#define FLAG_MASK	        0xf0
#define OPEN_FLAG		0x80
#define CLOSE_NORMAL_FLAG	0x40
#define CLOSE_ABNORMAL_FLAG	0x20
#define CLOSE_SHUTDOWN_FLAG	0x10

#define NEXUS_PROTO_NAME_UDP "udp"
#define UDP_PROTOCOL_INFO _nx_pr_udp_info
#define UDP_INTERFACE_NAME_KEY "udp_interface"

/*
 * MSG_FORMAT_AND_HEADER_SIZE
 *
 * This should be >1, so that the close messages get handled by
 * the error callback instead of the read callback.  And ideally
 * it would be big enough to include the message size field.  But
 * we do not know the size of the message size field until we know
 * the format, so the best we can do is guess.  Finally, it must be
 * smaller than the total message header size of any format.
 */
#define MSG_FORMAT_AND_HEADER_SIZE 9

#ifndef SO_REUSEPORT
#   define SO_REUSEPORT SO_REUSEADDR
#endif


/******************************************************************************
			       Type definitions
******************************************************************************/

/*
 * udp_outgoing_t
 *
 * This is an overload of nexus_proto_t.  It adds the
 * udp specific information to that structure.
 */
struct udp_outgoing_s
{
    nexus_proto_type_t		type;
    nexus_proto_funcs_t *	funcs;
    unsigned long		direct_custom_min_size;
    unsigned long		direct_custom_max_size;
    unsigned long		direct_pointer_min_size;
    unsigned long		direct_pointer_max_size;
    nexus_bool_t		can_use_iovec;
    unsigned long		reserved_header_size;
    struct udp_outgoing_s *	next;
    char *			host;
    unsigned short		port;
    struct sockaddr_in		addr;
    unsigned int                bytes_sent;
    unsigned short              seq_num;
    int				fd;
    int				state;
    int				fault_code;
    int				reference_count;
    char			read_buf[1];
};

enum
{
  OUTGOING_STATE_UNOPENED                = 0,
  OUTGOING_STATE_OPEN                    = 1, 
  OUTGOING_STATE_CLOSING                 = 2,
  OUTGOING_STATE_CLOSED                  = 3,
  OUTGOING_STATE_WRITE_ERROR             = 4,
  OUTGOING_STATE_FAULT                   = 5
};


/*
 * udp_incoming_t
 *
 * One of these is allocated for each incoming file descriptor.
 * It stores partial incoming messages until they are complete
 * and can be handled or queued.
 */
struct udp_incoming_s
{
    int				fd;
    unsigned short              port;
    int				state;
    int				format;
    unsigned long		msg_size;
    unsigned int                expected_seq_num;
    unsigned int                current_msg_num;
    char                        tmp_storage[UDP_HDR_SIZE];
    size_t			nbytes_read;
    unsigned long		nbytes_parsed;
    unsigned long		nbytes_unparsed;
    unsigned long		storage_size;
    nexus_byte_t *		storage;
    nexus_byte_t *		current;
    unsigned long		save_storage_size;
    nexus_byte_t *		save_storage;
    struct sockaddr             from_addr;
    int                         from_len;
    nexus_bool_t		dispatch_in_progress;
    struct _nexus_buffer_t *	dispatch_q_head;
    struct _nexus_buffer_t *	dispatch_q_tail;
    nexus_endpoint_t *		endpoint;
};

enum
{
    INCOMING_STATE_FORMAT,
    INCOMING_STATE_MSG_SIZE,
    INCOMING_STATE_HEADER,
    INCOMING_STATE_BODY
};

typedef struct udp_outgoing_s	udp_outgoing_t;
typedef struct udp_incoming_s	udp_incoming_t;

/*
 * Tables to map from fd to outgoings, incomings, and listeners.
 */
typedef enum
{
    FD_TABLE_TYPE_NONE,
    FD_TABLE_TYPE_OUTGOING,
    FD_TABLE_TYPE_INCOMING
} fd_table_type_t;

typedef struct _fd_table_t
{
    fd_table_type_t	type;
    void *		value;
} fd_table_t;


/******************************************************************************
			    Module specific macros
******************************************************************************/

/*
 * Some useful queue macros
 */
#define Enqueue(Qhead, Qtail, Item) \
{ \
    if (Qhead) \
    { \
	(Qtail)->next = (Item); \
	(Qtail) = (Item); \
    } \
    else \
    { \
	(Qhead) = (Qtail) = (Item); \
    } \
}

#define Dequeue(Qhead, Qtail, Item) \
{ \
    (Item) = (Qhead); \
    (Qhead) = (Qhead)->next; \
}

#define QueueNotEmpty(Qhead)	(Qhead)


/******************************************************************************
			  Module specific prototypes
******************************************************************************/

static nexus_proto_type_t
nexusl_pr_udp_proto_type(void);

static void
nexusl_pr_udp_init(int *argc,
		   char ***argv,
		   nexus_bool_t * add_to_my_mi_proto);

static void
nexusl_pr_udp_shutdown(nexus_bool_t shutdown_others);

static void
nexusl_pr_udp_abort(void);

static int
nexusl_pr_udp_send_rsr(struct _nexus_buffer_t *buffer);

static void
nexusl_pr_udp_increment_reference_count(nexus_proto_t *nproto);

static nexus_bool_t
nexusl_pr_udp_decrement_reference_count(nexus_proto_t *nproto);

static int
nexusl_pr_udp_get_my_mi_proto(nexus_byte_t **array,
			      int *size,
			      void *proto_info,
			      nexus_endpoint_t *endpoint);

static nexus_bool_t
nexusl_pr_udp_construct_from_mi_proto(nexus_proto_t **proto,
				      nexus_mi_proto_t *mi_proto,
				      nexus_byte_t *proto_array,
				      int size);

static int
nexusl_pr_udp_destroy_my_mi_proto(nexus_byte_t *proto_array,
				  int size);

static int
nexusl_pr_udp_test_proto(nexus_proto_t *proto);

static int
nexusl_pr_udp_direct_info_size(void);

static int
nexusl_pr_udp_outgoing_construct(char *host,
				 u_short port,
				 int ttl,
				 udp_outgoing_t ** outgoing_ret);

static void
nexusl_pr_udp_outgoing_free(udp_outgoing_t *outgoing);

static void
nexusl_pr_udp_outgoing_close(udp_outgoing_t *outgoing,
			     int new_state);

static int
nexusl_pr_udp_outgoing_address_setup(udp_outgoing_t * outgoing);


static void
nexusl_pr_udp_incoming_construct(int fd,
				 nexus_endpoint_t * endpoint);

static void
nexusl_pr_udp_incoming_close(udp_incoming_t *incoming);

static void
nexusl_pr_udp_incoming_free(udp_incoming_t *incoming);

static void
nexusl_pr_udp_process_fragment(void *arg, long n_read);

static void
nexusl_pr_udp_set_socket_size(int s);

/* callbacks */
static void
nexusl_pr_udp_read_callback(void *arg,
			    int fd);


/******************************************************************************
		       Define module specific variables
******************************************************************************/

static unsigned short udp_msg_num;

static int              fd_tablesize;
static fd_table_t       *fd_table;
static int              n_fds_open;

/*
 * My local udp connection information
 */
static char		udp_local_host[MAXHOSTNAMELEN];

/*
 * Udp specific stuff 
 */
static nexus_proto_funcs_t udp_proto_funcs =
{
    nexusl_pr_udp_proto_type,
    nexusl_pr_udp_init,
    nexusl_pr_udp_shutdown,
    nexusl_pr_udp_abort,
    NULL /* nexusl_pr_udp_poll */,
    NULL /* nexusl_pr_udp_poll_blocking */,
    nexusl_pr_udp_increment_reference_count,
    nexusl_pr_udp_decrement_reference_count,
    nexusl_pr_udp_get_my_mi_proto,
    nexusl_pr_udp_construct_from_mi_proto,
    nexusl_pr_udp_destroy_my_mi_proto,
    nexusl_pr_udp_test_proto,
    nexusl_pr_udp_send_rsr,
    nexusl_pr_udp_direct_info_size,
    NULL /* nexusl_pr_udp_direct_get */,
};


/*
 * _nx_pr_udp_info()
 *
 * Return the nexus_proto_funcs_t function table for this protocol module.
 *
 * This procedure is used for bootstrapping the protocol module.
 * The higher level Nexus code needs to call this routine to
 * retrieve the functions it needs to use this protocol module.
 */
void *
_nx_pr_udp_info(void)
{
    return((void *) (&udp_proto_funcs));
}
/* _nx_pr_udp_info() */


/*
 * nexusl_pr_udp_proto_type()
 *
 * Return the nexus_proto_type_t for this protocol module.
 */
static nexus_proto_type_t
nexusl_pr_udp_proto_type(void)
{
    return (NEXUS_PROTO_TYPE_UDP);
}
/* nexusl_pr_udp_proto_type() */


/*
 * nexusl_pr_udp_init()
 *
 * Initialize the UDP protocol.
 */
static void
nexusl_pr_udp_init(int *argc,
		   char ***argv,
		   nexus_bool_t * add_to_my_mi_proto)
{
    int i;
    int rc;
    
#ifndef BUILD_LITE
    nexus_mutex_init(&udp_mutex, (nexus_mutexattr_t *) NULL);
#endif /* BUILD_LITE */
    
    udp_done     = NEXUS_FALSE;
    udp_aborting = NEXUS_FALSE;

    udp_msg_num = 0;
    
    /* Initialize the fd_table_* tables */ 
    fd_tablesize = nexus_fd_tablesize();
    NexusMalloc(udp_init(),
		fd_table,
		fd_table_t *,
		(sizeof(fd_table_t) * fd_tablesize));
    for (i = 0; i < fd_tablesize; i++)
    {
	fd_table[i].type = FD_TABLE_TYPE_NONE;
	fd_table[i].value = NULL;
    }

    n_fds_open = 0;

    _nx_md_gethostname(udp_local_host, MAXHOSTNAMELEN);

    *add_to_my_mi_proto = NEXUS_FALSE;
}
/* nexusl_pr_udp_init() */


/*
 * nexusl_pr_udp_shutdown()
 *
 * This routine is called during normal shutdown of a process.
 *
 * Shutdown the udp protocol module by closing all off of the
 * outgoing's that have open fd's.  This will result in an orderly
 * shutdown of fd's, so that other nodes will not detect an
 * abnormal eof on the other ends of these fd's.
 *
 * It is ok if this procedure takes a little while to complete -- for
 * example to wait for another process to read off its fd and
 * free up some socket buffer space.
 *
 * An orderly closing of a connection is done as follows:
 *   1) Send the CLOSE_NORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd,
 *	if shutdown_others==NEXUS_FALSE.  Otherwise send the
 *	CLOSE_SHUTDOWN_FLAG.
 *   2) Then close the fd and its associated outgoing.
 */
static void
nexusl_pr_udp_shutdown(nexus_bool_t shutdown_others)
{
    int i;
    int fd;
    udp_outgoing_t *outgoing;
    udp_incoming_t *incoming;
    int rc;

    udp_enter();

    nexus_debug_printf(1, ("udp_shutdown(): inside\n"));

    udp_done = NEXUS_TRUE;

    for (i = 0; i < fd_tablesize; i++)
    {
	switch(fd_table[i].type)
	{
	case FD_TABLE_TYPE_OUTGOING:
	  nexus_debug_printf(1, ("udp_shutdown(): outgoing\n"));
	    outgoing = (udp_outgoing_t *) fd_table[i].value;
	    if (outgoing->state == OUTGOING_STATE_OPEN)
	    {
		nexusl_pr_udp_outgoing_close(outgoing, OUTGOING_STATE_CLOSED);
		nexusl_pr_udp_outgoing_free(outgoing);
	    }
	    break;
	case FD_TABLE_TYPE_INCOMING:
	  nexus_debug_printf(1, ("udp_shutdown(): incoming\n"));	
	    incoming = (udp_incoming_t *) fd_table[i].value;
	    nexusl_pr_udp_incoming_close( incoming );
	    nexusl_pr_udp_incoming_free( incoming );
	    break;
	default:
	    /* Nothing registered for this fd */
	    break;
	}
    }
    udp_exit();

}
/* nexusl_pr_udp_shutdown() */


/*
 * nexusl_pr_udp_abort()
 *
 * This routine is called during the _abnormal_ shutdown of a process.
 *
 * Shutdown the udp protocol module by closing all off of the
 * outgoing's that have open fd's.  This will hopefully result
 * in a reasonably  orderly shutdown of fd's, so that
 * other nodes do not spew error messages when they detect
 * closed sockets.
 *
 * However, this procedure should not wait to do things cleanly,
 * like udp_shutdown().  If it cannot close a fd cleanly (first
 * write a CLOSE_ABNORMAL_FLAG to it), then it should close the
 * fd anyway and get on with things.  The highest priority is
 * to just get shut down.
 *
 * The closing of a connection is done as follows:
 *   1) Send the CLOSE_ABNORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd.
 *	If the write would block, give up and go on to next step.
 *   2) Then close the fd.
 */
static void
nexusl_pr_udp_abort()
{
    udp_outgoing_t *outgoing;
    udp_incoming_t *incoming;
    int i;

    udp_done     = NEXUS_TRUE;
    udp_aborting = NEXUS_TRUE;

    for (i = 0; i < fd_tablesize; i++)
    {
	switch(fd_table[i].type)
	{
	case FD_TABLE_TYPE_OUTGOING:
	    outgoing = (udp_outgoing_t *) fd_table[i].value;
	    if( outgoing->state == OUTGOING_STATE_OPEN )
	    {
		nexusl_pr_udp_outgoing_close(outgoing, OUTGOING_STATE_CLOSED);
		nexusl_pr_udp_outgoing_free(outgoing);
	    }
	    break;
	case FD_TABLE_TYPE_INCOMING:
	    incoming = (udp_incoming_t *) fd_table[i].value;
	    nexusl_pr_udp_incoming_close(incoming);
	    nexusl_pr_udp_incoming_free(incoming);
	    break;
	default:
	    /* Nothing registered for this fd */
	    break;
	}
    }

}
/* nexusl_pr_udp_abort() */


/*
 * nexusl_pr_udp_send_rsr()
 *
 * Generate a remote service request message to the node and context
 * saved in the 'nexus_buffer'.
 */
static int
nexusl_pr_udp_send_rsr(struct _nexus_buffer_t *buffer)
{
    int            rc;
    udp_outgoing_t  *outgoing;
    int            fd;
    size_t         total_size;
    size_t         size;
    nexus_byte_t   *buf;
    nexus_bool_t   done;
    
    done    = NEXUS_FALSE;
    
    nexus_debug_printf(0, ("nexusl_pr_udp_send_rsr(): entering, "
			   "buffer=0x%08x\n", (unsigned) buffer));

    if (buffer->n_direct > 0)
    {
	nexus_mutex_init(&(buffer->barrier.mutex), (nexus_mutexattr_t *) NULL);
	nexus_cond_init(&(buffer->barrier.cond), (nexus_condattr_t *) NULL);
	buffer->barrier.count = buffer->n_direct;
	buffer->using_barrier = NEXUS_TRUE;
    }

    udp_enter();
    
    outgoing = (udp_outgoing_t *) buffer->proto;
    NexusAssert2((outgoing->type == NEXUS_PROTO_TYPE_UDP),
		 ("udp_send_rsr(): Internal error: proto_type is not NEXUS_PROTO_TYPE_UDP\n"));

    NexusAssert2((outgoing->fd >= 0), ("udp_send_rsr(): Internal error: Failed to open outgoing\n"));


    if (outgoing->fault_code != NEXUS_FAULT_NONE)
    {
	goto fn_exit;
    }

    /* By the time we get here, everything is already in the proto
     * including the host and port combination we have to send to.
     * this is connectionless now so use sendto.
     */

    /* don't worry about iovec or direct segments right now */
    
    total_size = buffer->base_segments->size_used;

    outgoing->bytes_sent = 0;
    outgoing->seq_num = 0;
    
    /*
     * for the first packet, move current back to make room for
     * the header
     *
     * udp_outgoing_t->reserved_header_size is set to UDP_HDR_SIZE
     * to guarantee us that there is room saved in front of current
     * so that this operation does not underflow the array.
     */
    buffer->base_segments->current -= UDP_HDR_SIZE;

    done = NEXUS_FALSE;
    while(!done)
    {
	buf = buffer->base_segments->current;
	
	/* figure out the size of the packet we're sending */
	if (total_size - outgoing->bytes_sent + UDP_HDR_SIZE
	    <= UDP_MAX_TRANSMISSION_UNIT)
	{
	    size = total_size - outgoing->bytes_sent + UDP_HDR_SIZE;
	}
	else
	{
	    size = UDP_MAX_TRANSMISSION_UNIT;
	}

	/* put on the header */
	
	buf[0] = (char) udp_msg_num;
	buf[1] = (char) ((outgoing->seq_num >> 16) & 0xff);
	buf[2] = (char) ((outgoing->seq_num >> 8) & 0xff);
	buf[3] = (char) (outgoing->seq_num & 0xff);

	/* send the packet */
	rc = sendto(outgoing->fd,
		    (char*) buf,
		    size,
		    0,
		    (struct sockaddr*)&outgoing->addr,
		    sizeof(outgoing->addr));

	if (rc >= 0)
	{
	    /*
	     *  don't count the header as part of
	     *  the total number of bytes sent
	     */
	    outgoing->bytes_sent += size - UDP_HDR_SIZE;

	    nexus_debug_printf(0, ("tot_size[%d] pkt-size[%d] "
				   "real-bytes-sent[%d] seq[%d]\n",
				   total_size, size, outgoing->bytes_sent,
				   outgoing->seq_num));
		
	    /* are we done */
	    if( outgoing->bytes_sent < total_size )
	    {
		/* move current forward */	  
		buffer->base_segments->current +=
		    (UDP_MAX_TRANSMISSION_UNIT - UDP_HDR_SIZE);
		outgoing->seq_num++;
	    }
	    else
	    {
		buffer->current_base_segment
		    = (nexus_base_segment_t *) NULL;
		done = NEXUS_TRUE;
	    }
	}
	else
	{
	    int				save_errno;

	    save_errno = errno;
	    
	    if (save_errno == ECONNREFUSED)
	    {
		nexus_debug_printf(1, ("nexusl_pr_udp_send_rsr(): "
				       "sendto() connection refused\n"));
		outgoing->fault_code = NEXUS_FAULT_PROCESS_DIED;
		nexusl_pr_udp_outgoing_close(outgoing, OUTGOING_STATE_FAULT);
		goto fn_exit;
	    }
	    else if (save_errno != ENOBUFS)
	    {
		nexus_debug_printf(1, ("nexusl_pr_udp_send_rsr(): "
				   "sendto() failed (errno=%d\n", save_errno));
		outgoing->fault_code = NEXUS_FAULT_BAD_PROTOCOL;
		nexusl_pr_udp_outgoing_close(outgoing, OUTGOING_STATE_FAULT);
		goto fn_exit;
	    }
	    else /* save_errno == ENOBUFS */
	    {
		/*
		 * We are out of buffers; just drop the rest of the message
		 * on the floor and return
		 */
		nexus_debug_printf(3, ("nexusl_pr_udp_send_rsr(): "
				   "sendto() ran out of buffer space, "
				   "dropping message...\n"));
		buffer->current_base_segment = (nexus_base_segment_t *) NULL;
                done = NEXUS_TRUE;
	    }
	}
    }
    
    udp_msg_num++;

  fn_exit:
    udp_exit();

    nexus_debug_printf(0, ("nexusl_pr_udp_send_rsr(): exiting\n"));

    return (outgoing->fault_code);
}
/* nexusl_pr_udp_send_rsr() */


/*
 * nexusl_pr_udp_outgoing_construct()
 *
 * Construct a udp_outgoing_t for the given host and port. Look up in the
 * outgoing table to see if one already exists. If it does, bump its reference
 * count and return that one. Otherwise create one, insert into the
 * table with a reference count of 1 and return it.
 */
static int
nexusl_pr_udp_outgoing_construct(char *host,
				 u_short port,
				 int ttl,
				 udp_outgoing_t ** outgoing_ret)
{
    int					rc;
    udp_outgoing_t *			outgoing;

    nexus_debug_printf(3,("nexusl_pr_udp_outgoing_construct(): enter, "
			  "host=%s, port=%d, ttl=%d\n",
			  host, port, ttl));

    rc = 0;

    outgoing = NULL;
    /*
     * Perform lookup of host and port here...
     */

    if(outgoing == NULL)
    {
	char				ip_mc_loop = 0;
	char				ip_mc_ttl;
	struct ip_mreq			imr;
	unsigned			ip_daddr;
	unsigned			tmp_u;
	int				new_fd;

	NexusMalloc(outgoing_construct(), outgoing, udp_outgoing_t *,
		    sizeof(udp_outgoing_t));

	outgoing->type = NEXUS_PROTO_TYPE_UDP;
	outgoing->funcs = &udp_proto_funcs;
	outgoing->direct_custom_min_size = NEXUS_UDP_DIRECT_CUSTOM_MIN_SIZE;
	outgoing->direct_custom_max_size = NEXUS_DC_MAX_U_LONG;
	outgoing->direct_pointer_min_size = NEXUS_DC_MAX_U_LONG;
	outgoing->direct_pointer_max_size = NEXUS_DC_MAX_U_LONG;
	outgoing->can_use_iovec = NEXUS_FALSE;
	outgoing->reserved_header_size = UDP_HDR_SIZE;
	outgoing->host = _nx_copy_string(host);
	outgoing->port = port;
	outgoing->fd = -1;
	outgoing->seq_num = 0;
	outgoing->bytes_sent = 0;
	outgoing->state = OUTGOING_STATE_UNOPENED;
	outgoing->fault_code = NEXUS_FAULT_NONE;
	outgoing->reference_count = 1;

	/* note: no need to bind, this will be a send-only
	 * socket.  The receivers are instantiated in the
	 * epattr
	 */

	if((outgoing->fd = socket( AF_INET, SOCK_DGRAM, 0 )) < 0 )
	{
	    rc = -1 /* NEXUS_FAULT_SPACE */;
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_outgoing_construct(): "
				"socket() failed: %s\n",
				_nx_md_system_error_string(errno)));
	    goto fn_exit;
	}

	/* check to see if we've got a multicast address, if so
	 * we have to join the group
	 */
	if((sscanf(host, "%u.%*u.%*u.%u", &ip_daddr, &tmp_u) == 2)
	   && (ip_daddr >= 224 && ip_daddr <= 239))
	{
	    ip_mc_ttl = (char) ((ttl > 0) ? ttl : IP_DEFAULT_MULTICAST_TTL);
	    if (setsockopt(outgoing->fd, IPPROTO_IP, IP_MULTICAST_TTL,
			   &ip_mc_ttl, sizeof(ip_mc_ttl)) < 0)
	    {
		rc = -1 /* NEXUS_FAULT_BADPARAM */;
		nexus_debug_printf(1,
				   ("nexusl_pr_udp_outgoing_construct(): "
				    "setsockopt(MULTICAST_TTL) failed: %s\n",
				    _nx_md_system_error_string(errno)));
		goto fn_exit;
	    }

	    if (setsockopt(outgoing->fd, IPPROTO_IP, IP_MULTICAST_LOOP,
			   &ip_mc_loop, sizeof(ip_mc_loop)) == -1)
	    {
		rc = -1 /* NEXUS_FAULT_BADPARAM */;
		nexus_debug_printf(1,
				   ("nexusl_pr_udp_outgoing_construct(): "
				    "setsockopt(MULTICAST_LOOP) failed: %s\n",
				    _nx_md_system_error_string(errno)));
		goto fn_exit;
	    }
 
	    if ((int) (imr.imr_multiaddr.s_addr = inet_addr(host)) == -1)
	    {
		rc = -1 /* NEXUS_FAULT_BADPARAM */;
		nexus_debug_printf(1,
				   ("nexusl_pr_udp_outgoing_construct(): "
				    "inet_addr() failed: %s\n"));
		goto fn_exit;
	    }
	    imr.imr_interface.s_addr = INADDR_ANY;

	    if( setsockopt(outgoing->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			   (char *) &imr, sizeof(struct ip_mreq) ) == -1 )
	    {
		rc = -1 /* NEXUS_FAULT_BADPARAM */;
		nexus_debug_printf(1,
				   ("nexusl_pr_udp_outgoing_construct(): "
				    "setsockopt(IP_ADD_MEMBER) failed: %s\n",
				    _nx_md_system_error_string(errno)));
		goto fn_exit;
	    }
	}

	/*
	 * setup a sockaddr_in struct ahead of time so when I'm
	 * ready to send all I have to do is send.
	 */
	if((rc = nexusl_pr_udp_outgoing_address_setup(outgoing)) != 0)
	{
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_outgoing_construct(): "
				"address_setup failed\n"));
	    goto fn_exit;
	}

	fd_table[outgoing->fd].type = FD_TABLE_TYPE_OUTGOING;
	fd_table[outgoing->fd].value = (void *) outgoing;
    }

    *outgoing_ret = outgoing;
    NexusAssert(rc == 0);

  fn_exit:
    if (rc != 0 && outgoing->fd >= 0)
    {
	close(outgoing->fd);
    }

    nexus_debug_printf(3,("nexusl_pr_udp_outgoing_construct(): exit\n"));

    return(rc);
}
/* nexusl_pr_udp_outgoing_construct() */


/*
 * nexusl_pr_udp_outgoing_free()
 *
 * Free the passed 'outgoing'.
 */
static void
nexusl_pr_udp_outgoing_free(udp_outgoing_t *outgoing)
{
    if (outgoing->host != (char *) NULL)
    {
	NexusFree(outgoing->host);
    }
    NexusFree(outgoing);
}
/* nexusl_pr_udp_outgoing_free() */


/*
 * nexusl_pr_udp_outgoing_close()
 *
 * Close a outgoing:
 *   1) Remove the outgoing from the fd_table table
 *   2) Close the fd
 *   3) Modify outgoing data structure
 */
static void
nexusl_pr_udp_outgoing_close(udp_outgoing_t *outgoing,
			     int new_state)
{
    nexus_debug_printf(1, ("outgoing_close(): inside\n"));

#ifdef BUILD_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(outgoing->fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(outgoing->fd, (struct sockaddr *) &remote, &len);

	nexus_printf("outgoing_close(): closing outgoing %x %s/%hu fd=%d local=%hu remote=%hu\n",
		     outgoing,
		     outgoing->host,
		     outgoing->port,
		     outgoing->fd,
		     htons(local.sin_port),
		     htons(remote.sin_port));
    }
#endif

    outgoing->state = OUTGOING_STATE_CLOSING;
    nexus_fd_unregister(outgoing->fd, NULL);
    nexus_fd_close(outgoing->fd);
    outgoing->state = new_state;
    outgoing->fd = -1;

    fd_table[outgoing->fd].type = FD_TABLE_TYPE_NONE;
    fd_table[outgoing->fd].value = NULL;
    n_fds_open--;
    
}
/* nexusl_pr_udp_outgoing_close() */


/*
 * nexusl_pr_udp_increment_reference_count()
 *
 * Increase the reference count on the associated proto and copy the
 * pointer to the nexus_proto_t
 *
 */
static void
nexusl_pr_udp_increment_reference_count(nexus_proto_t *nproto)
{
    udp_outgoing_t *outgoing = (udp_outgoing_t *) nproto;
    udp_enter();
    outgoing->reference_count++;
    udp_exit();
}
/* nexusl_pr_udp_increment_reference_count() */


/*
 * nexusl_pr_udp_decrement_reference_count()
 *
 * Decrement the reference count for this proto.  If it goes to 0
 * then close the fd used by this outgoing.
 *
 * Return NEXUS_TRUE if this functioin frees the proto.
 */
static nexus_bool_t
nexusl_pr_udp_decrement_reference_count(nexus_proto_t *nproto)
{
    udp_outgoing_t *outgoing = (udp_outgoing_t *) nproto;
    nexus_bool_t proto_freed = NEXUS_FALSE;
    udp_enter();
    outgoing->reference_count--;
    
    NexusAssert2((outgoing->reference_count >= 0),
		 ("nexusl_pr_udp_decrement_reference_count(): "
		  "Internal error: Reference count < 0\n"));
	       
    if (outgoing->reference_count == 0)
    {
	if (outgoing->state == OUTGOING_STATE_OPEN)
	{
	    nexusl_pr_udp_outgoing_close(outgoing, OUTGOING_STATE_CLOSED);
	}

	nexusl_pr_udp_outgoing_free(outgoing);
	proto_freed = NEXUS_TRUE;
    }
    udp_exit();
    return(proto_freed);
    
}
/* nexusl_pr_udp_decrement_reference_count() */


/*
 * nexusl_pr_udp_get_my_mi_proto()
 *
 * Return in 'array' and 'size' a byte array containing
 * enough information to enable another process to connect
 * to this one.  This byte array will be assembled with those
 * of the other protocol modules and placed in a nexus_mi_proto_t.
 *
 * Fill in 'array' with:
 *	port  (4 byte, big endian integer)
 *	hostname (null terminated string)
 *
 * If the hostname is the same as my node name, then just place an
 * empty string in the hostname field.
 */
static int
nexusl_pr_udp_get_my_mi_proto(nexus_byte_t **array,
			      int *size,
			      void *proto_info,
			      nexus_endpoint_t *endpoint)
{
    int					rc;
    int					i;
    int					tmp_int;
    char *				host;
    int					host_length;
    int					fd;
    int					port;
    int					len;
    struct sockaddr_in			my_addr;
    int					save_error;
    udp_incoming_t *			incoming;
    nexus_proto_info_udp_t *		udp_info;

    nexus_debug_printf(0, ("udp_get_my_mi_proto(): inside\n"));

    rc = 0;

    udp_info = (nexus_proto_info_udp_t *) proto_info;

    if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
	rc = -1 /* NEXUS_FAULT_SPACE */;
	nexus_debug_printf(1,
			   ("nexusl_pr_udp_get_my_mi_proto(): "
			    "socket() call failed: %s\n",
			    _nx_md_system_error_string(errno)));
		goto fn_exit;
    }

    if(udp_info)
    {
	struct ip_mreq			imr;
	unsigned			tmp_u;
	int				so_reuse = 1;
	char				ip_mc_loop = 0;

	host = udp_info->host;
	port = udp_info->port;

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT,
		       (char *) &so_reuse, sizeof(so_reuse)) == -1)
	{
	    rc = -1 /* NEXUS_FAULT_BADPARAM */;
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_get_my_mi_proto(): "
				"setsockopt(REUSEPORT) failed: %s\n",
				_nx_md_system_error_string(errno)));
	    goto fn_exit;
	}

	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP,
		       &ip_mc_loop, sizeof(ip_mc_loop)) == -1)
	{
	    rc = -1 /* NEXUS_FAULT_BADPARAM */;
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_get_my_mi_proto(): "
				"setsockopt(MULTICAST_LOOP) failed: %s\n",
				_nx_md_system_error_string(errno)));
	    goto fn_exit;
	}
 
	if ((int)(imr.imr_multiaddr.s_addr = inet_addr(host)) == -1)
	{
	    rc = -1 /* NEXUS_FAULT_BADPARAM */;
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_get_my_mi_proto(): "
				"inet_addr() failed: %s\n"));
	    goto fn_exit;
	}
	imr.imr_interface.s_addr = INADDR_ANY;

	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       (char *)&imr, sizeof(struct ip_mreq)) == -1)
	{
	    rc = -1 /* NEXUS_FAULT_BADPARAM */;
	    nexus_debug_printf(1,
			       ("nexusl_pr_udp_get_my_mi_proto(): "
				"setsockopt(IP_ADD_MEMBER) failed: %s\n",
				_nx_md_system_error_string(errno)));
	    goto fn_exit;
	}
    }
    else if (strcmp(udp_local_host, _nx_my_node.name) == 0)
    {
	host = "";
	port = 0;
    }
    else
    {
	host = udp_local_host;
	port = 0;
    }

    bzero((char*) &my_addr, sizeof(my_addr));
    my_addr.sin_family = AF_INET;
    my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    my_addr.sin_port = htons(port);

    if( bind(fd, (struct sockaddr*) &my_addr, sizeof(my_addr)) < 0)
    {
	rc = -1 /* NEXUS_FAULT_ADDR_INUSE */;
	nexus_debug_printf(1,
			   ("nexusl_pr_udp_get_my_mi_proto(): "
			    "bind() failed: %s\n",
			    _nx_md_system_error_string(errno)));
	goto fn_exit;
    }
    
    nexusl_pr_udp_set_socket_size( fd );

    /* figure out the local port number */
    len = sizeof(my_addr);
    if( (getsockname( fd, (struct sockaddr *)&my_addr, &len)) < 0)
    {
	rc = -1 /* NEXUS_FAULT_BADPARAM */;
	nexus_debug_printf(1,
			   ("nexusl_pr_udp_get_my_mi_proto(): "
			    "getsockname() failed: %s\n",
			    _nx_md_system_error_string(errno)));
	goto fn_exit;
    }
    port = (u_short) my_addr.sin_port;

    nexus_debug_printf(5, ("nexusl_pr_udp_get_my_mi_proto(): "
			   "calling incoming_construct(%d, %d)\n", fd, port));
    nexusl_pr_udp_incoming_construct(fd, endpoint);
    incoming = fd_table[fd].value;

    nexus_fd_register_for_select( fd,
				  nexusl_pr_udp_read_callback,
				  NULL,
				  NULL,
				  incoming);
    
    
    host_length = (strlen(host) + 1);
    *size = 12 + host_length;
    NexusMalloc(udp_get_my_mi_proto(),
		*array,
		nexus_byte_t *,
		*size);

    i = 0;
    PackInt4(*array, i, fd);
    tmp_int = (int) port;
    PackInt4(*array, i, tmp_int);
    if (udp_info != NULL)
    {
	PackInt4(*array, i, udp_info->ttl);
    }
    else
    {
	tmp_int = 0;
	PackInt4(*array, i, tmp_int);
    }
    memcpy(&((*array)[i]), host, host_length);

  fn_exit:
    if (rc != 0)
    {
	if (fd >= 0)
	{
	    close(fd);
	}

	*size = 0;
	*array = NULL;
    }

    nexus_debug_printf(1, ("udp_get_my_mi_proto(): host[%s]\n", host));

    return(rc);
} 
/* nexusl_pr_udp_get_my_mi_proto() */


/*
 * nexusl_pr_udp_construct_from_mi_proto()
 *
 * From the passed machine independent protocol list ('mi_proto'), plus
 * the udp specific entry from that list ('proto_array' and 'size'),
 * see if I can use the information to create a nexus_proto_t object
 * that can be used to connect to the node:
 *	- If I cannot use this protocol to attach to the node, then
 *		return NEXUS_FALSE.  (This option is useful if two nodes
 *		both speak a particular protocol, but they cannot
 *		talk to each other via that protocol.  For example,
 *		on two MPP's, the nodes within a single MPP can
 *		talk to each other via the native messaging protocol,
 *		but cannot talk to the nodes on the other MPP
 *		using that native protocol.)
 *	- If this udp protocol points to myself, and thus the local
 *		protocol module should be used, then set
 *		*proto=NULL, and return NEXUS_TRUE.
 *	- Otherwise, construct a udp protocol object for this mi_proto
 *		and put it in *proto.  Then return NEXUS_TRUE.
 *
 * The 'proto_array' should contain:
 *	port  (4 byte, big endian integer)
 *	hostname (null terminated string)
 *
 * If the hostname is an empty string, then use the node name
 * from the mi_proto.
 */
static nexus_bool_t
nexusl_pr_udp_construct_from_mi_proto(nexus_proto_t **proto,
				      nexus_mi_proto_t *mi_proto,
				      nexus_byte_t *proto_array,
				      int size)
{
    char *				host;
    u_short				port;
    int					ttl;
    int					i;
    int					tmp_int;
    nexus_bool_t			rc;

    rc = NEXUS_TRUE;

    NexusAssert2((size >= 13),
		 ("udp_construct_from_mi_proto(): "
		  "Invalid udp information in mi_proto\n"));
		 
    /* Get the port and hostname */
    i = 0;
    UnpackInt4(proto_array, i, tmp_int);
    UnpackInt4(proto_array, i, tmp_int);
    port = (u_short) tmp_int;
    UnpackInt4(proto_array, i, ttl);
    host = (char *) &(proto_array[i]);
    if (host[0] == '\0')
    {
	int node_number, context_differentiator;
	i = 0;
	UnpackMIProtoHeader(mi_proto->array, i,
			    node_number,
			    context_differentiator,
			    host);
	NexusAssert2((strlen(host) > 0),
		     ("udp_construct_from_mi_proto(): "
		      "Invalid node name field in mi_proto\n"));
    }

    /*
     * TODO: how do we know if this proto points to itself anymore,
     * the test used to look at the udp_local_port information
     * but that's not used anymore.
     *
     * (old comment below, leave as a reminder)
     * Test to see if this proto points to myself.
     * If it does, then return the _nx_local_proto.
     */

    udp_enter();
    if (nexusl_pr_udp_outgoing_construct(host, port, ttl,
					 (udp_outgoing_t **) proto) != 0)
    {
	rc = NEXUS_FALSE;
    }
    udp_exit();

    return (rc);
}
/* nexusl_pr_udp_construct_from_mi_proto() */


/*
 * nexusl_pr_udp_destroy_my_mi_proto()
 */
int
nexusl_pr_udp_destroy_my_mi_proto(nexus_byte_t *proto_array,
				  int size)
{
    int					fd;
    int					tmp_i;
    int					rc;
    udp_incoming_t *			incoming;
    

    udp_enter();

    rc = 0;

    tmp_i = 0;
    UnpackInt4(proto_array, tmp_i, fd);

    if (fd <0 || fd >= fd_tablesize)
    {
	if (_nx_fault_tolerant)
	{
	    nexus_debug_printf(1, ("nexusl_pr_udp_destroy_my_mi_proto(): "
				   "proto_array contains bad FD=%d\n", fd));
	    rc = -1;
	    goto fn_exit;
	}
	else
	{
	    nexus_fatal("nexusl_pr_udp_destroy_my_mi_proto(): "
			"proto_array contains bad FD=%d\n", fd);
	}
    }

    if (fd_table[fd].type == FD_TABLE_TYPE_INCOMING)
    {
	if (_nx_fault_tolerant)
	{
	    nexus_debug_printf(1, ("nexusl_pr_udp_destroy_my_mi_proto(): "
				   "fd_table[%d].type != INCOMING\n", fd));
	    rc = -1;
	    goto fn_exit;
	}
	else
	{
	    nexus_fatal("nexusl_pr_udp_destroy_my_mi_proto(): "
			"fd_table[%d].type != INCOMING\n", fd);
	}
    }

    incoming = (udp_incoming_t *) fd_table[fd].value;

    nexusl_pr_udp_incoming_close(incoming);
    nexusl_pr_udp_incoming_free(incoming);

  fn_exit:
    udp_exit();
    return(rc);
}
/* nexusl_pr_udp_destroy_my_mi_proto() */


/*
 * nexusl_pr_udp_test_proto()
 */
static int
nexusl_pr_udp_test_proto(nexus_proto_t *proto)
{
    udp_outgoing_t *outgoing = (udp_outgoing_t *) proto;
    return(outgoing->fault_code);
}
/* nexusl_pr_udp_test_proto() */


/*
 * nexusl_pr_udp_direct_info_size()
 */
static int
nexusl_pr_udp_direct_info_size(void)
{
    /* TODO: This needs to be filled in */
    return(0);
}
/* nexusl_pr_udp_direct_info_size() */


/*
 * nexusl_pr_udp_incoming_construct()
 *
 * Construct a udp_incoming_t for the given file descriptor, 'fd'.
 */
static void
nexusl_pr_udp_incoming_construct(int fd,
				 nexus_endpoint_t * endpoint)
{
    udp_incoming_t *incoming;

    nexus_debug_printf(1, ("nexusl_pr_udp_incoming_construct(), enter, "
			   "fd=%d\n", fd));

    NexusMalloc(incoming_construct(),
		incoming,
		udp_incoming_t *,
		sizeof(udp_incoming_t));
    NexusMalloc(incoming_construct(),
		incoming->storage,
		nexus_byte_t *,
		UDP_INCOMING_DEFAULT_SIZE);

    incoming->fd			= fd;
    incoming->expected_seq_num		= 0;
    incoming->current_msg_num		= 0;    
    incoming->nbytes_read		= 0;
    incoming->nbytes_parsed		= 0;
    incoming->nbytes_unparsed		= 0;
    incoming->storage_size		= UDP_INCOMING_DEFAULT_SIZE;
    incoming->current			= incoming->storage;
    incoming->save_storage_size		= 0;
    incoming->dispatch_in_progress	= NEXUS_FALSE;
    incoming->dispatch_q_head		= (struct _nexus_buffer_t *) NULL;
    incoming->dispatch_q_tail		= (struct _nexus_buffer_t *) NULL;
    incoming->endpoint			= endpoint;

    fd_table[fd].type = FD_TABLE_TYPE_INCOMING;
    fd_table[fd].value = (void *) incoming;
    
    nexus_debug_printf(1, ("nexusl_pr_udp_incoming_construct(), exit\n"));

}
/* nexusl_pr_udp_incoming_construct() */


/*
 * nexusl_pr_udp_incoming_close()
 *
 * Close an incoming connection:
 *   1) Remove the udp_incoming_t from the fd_table table
 *   2) Close the fd
 *   3) Put the udp_incoming_t back on the free list.
 */
static void
nexusl_pr_udp_incoming_close(udp_incoming_t *incoming)
{
#ifdef BUILD_DEBUG
    if (NexusDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(incoming->fd, (struct sockaddr *) &local, &len);

	nexus_printf("incoming_close(): closing %x fd=%d local=%hu\n",
		     incoming,
		     incoming->fd,
		     htons(local.sin_port));
    }
#endif

    nexus_fd_unregister(incoming->fd, NULL);
    nexus_fd_close(incoming->fd);

    fd_table[incoming->fd].type = FD_TABLE_TYPE_NONE;
    fd_table[incoming->fd].value = NULL;
    n_fds_open--;
    
}
/* nexusl_pr_udp_incoming_close() */


/*
 * nexusl_pr_udp_incoming_free()
 */
static void
nexusl_pr_udp_incoming_free(udp_incoming_t *incoming)
{
    NexusFree(incoming->storage);
    NexusFree(incoming);
}
/* nexusl_pr_udp_incoming_free() */


/*
 * nexusl_pr_udp_set_socket_size()
 * TODO: figure out if this should be part of the proto->info stuff
 *       that Carl, Steve and I were talking about (buffer sizes)
 */
#define MAX_SOCKET_SEND_BUFFER		9000
#define MAX_SOCKET_RECEIVE_BUFFER	18032

static void
nexusl_pr_udp_set_socket_size(int fd)
{
    int size;
    int sock_size;
    int sock_opt_len = sizeof(int);
    int save_error;
    int flags;
    int rc = 0;
    

    size = MAX_SOCKET_SEND_BUFFER;
    if( setsockopt( fd, SOL_SOCKET, SO_SNDBUF, (char *) &size, sizeof(int) ) < 0 )
    {
	nexus_printf("udp_set_socket_size: setsockopt failed\n");
	save_error = errno;
	udp_fatal("udp_set_socket_size: setsockopt failed. errno = %d\n",
		 save_error);
	sock_size = -1;
    }

    if ( getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &sock_size, 
		   &sock_opt_len) < 0 ) 
    {
	save_error = errno;
        udp_fatal("udp_set_socket_size: getsockopt failed. errno = %d\n",
		  save_error);
        sock_size = -1;
    }
    
    if ( sock_size != size )
    {
        nexus_printf("udp_set_socket_size: set to %d instead of %d\n",
		     sock_size, size);
    }

    size = MAX_SOCKET_RECEIVE_BUFFER;

    /* bug fix, 7/5/95 cdematt, there was no & in front of size... */
    
    setsockopt( fd, SOL_SOCKET, SO_RCVBUF, (char *) &size, sizeof(int) );
    if ( getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &sock_size, 
		   &sock_opt_len) < 0 ) 
    {
	save_error = errno;
        udp_fatal("udp_set_socket_size: getsockopt failed. errno = %d\n",
		  save_error);
        sock_size = -1;
    }
    if ( sock_size != size )
    {
        nexus_printf("udp_set_socket_size: set to %d instead of %d\n",
		     sock_size, size);
    }

    /* set nonblocking */
    nexus_fd_setup_nonblocking(fd);

}
/* nexusl_pr_udp_set_socket_size */


/*
 * nexusl_pr_udp_outgoing_address_setup()
 */
static int
nexusl_pr_udp_outgoing_address_setup(udp_outgoing_t * outgoing)
{
    struct hostent			he;
    struct hostent *			hp;

#   ifdef HAVE_GETHOSTBYNAME_R_5
    {
	char				hp_tsdbuffer[500];
	int				hp_errnop;

	hp = gethostbyname_r(outgoing->host, &he,
			     hp_tsdbuffer, 500, &hp_errnop);
    }
#   elif defined(HAVE_GET_HOSTBYNAME_R_3)
    {
	struct hostent_data hp_data;

	hp = &he;
	if (gethostbyname_r(outgoing->host, hp, &hp_data) != 0)
	{
	    hp = NULL;
	}
    }
#   else
    {
	/*
	 * On SunOS 4.1.X, gethostbyname() calls sprintf() :(
	 */
	nexus_stdio_lock();
	hp = gethostbyname(outgoing->host);
	nexus_stdio_unlock();
    }
#   endif

    memset(&outgoing->addr, '\0', sizeof(outgoing->addr));
    outgoing->addr.sin_port = htons(outgoing->port);

    if (hp == NULL)
    {
	/*
	 * gethostbyname() on many machines does the right thing for IP
	 * addresses (e.g., "140.221.7.13").  But on some machines
	 * (e.g., SunOS 4.1.x) it doesn't.  So hack it in this case.
	 */
	if (isdigit(outgoing->host[0]))
	{
	    outgoing->addr.sin_family = AF_INET;
	    outgoing->addr.sin_addr.s_addr = inet_addr(outgoing->host);
	    if ((int) outgoing->addr.sin_addr.s_addr == -1)
	    {
		return(-1);
	    }
	}
	else
	{
	    return(-1);
	}
    }
    else
    {
	memcpy(&outgoing->addr.sin_addr, hp->h_addr, hp->h_length);
	outgoing->addr.sin_family = hp->h_addrtype;
    }

    return(0);
}
/* nexusl_pr_udp_outgoing_address_setup() */


/*
 * nexusl_pr_udp_read_callback()
 */
static void
nexusl_pr_udp_read_callback(void *arg,
			    int fd)
{
    udp_incoming_t *incoming = (udp_incoming_t *) arg;
    long n_read;
    int save_errno;
    nexus_bool_t done;

    nexus_debug_printf(1,("udp_read_callback(): entering\n"));

    udp_enter();

    done = NEXUS_FALSE;
    while (!done)
    {
	nexus_debug_printf(1,
			   ("udp_read_callback(): "
			    "calling recvfrom, fd=%i, current=%lx, size=%i\n",
			    fd,
			    (incoming->current),
			    (incoming->storage_size - incoming->nbytes_read)));

	n_read = recvfrom(fd,
			  (char*)incoming->current,
			  (incoming->storage_size - incoming->nbytes_read),
			  0,
			  &incoming->from_addr,
			  &incoming->from_len);
	save_errno = errno;
	nexus_debug_printf(1,("udp_read_callback(): "
			      "read returned n_read=%li\n", n_read));

#ifdef BUILD_DEBUG
	if (NexusDebug(7))
	{
	    int i;
	    unsigned char *start;
	    
	    start = (unsigned char *)(incoming->current);

	    nexus_printf("Received %d bytes of data on fd %d\n",
			 n_read,
			 fd);

	    nexus_printf("First %d bytes:  ", NEXUS_MIN(n_read, 100));
	    nexus_stdio_lock();
	    for (i = 0; i < NEXUS_MIN(n_read, 100); i++)
	    {
		printf("%2x  ", *(start + i));
	    }
	    printf("\n"); fflush(stdout);
	    nexus_stdio_unlock();
	}
#endif

	/*
	 * n_read: is > 0 if it successfully read some bytes
	 *         is < 0 on error -- need to check errno
	 */
	
	if (n_read > UDP_HDR_SIZE)
	{
	    nexusl_pr_udp_process_fragment(incoming, n_read);
	}
	else if (n_read < 0)
	{
	    if (save_errno == EINTR)
	    {
		/* Try again */
	    }
	    else if (save_errno == EAGAIN || save_errno == EWOULDBLOCK)
	    {
		done = NEXUS_TRUE;
	    }
	    else
	    {
		nexusl_pr_udp_incoming_close(incoming);
		udp_exit();
		if (_nx_fault_detected(NEXUS_FAULT_READ_FAILED) != 0)
		{
		    nexus_fatal("udp_read_callback(): "
				"Internal error: Read failed with errno=%i\n",
				save_errno);
		}
		else
		{
		    nexus_debug_printf(1,
				       ("nexusl_pr_udp_read_callback(): read "
					"failed, errno=%d\n", save_errno));
		}
		udp_enter();
		nexusl_pr_udp_incoming_free(incoming);
	    }
	}
	else /* n_read < UDP_HDR_SIZE */
	{
	    nexusl_pr_udp_incoming_close(incoming);
	    udp_exit();
	    if (_nx_fault_detected(NEXUS_FAULT_READ_FAILED) != 0)
	    {
		nexus_fatal("udp_read_callback(): "
			    "Internal error: Read failed with n_read=%d\n",
			    n_read);
	    }
	    else
	    {
		nexus_debug_printf(1, ("nexusl_pr_udp_read_callback(): read "
				       "read failed, n_read=%d\n", n_read));
	    }
	    udp_enter();
	    nexusl_pr_udp_incoming_free(incoming);
	}
    }

    nexus_fd_register_for_select(incoming->fd,
				 nexusl_pr_udp_read_callback,
				 NULL,
				 NULL,
				 incoming);
    
    udp_exit();
	
    nexus_debug_printf(1,("udp_read_callback(): exiting\n"));
}
/* nexusl_pr_udp_read_callback */


/*
 * nexusl_pr_udp_process_fragment()
 *
 * Assume udp_enter() has already been called before entering this function.
 */
static void
nexusl_pr_udp_process_fragment(void *arg, long n_read)
{
    udp_incoming_t *incoming = (udp_incoming_t *) arg;
    nexus_bool_t done;
    int sizeof_u_long;
    unsigned long tmp_u_long;
    struct _nexus_buffer_t *buffer;
    nexus_bool_t message_enqueued = NEXUS_FALSE;
    unsigned int seq_num = 0;
    unsigned short msg_num;
    
    if (udp_done)
    {
	/* just return here, udp_close and udp_abort take care of
	 * freeing the incomings and closing the fds
	 */
	return;
    }
    
    incoming->nbytes_read += n_read;
    incoming->nbytes_unparsed += n_read;

    msg_num = (unsigned short) *(incoming->current);
    incoming->current[0] = (char)0;
    memcpy( &seq_num, incoming->current, UDP_HDR_SIZE);

    nexus_debug_printf(1,("process_fragment(): msg_num[%d] seq_num[%d]\n",
			  msg_num, seq_num));

    /* this has to be here, before current gets moved to make sure
     * the contents of the buffer get restored correctly
     */
    if( seq_num == 0 )
    {
	incoming->current_msg_num = msg_num;
    }
    else
    {
	memcpy( incoming->current, incoming->tmp_storage, UDP_HDR_SIZE);
    }

    incoming->current += UDP_HDR_SIZE;
    incoming->nbytes_parsed += UDP_HDR_SIZE;
    incoming->nbytes_unparsed -= UDP_HDR_SIZE;

    if( seq_num != incoming->expected_seq_num ||
	msg_num != incoming->current_msg_num )
    {
	nexus_debug_printf(1,("nexusl_pr_udp_process_fragment(): "
			      "out-of-sequence packet/msg received"));
	nexus_debug_printf(1,("expected_seq_num[%d] seq_num[%d]\n",
			      incoming->expected_seq_num, seq_num));
	nexus_debug_printf(1,("msg_num[%d] current_msg_num[%d]\n",
			      msg_num, incoming->current_msg_num));

	/* drop everything else until we get the next message */
	incoming->current = incoming->storage;
	incoming->nbytes_read = 0;
	incoming->nbytes_parsed = 0;
	incoming->nbytes_unparsed = 0;
	incoming->expected_seq_num = 0;
	return;
    }
	    
    incoming->expected_seq_num++;

    if( seq_num == 0 )
    {
	/* get the format */
	incoming->format = (int) *(incoming->current);
	incoming->current++;
	incoming->nbytes_parsed++;
	incoming->nbytes_unparsed--;

	/* Get the message size */
	sizeof_u_long = nexus_dc_sizeof_remote_u_long(1, incoming->format);
	/* Get the message size */
	nexus_dc_get_u_long(&(incoming->current),
			    &(incoming->msg_size),
			    1,
			    incoming->format);

	/* don't increment current, it was moved in nexus_dc_get_u_long */
	incoming->nbytes_parsed += sizeof_u_long;
	incoming->nbytes_unparsed -= sizeof_u_long;

	if (incoming->msg_size > incoming->storage_size)
	{
	    /*
	     * The message is too big to fit in this storage.
	     * So allocate new storage that is big enough,
	     * copy the message into the new storage.
	     */
	    nexus_debug_printf(4, ("process_fragment(): msg too big\n"));
	    incoming->save_storage_size = incoming->storage_size;
	    incoming->save_storage = incoming->storage;

	    /* Allocate new storage */
	    NexusMalloc(process_fragment(),
			incoming->storage,
			nexus_byte_t *,
			(incoming->msg_size + UDP_HDR_SIZE));
	    incoming->storage_size = (incoming->msg_size + UDP_HDR_SIZE);

	    /* Copy message to new storage */
	    tmp_u_long = (incoming->nbytes_parsed
			  + incoming->nbytes_unparsed);
	    memcpy(incoming->storage,
		   incoming->save_storage,
		   tmp_u_long);
	    incoming->current = incoming->storage + incoming->nbytes_parsed;
	}
    }

    incoming->current += incoming->nbytes_unparsed;
    incoming->nbytes_parsed += incoming->nbytes_unparsed;
    incoming->nbytes_unparsed = 0;

    nexus_debug_printf(1, ("process_fragment(): seq[%d] hv[%d] nd[%d]\n",
		 seq_num,
		 incoming->nbytes_parsed + UDP_HDR_SIZE,
		 incoming->msg_size));
    if ((incoming->nbytes_parsed + UDP_HDR_SIZE) < incoming->msg_size)
    {
	/*
	 * We do not yet have all the fragments for this message.
	 * So save off the last UDP_HDR_SIZE bytes of the buffer,
	 * and setup incoming to expect another fragment and return.
	 */

	incoming->current -= UDP_HDR_SIZE;
	incoming->nbytes_parsed -= UDP_HDR_SIZE;
	
	memcpy(incoming->tmp_storage, incoming->current, UDP_HDR_SIZE);
    }
    else
    {
	/*
	 * We have a complete message.
	 * So dispatch it, and reset incoming for next message.
	 */

	nexus_debug_printf(5, ("process_fragment():got all the fragments\n"));
	
	/*
	 * Enqueue the message for dispatch, there is a header
	 * at the beginning of incoming->storage, so be careful
	 * to pass the correct address when creating the buffer
	 */
	_nx_buffer_create_from_raw(incoming->storage,
				   incoming->storage_size,
				   UDP_HDR_SIZE,
				   incoming->msg_size,
				   incoming->endpoint,
				   &buffer);
	Enqueue(incoming->dispatch_q_head,
		incoming->dispatch_q_tail,
		buffer);
	message_enqueued = NEXUS_TRUE;
		
	/* Reset the incoming state to INCOMING_STATE_HEADER */
	if (incoming->save_storage_size > 0)
	{
	    /* There is a buffer in the save_storage to use */
	    incoming->storage_size = incoming->save_storage_size;
	    incoming->storage = incoming->save_storage;
	    incoming->save_storage_size = 0;
	}
	else
	{
	    /* Must allocate new storage */
	    incoming->storage_size = UDP_INCOMING_DEFAULT_SIZE;
	    NexusMalloc(process_fragment(),
			incoming->storage,
			nexus_byte_t *,
			UDP_INCOMING_DEFAULT_SIZE);
	}

	incoming->current = incoming->storage;
	incoming->nbytes_read = 0;
	incoming->nbytes_parsed = 0;
	incoming->nbytes_unparsed = 0;
	incoming->expected_seq_num = 0;
    }
	    
    if (message_enqueued && !incoming->dispatch_in_progress)
    {
	incoming->dispatch_in_progress = NEXUS_TRUE;
	while (QueueNotEmpty(incoming->dispatch_q_head))
	{
	    Dequeue(incoming->dispatch_q_head,
		    incoming->dispatch_q_tail,
		    buffer);
	    udp_exit();
	    _nx_buffer_dispatch(buffer);
	    udp_enter();
	}
	incoming->dispatch_in_progress = NEXUS_FALSE;
    }
}
/* nexusl_pr_udp_process_fragment() */
