/*
 * Nexus
 * Authors:     Stuart Martin
 *              Argonne National Laboratory
 *
 * ping_pong.c  - Event Driver test routine 
 *
 *                This program will create a client and server to pass a message
 *                back and forth for a number of iterations.  Additionally, if a make
 *                dc is used to compile then this program will reformat the message into
 *                a byte string using the nexus data_conversion function calls. 
 */

#include <stdio.h>
#include <malloc.h>
#include <nexus_ed.h>
#include <nexus_dc.h>
#include <sys/param.h>
#include <sys/time.h>

/*
 * ed_args_t
 */
typedef struct _ed_args_t
{
    int                         fd;
    nexus_bool_t                done;
    nexus_bool_t                have_connection;
    nexus_bool_t                running_as_server;
    int                         msg_size;
    int                         n_iterations;
} ed_args_t;

static void             dc_puts(char *msg_buf);

static void             listener_callback(void *arg,
					  int fd);
static void             ed_read_callback(void *arg,
                                         int fd,
                                         char *buf,
                                         size_t nbytes,
                                         char **new_buf,
                                         size_t *new_max_nbytes,
                                         size_t *new_wait_for_nbytes);
static void             ed_read_error_callback(void *arg,
                                               int fd,
                                               char *buf,
                                               size_t nbytes,
                                               int error);
static void 		ed_write_callback(void *arg,
	                                  int fd,
                                          char *buf,
                                          size_t nbytes);
static void             ed_write_error_callback(void *arg,
                                                int fd,
                                                char *buf,
                                                size_t nbytes,
                                                int error);
/*
 * MAIN
 */
int main (int argc, char **argv)
{
    char             srv_host[MAXHOSTNAMELEN];
    unsigned short   srv_port;
#ifdef USE_DATA_CONVERSION
    nexus_byte_t     *msg_buf, *buffer;
    nexus_byte_t     type;
    int              size,count;
    char             *put_char_str= "Ping Pong test";
#else  
    char             *msg_buf;
#endif 
    int              save_n_iterations;
    int              rc;
    ed_args_t        args;
    double           avg_secs = 0.0;
    double           tot_secs = 0.0;
    struct timeval   i1, i2;

    if (argc != 4 && argc != 6)
    {
        /* invalid parameters passed */
        printf("Usage: %s <msg_size> <iterations> <cli or srv> <port> <host>\n", argv[0]);
        exit(1);
    }

#ifdef USE_DATA_CONVERSION
    type= (nexus_byte_t)(NEXUS_DC_FORMAT_LOCAL);
    size= nexus_dc_sizeof_byte(1);
    size+= nexus_dc_sizeof_int(1);
    size+= nexus_dc_sizeof_char(strlen(put_char_str));
    buffer= (nexus_byte_t *)malloc(size);
    args.msg_size = size;
    msg_buf= buffer;
    nexus_dc_put_byte(&buffer, &type, 1);
    count= strlen(put_char_str);
    nexus_dc_put_int(&buffer, &count, 1);
    nexus_dc_put_char(&buffer, put_char_str, strlen(put_char_str));
#else  
    args.msg_size = atoi(argv[1]);
    msg_buf = (char *)malloc(args.msg_size);
#endif 

    save_n_iterations = args.n_iterations = atoi(argv[2]);
    args.running_as_server = NEXUS_FALSE;
    if (strcmp(argv[3], "srv") == 0 ) 
    {
        if (argc != 4)
        {
            printf("Usage: %s <msg_size> <iterations> <cli or srv>\n", argv[0]);
            printf("      note: above args required when running as srv.\n");
            exit(1);
        }
        args.running_as_server = NEXUS_TRUE;
    }
    else if (strcmp(argv[3], "cli") == 0 ) 
    {
        if (argc != 6)
        {
            printf("Usage: %s <msg_size> <iterations> <cli or srv> <port> <host>\n", argv[0]);
            printf("      note: above args required when running as cli.\n");
            exit(1);
        }
        srv_port = atoi(argv[4]);
        strcpy(srv_host, argv[5]);
    }
    else 
    {
        printf("Usage: %s <msg_size> <iterations> <cli or srv>\n", argv[0]);
        printf("      note: the 3rd arg must be \"cli\" or \"srv\"\n");
        exit(1);
    }

    printf("----------------------------------------------------------------\n");
    printf("\tmessage size = %d \n", args.msg_size);
    printf("\titerations   = %d \n", args.n_iterations);
    printf("\trunning as %s\n", argv[3]);

    rc = nexus_fd_init(&argc, &argv);
    if (rc != 0)
    {
        fprintf(stderr, " nexus_fd_init failed\n");
	exit(1);
    } 

    args.have_connection = NEXUS_FALSE;
    args.done = NEXUS_FALSE;

    if (args.running_as_server)
    {
        srv_port = 0;

        if ( gethostname(srv_host, MAXHOSTNAMELEN) != 0 )
        {
            fprintf(stderr, "gethostname call failed\n");
            exit(1);
        }

        rc = nexus_fd_create_listener(&srv_port, -1, listener_callback, (void *) &args);
        if (rc != 0)
        {
            fprintf(stderr, " nexus_fd_create_listener failed\n");
            exit(1);
        }

        fprintf(stdout, "\tserver port is - %hi\n", srv_port);
        fprintf(stdout, "\t       host is - %s\n", srv_host);

        while(!args.have_connection)
            nexus_fd_handle_events(NEXUS_FD_POLL_BLOCKING_ALL, NULL);

        rc = nexus_fd_register_for_read(args.fd,
#ifdef USE_DATA_CONVERSION
                                        (char *)msg_buf,
#else  
                                        msg_buf,
#endif 
                                        args.msg_size,
                                        args.msg_size,
                                        ed_read_callback,
                                        ed_read_error_callback,
                                        (void *) &args);
        if (rc != 0)
        {
            fprintf(stderr, " nexus_fd_register_for_read failed\n");
            exit(1);
        }
    }
    else
    {
        rc = nexus_fd_connect(srv_host, srv_port, &(args.fd) );
        if (rc != 0)
        {
            fprintf(stderr, " nexus_fd_connect failed\n");
            exit(1);
        }

        rc = nexus_fd_register_for_write(args.fd,
#ifdef USE_DATA_CONVERSION
                                         (char *)msg_buf,
#else  
                                         msg_buf,
#endif 
                                         args.msg_size,
                                         ed_write_callback,
                                         ed_write_error_callback,
                                         (void *) &args);
        if (rc != 0)
        {
            fprintf(stderr, " nexus_fd_register_for_write failed\n");
        }
    }

    gettimeofday( &i1, NULL );

    while(!args.done)
        nexus_fd_handle_events(NEXUS_FD_POLL_BLOCKING_ALL, NULL);

    gettimeofday( &i2, NULL );

    tot_secs = ((i2.tv_sec - i1.tv_sec) * 1000000 + i2.tv_usec - i1.tv_usec);
    if (i2.tv_sec > i1.tv_sec)
        avg_secs = tot_secs / 1000000 / save_n_iterations;
    else
        avg_secs = (i2.tv_usec - i1.tv_usec) / save_n_iterations; 

    nexus_fd_close(args.fd);

    if (args.running_as_server)
    {
        rc = nexus_fd_close_listener(srv_port);
        if (rc != 0)
        {
            fprintf(stderr, " nexus_fd_close_listener failed\n");
            exit(1);
        }
    }

    printf("\t-----------------------------------------\n");
    printf("\taverage seconds per iteration = %7.7f\n", avg_secs);

    exit(0);
} /* main */ 


/*
 * listener_callback()
 */
static void listener_callback(void *arg, int fd)
{
    ed_args_t *ed_args = (ed_args_t *) arg;
    ed_args->have_connection = NEXUS_TRUE;
    ed_args->fd = fd;
} /* listener_callback() */


/*
 * ed_read_callback()
 */
static void ed_read_callback(void *arg,
                             int fd,
                             char *buf,
                             size_t nbytes_read,
                             char **new_buf,
                             size_t *new_max_nbytes,
                             size_t *new_wait_for_nbytes)
{
    int  rc, count;
    ed_args_t *ed_args = (ed_args_t *) arg;

#ifdef USE_DATA_CONVERSION
    nexus_byte_t bformat;
    int format;
    char test_char_str[15];
    nexus_byte_t *tmp_buf = (nexus_byte_t *) buf;
#endif 

    if (!ed_args->running_as_server)
    {
        /* a full iteration for the client (who starts off with the write)  
         * is after it has completed it's read 
         */
        ed_args->n_iterations--;
        if (ed_args->n_iterations == 0)
        {
            ed_args->done = NEXUS_TRUE;
        }
    }

    if (!ed_args->done)
    {
#ifdef USE_DATA_CONVERSION
        nexus_dc_get_byte(&tmp_buf, &bformat, 1, NEXUS_DC_FORMAT_LOCAL);
        format= (int)bformat;
        printf("in dc_gets: format= %i\n", format);
        nexus_dc_get_int(&tmp_buf, &count, 1, format);
        printf("in dc_gets: count= %i\n",count);
        nexus_dc_get_char(&tmp_buf, test_char_str, count, format);
        *(test_char_str+count)= '\0';
        printf("int dc_gets: string= %s\n",test_char_str);
#endif

        rc = nexus_fd_register_for_write(fd,
                                         buf,
                                         ed_args->msg_size,
                                         ed_write_callback,
                                         ed_write_error_callback,
                                         (void *) ed_args);
        if (rc != 0)
            fprintf(stderr, " nexus_fd_register_for_write failed\n");
    }

} /* ed_read_callback() */


/*
 * ed_read_error_callback()
 */
static void ed_read_error_callback(void *arg,
                                   int fd,
                                   char *buf,
                                   size_t nbytes,
                                   int error)
{
    if (error != 0)
    {
        fprintf(stderr, " ed_read_error failed\n");
    }
} /* ed_read_error_callback() */

/*
 * ed_write_callback()
 */
static void ed_write_callback(void *arg,
                             int fd,
                             char *buf,
                             size_t nbytes)
{
    int rc;

    ed_args_t *ed_args = (ed_args_t *) arg;

    if (ed_args->running_as_server)
    {
        /* a full iteration for the server (who is waiting to do a read) 
         * is after it has completed it's write
         */
        ed_args->n_iterations--;
        if (ed_args->n_iterations == 0)
            ed_args->done = NEXUS_TRUE;
    }

    if (!ed_args->done)
    {
        rc = nexus_fd_register_for_read(fd,
                                        buf,
                                        ed_args->msg_size,
                                        ed_args->msg_size,
                                        ed_read_callback,
                                        ed_read_error_callback,
                                        (void *) ed_args);
        if (rc != 0)
            fprintf(stderr, " nexus_fd_register_for_read failed\n");
    }
} /* ed_write_callback() */

/*
 * ed_write_error_callback()
 */
static void ed_write_error_callback(void *arg,
                                    int fd,
                                    char *buf,
                                    size_t nbytes,
                                    int error)
{
    if (error != 0)
    {
        fprintf(stderr, " ed_write_error failed\n");
    }
} /* ed_write_error_callback() */
