/**************************************************************************
*
* Program:      Nexus-Talk Client
*               talk_client.c
*
* Written at:   High-Performance Computing Laboratory
*               Department of Computer Science
*               Northern Illinois University
*               DeKalb, IL
*
* Authors:      Robert Bach
*               John Bresnahan
*
* Date Written: 2/97
*
* Description:
*
*     This is the Nexus version of the Unix "talk" application.  As such, it
*     is a curses-based program.  It is intended to be an example of how
*     to use the Nexus "attach" functionality in a client/server application.
* 
*     This is the talk client application.  Before running the client the 
*     talk server must be started first.  The "welcome" screen of the
*     talk server displays the attach URL that clients wishing to attach
*     to it must use.
*
*     Once the talk client has that URL, it uses it to ask the talk server
*     for a connection using nexus_attach().  The talk server may grant
*     the request, in which case nexus_attach() returns with 0 and we
*     get a startpoint that points back to the server.  If the request
*     is denied, nexus_attach() returns with a non-zero value.
*
*     Note that if the request is granted, the client has a startpoint
*     that points back to the server.  The server does not have a startpoint
*     that points back to the client.  If the client wishes the server
*     to have a startpoint that points back to the client, it may use
*     the startpoint it just received that points to the server to send
*     a startpoint to the server.
*
*     This program needs the attach URL generated by the talk server.
*     When the talk server starts executing it not only displays the
*     attach URL on its "welcome" screen, but also writes it to
*     /tmp/talk_url as well.  If the talk client can access that same
*     file, then it will.  If the talk client cannot access that file
*     (e.g., it's being executed on a remote machine), then the user
*     may pass the attach URL to the talk client as a command line argument.
*
* To run this program:
*
*         If the talk client has access to /tmp/talk_url created 
*         by the talk server.
*
*             talk_client
*
*     or
*
*         If the talk client does not have access to /tmp/talk_url
*         created by the talk server.  Then the attach URL displayed
*         on the talk server "welcome" screen must be passed as a
*         command line argument.  For example, if the attach URL
*         is "x-nexus://olympus:44431/"
*
*             talk_client -a x-nexus://olympus:44431/
*
*     The talk session is terminated when either side enters control-x.
*
* Output:
*
*     A curses-based talk session.
*
**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <nexus.h>
#include <curses.h>
#include "display.h"

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

static void reply_handler(nexus_endpoint_t *endpoint, 
			    nexus_buffer_t *buffer,
			    nexus_bool_t is_non_threaded_handler);
static void receive_char_handler(nexus_endpoint_t *endpoint, 
				    nexus_buffer_t *buffer, 
				    nexus_bool_t is_non_threaded_handler);

/*********************************/
/*                               */
/* Nexus Handler Functions Table */
/*                               */
/*********************************/

#define REPLY_HANDLER_ID        0
#define RECEIVE_CHAR_HANDLER_ID 1

static nexus_handler_t handlers[] = 
{ 
    {NEXUS_HANDLER_TYPE_NON_THREADED, reply_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED, receive_char_handler}
};

/*******************/
/*                 */
/* Data Structures */
/*                 */
/*******************/

/* Monitor                                                 */
/*                                                         */
/* we can simulate a monitor by defining a struct that has */
/* at least a nexus_mutex_t to provide mutually exclusive  */
/* access to the monitor.                                  */
/* in this application we also need a monitor condition    */
/* variable and a flag.                                    */

typedef struct
{
    /* for mutually exclusive access to monitor */
    nexus_mutex_t mutex;

    /* a monitor condition variable and flag */
    nexus_cond_t cond;
    nexus_bool_t done;
} monitor_t;

/********************/
/*                  */
/* Global Variables */
/*                  */
/********************/

static nexus_endpointattr_t	EpAttr;
static nexus_endpoint_t		GlobalEndpoint;

static nexus_startpoint_t	TalkServerSP;
static monitor_t		Monitor;
nexus_bool_t			RemoteKill;

static struct UserScreen 	User1;
static struct UserScreen 	User2;

/*******************/
/*                 */
/* Local Functions */
/*                 */
/*******************/

static void attach_to_server(char *attach_url);
static nexus_bool_t read_attach_file(char *url, char *file);

/****************************/
/*                          */
/* Required Nexus Functions */
/*                          */
/****************************/

/*
 * NexusExit() (required of all Nexus programs)
 */
void NexusExit(void)
{
} /* end NexusExit() */

/*
 * NexusAcquiredAsNode() (required of all Nexus programs)
 */
int NexusAcquiredAsNode(nexus_startpoint_t *startpoint)
{
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);
    return 0;
} /* end NexusAcquiredAsNode() */

/*
 * NexusBoot() (required of all Nexus programs)
 */
int NexusBoot(nexus_startpoint_t *startpoint)
{
    /****************************************/
    /* NexusBoot() must bind its startpoint */
    /* arg to an endpoint in this context   */
    /****************************************/

    /*
    ** before you can bind a startpoint to an endpoint, that endpoint  
    ** must be initialized.  endpoints are initialized only once,     
    ** and once initialized, may be bound to several startpoints.    
    **
    ** initializing an endpoint requires an endpointattr.  again, before you 
    ** can use an endpointattr to initialize an endpoint, the endpointattr
    ** must also be initialized.  endpointattrs are also intialized only
    ** once, and once initialized, may be used to initialize many endpoints. 
    **
    ** So, the sequence of events to bind a startpoint to an endpoint are:
    **
    **    1. initialize an endpointattr 
    **       - done once per endpointattr
    **       - once done, this endpointattr can be used to 
    **         initialize several endpoints
    **    2. initialize endpoint using an initialized enpointattr
    **       - done once per endpoint
    **       - once done, this endpoint can be bound to several startpoints
    **    3. bind startpoint to initialized endpoint
    */

    /* initializing an endpointattr */
    nexus_endpointattr_init(&EpAttr);
    nexus_endpointattr_set_handler_table(&EpAttr, 
				    handlers, 
				    sizeof(handlers)/sizeof(nexus_handler_t));

    /* initializing endpoint ... no particular address */
    nexus_endpoint_init(&GlobalEndpoint, &EpAttr);

    /* binding startpoint to endpoint */
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);

    return NEXUS_SUCCESS;

} /* end NexusBoot() */

/********/
/*      */
/* MAIN */
/*      */
/********/

int main(int argc, char **argv)
{
    nexus_startpoint_t my_sp;
    int i, ch;
    nexus_buffer_t buffer;
    char attach_url[1024];
    nexus_node_t *nodes;
    int n_nodes;

    /***************************************************/
    /* calls to nexus_init() and nexus_start() are     */
    /* required at the beginning of all Nexus programs */
    /***************************************************/

    nexus_init(&argc, 
                &argv,
               "NEXUS_ARGS", /* conf info env variable          */
               "nx",         /* package designator              */
               NULL,         /* package args init callback func */
               NULL,         /* usage message callback func     */
               NULL,         /* new process params func         */
               NULL,         /* module list                     */
               &nodes,
               &n_nodes);

    /* at this point :                                       */
    /*       nodes[0].starpoint is NULL                      */

    nexus_start();

    /* See if we have 1 node */
    if (n_nodes != 1)
    {
        nexus_stdio_lock();
        fprintf(stderr, "ERROR: Must run with exactly 1 node.\n");
        nexus_stdio_unlock();
        nexus_abort();
    } /* endif */

    /******************/
    /* Initialization */
    /******************/

    /* getting attach_url */

    /* checking for optional [-a attach_url] */
    if (argc > 1)
    {
	if (argc != 3 || strcmp(argv[1], "-a"))
	{
	    /* must always nexus_stdio_lock/nexus_stdio_unlock */
	    /* when doing any I/O */
	    nexus_stdio_lock();
	    fprintf(stderr, "usage: %s [-a attach_url]\n", argv[0]);
	    nexus_stdio_unlock();
	    nexus_abort();
	}
	else
	{
	    /* optional [-a attach_url] supplied */
	    strcpy(attach_url, argv[2]);
	} /* endif */
    } 
    else
    {
	/* no command line args ... must get attach_url from URL_FILE */
	if (!read_attach_file(attach_url, URL_FILE))
	{
	    /* must always nexus_stdio_lock/nexus_stdio_unlock */
	    /* when doing any I/O */
	    nexus_stdio_lock();

	    fprintf(stderr, 
		    "ERROR: Could not read attach_url from file: %s\n", 
		    URL_FILE);
	    fprintf(stderr, 
		    "talk_server may not be running on this file system.\n");
	    fprintf(stderr, 
		"If you know your party's attach_url type: %s -a attach_url\n",
		argv[0]);

	    nexus_stdio_unlock();
	    nexus_abort();
	} /* endif */
    } /* endif */

    /* initializing remote kill flag */
    RemoteKill = NEXUS_FALSE;

    /* Setup monitor that I'll suspend on */
    nexus_mutex_init(&(Monitor.mutex), (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&(Monitor.cond), (nexus_condattr_t *) NULL);

    /* entering the monitor and clearing the flag */
    nexus_mutex_lock(&(Monitor.mutex));
    Monitor.done = NEXUS_FALSE;
    nexus_mutex_unlock(&(Monitor.mutex));

    /*************************/
    /* attach to talk_server */
    /*************************/

    attach_to_server(attach_url);

    /* I now have a startpoint to the talk_server, TalkServerSP */

    /*********************************/
    /* send talk_server a startpoint */
    /* to myself and wait for ACK    */
    /*********************************/

    nexus_startpoint_bind(&my_sp, &GlobalEndpoint);

    nexus_buffer_init(&buffer, nexus_sizeof_startpoint(&my_sp, 1), 0);
    nexus_put_startpoint_transfer(&buffer, &my_sp, 1);

    /* sending startpoint to talk_server and invoking attach_handler() there */
    nexus_send_rsr(&buffer, 
		    &TalkServerSP, 
		    REPLY_HANDLER_ID, 
		    NEXUS_TRUE, 
		    NEXUS_FALSE);

    /* waiting for ACK */

    /* the implementation of nexus_cond_wait() and nexus_cond_signal()     */
    /* makes it possible for nexus_cond_wait() to experience a 'false      */
    /* wakeup', i.e., return without having had a nexus_cond_signal()      */
    /* applied to it.                                                      */
    /*                                                                     */
    /* this is why we must wait on a condition variable in the manner      */
    /* we do below, with a loop and a nexus_bool_t done, rather than       */
    /* simply:                                                             */
    /*     nexus_mutex_lock(&(Monitor.mutex));                             */
    /*     nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));             */
    /*     nexus_mutex_unlock(&(Monitor.mutex));                           */

    nexus_mutex_lock(&(Monitor.mutex));
    while (!(Monitor.done))
    {
       nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));
    } /* endwhile */
    nexus_mutex_unlock(&(Monitor.mutex));

    /* at this point, flag in monitor has been set      */
    /* ... we have successfully attached to talk_server */
    /* and have sent him a startpoint to ourself        */

    /******************************************************/
    /* talk session --- ends when either side hits cntl-X */
    /******************************************************/

    InitScreen();
    StartTalk(&User1, &User2, attach_url);
  
    ch = 0;
    while (ch != CTRL_X && !RemoteKill)
    {
	ch = getchar();
	fflush(stdin);
	if(!RemoteKill)
	{ 
	    /* displaying single character on screen */
	    AddChar(&User1, ch);

            /* sending single character to talk_server */
	    nexus_buffer_init(&buffer, nexus_sizeof_int(1), 0);
	    nexus_put_int(&buffer, &ch, 1);
	    nexus_send_rsr(&buffer, 
			&TalkServerSP, 
			RECEIVE_CHAR_HANDLER_ID, 
			NEXUS_TRUE, 
			NEXUS_FALSE);
	} /* endif */
    } /* endwhile */

    /************/ 
    /* Shutdown */
    /************/ 

    endwin();
    ThankYou();

    /********************/ 
    /* Initial Clean-up */
    /********************/ 

    nexus_startpoint_destroy(&TalkServerSP);

    nexus_mutex_destroy(&(Monitor.mutex));
    nexus_cond_destroy(&(Monitor.cond));

    /*********************/
    /* Standard Clean-up */
    /*********************/

    nexus_endpoint_destroy(&GlobalEndpoint);
    nexus_endpointattr_destroy(&EpAttr);

    /* Free the node list */
    for(i = 0; i < n_nodes; i++)
    {
	nexus_free(nodes[i].name);
	/* OK to call startpoint destroy on nodes[0].startpoint */
	/* even though nodes[0].startpoint known to be NULL     */
	nexus_startpoint_destroy(&(nodes[i].startpoint));
    } /* endfor */
    nexus_free(nodes);
    
    /* Terminate master process */
    nexus_context_destroy(NEXUS_FALSE);

    nexus_printf("main(): ERROR: We should never get here.\n");

    return 0;

} /* end main() */

/*******************/
/*                 */
/* Local Functions */
/*                 */
/*******************/

/*
 * attach_to_server()
 *
 * Attachment in Nexus is done by making a request to attach to another
 * process designated by a URL.  If the request is granted, we get a 0
 * return code and a startpoint that points back to the granting process.
 *
 * NOTE: We get a startpoint to the process we just attached to, they
 *       do not get a startpoint back to us.  If we need them to have
 *       a startpoint back to us, we need to send them one using the
 *       startpoint we just received ... the one that points back to them.
 *
 * If the request is denied, we get a non-zero return code back.
 *
 */
static void attach_to_server(char *attach_url)
{
    int rc;

    rc = nexus_attach(attach_url, &TalkServerSP);

    /* if things went OK, we now have a startpoint */
    /* to the talk_server, TalkServerSP            */

    if (rc != 0)
    {
	/* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
	nexus_stdio_lock();
	fprintf(stderr, 
		"ERROR: Could not attach to server using url: %s rc %d\n", 
		attach_url, rc);
	nexus_stdio_unlock();
	nexus_abort();
    } /* endif */

} /* end attach_to_server() */

/*
 * read_attach_file()
 */
static nexus_bool_t read_attach_file(char *attach_url, char *file)
{
    FILE *fp;
    nexus_bool_t rc;

    /* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
    nexus_stdio_lock();
    if ((fp = fopen(file, "r")) == (FILE *) NULL)
	rc = NEXUS_FALSE;
    else
    {
	rc = NEXUS_TRUE;

	fscanf(fp, "%s", attach_url);

	fclose(fp);
    } /* endif */
    nexus_stdio_unlock();

    return rc;

} /* end read_attach_file() */

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

/*
 * reply_handler()
 */
static void reply_handler(nexus_endpoint_t *endpoint, 
			    nexus_buffer_t *buffer,
			    nexus_bool_t is_non_threaded_handler)
{
    /* clean-up */
    nexus_buffer_destroy(buffer);

    /* the implementation of nexus_cond_wait() and nexus_cond_signal()  */
    /* makes it possible for nexus_cond_wait() to experience a 'false   */
    /* wakeup', i.e., return without having had a nexus_cond_signal()   */
    /* applied to it.                                                   */
    /*                                                                  */
    /* this is why we must signal a condition variable in the manner    */
    /* we do below, with a nexus_bool_t done, rather than simply:       */
    /*     nexus_mutex_lock(&(Monitor.mutex));                          */
    /*     nexus_cond_signal(&(Monitor.cond));                          */
    /*     nexus_mutex_unlock(&(Monitor.mutex));                        */

    nexus_mutex_lock(&(Monitor.mutex));
    Monitor.done = NEXUS_TRUE;
    nexus_cond_signal(&(Monitor.cond));
    nexus_mutex_unlock(&(Monitor.mutex));   

} /* end reply_handler() */

/*
 * receive_char_handler()
 */
static void receive_char_handler(nexus_endpoint_t *endpoint, 
				    nexus_buffer_t *buffer, 
				    nexus_bool_t is_non_threaded_handler)
{
    int ch;

    /* extracting character from buffer */
    nexus_get_int(buffer, &ch, 1);

    if (ch == CTRL_X) 
    {
        /* remote user hit ctrl-X */
	RemoteKill = NEXUS_TRUE;
	PartyLeft(&User1);
    }
    else
        /* writing character to screen */
	AddChar(&User2, ch);

    /* clean-up */
    nexus_buffer_destroy(buffer);

} /* end receive_char_handler() */
