/**************************************************************************
*
* Program:      Nexus-Talk Server
*               talk_server.c
*
* Written at:   High-Performance Computing Laboratory
*               Department of Computer Science
*               Northern Illinois University
*               DeKalb, IL
*
* Authors:      Robert Bach
*               John Bresnahan
*
* Date Written: 2/97
*
* Description:
* 
*     This is the Nexus version of the Unix "talk" application.  As such, it
*     is a curses-based program.  It is intended to be an example of how
*     to use the Nexus "attach" functionality in a client/server application.
* 
*     The server in this example is designed to service exactly one client.
*     Once that connection has been established, no other clients may
*     attach to this server.  Further, once the client/server connection
*     has been established and the session runs to completion, upon
*     disconnecting BOTH the server and the client terminate.  
*     
*     This is unlike traditional client/server applications where a server
*     my service a number of clients simultaneously and where the server
*     does not terminate after servicing exactly one client.  But, this
*     is an example whose purpose is to demonstrate the Nexus attach and
*     detach functionality.  Extending this example to allow multiple 
*     clients and establish an omnipresent server should be straightforward.
*
*     Attachment in Nexus is performed by one process "allowing attachments",
*     nexus_allow_attach().  This returns a <hostname> and <port> from which
*     we can construct a URL, an attach ULR.  For example, the attach URL 
*     constructed by a hostname=olympus and port=44431 would look like this:
*
*                     x-nexus://olympus:44431/
*     
*     It is this attach URL that the other process must pass (as a character
*     string) to nexus_attach() when making a request to attach to us.
*     When such a request is made, our attachment approval function,
*     attach_requested() is automatically (asynchronously) executed.  It
*     is that routine where we decide whether or not to allow attachment.
*     If permission is granted, we give the requesting process a startpoint
*     to ourself.  Note that we do not get a startpoint to the process
*     we grant attach permission to.  If they want to give us a startpoint
*     they can do so by using the startpoint we just gave them.
* 
*     We can also stop listening for attachments by calling 
*     nexus_disallow_attach().
*
* To run this program:
* 
*     talk_server
*
*     The talk session is terminated when either side enters control-x.
* 
* Output:
*
*     A curses-based talk session.
*
**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <nexus.h>
#include <curses.h>
#include "display.h"

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

static void attach_handler(nexus_endpoint_t *endpoint, 
			    nexus_buffer_t *buffer, 
			    nexus_bool_t is_non_threaded_handler);
static void receive_char_handler(nexus_endpoint_t *endpoint, 
				nexus_buffer_t *buffer,
				nexus_bool_t is_non_threaded_handler);

/*********************************/
/*                               */
/* Nexus Handler Functions Table */
/*                               */
/*********************************/

#define ATTACH_HANDLER_ID       0
#define RECEIVE_CHAR_HANDLER_ID 1

static nexus_handler_t handlers[] = 
{ 
    {NEXUS_HANDLER_TYPE_NON_THREADED, attach_handler},
    {NEXUS_HANDLER_TYPE_NON_THREADED, receive_char_handler}
};

/*******************/
/*                 */
/* Data Structures */
/*                 */
/*******************/

/* Monitor                                                 */
/*                                                         */
/* we can simulate a monitor by defining a struct that has */
/* at least a nexus_mutex_t to provide mutually exclusive  */
/* access to the monitor.                                  */
/* in this application we also need a monitor condition    */
/* variable and two flags.                                 */

typedef struct
{
    /* for mutually exclusive access to monitor */
    nexus_mutex_t mutex;

    /* a monitor condition variable and flag */
    nexus_cond_t  cond;
    nexus_bool_t  done;
    nexus_bool_t  connection_established;
} monitor_t;

/********************/
/*                  */
/* Global Variables */
/*                  */
/********************/

static nexus_endpointattr_t	EpAttr;
static nexus_endpoint_t 	GlobalEndpoint;

static nexus_startpoint_t	TalkClientSP;
static monitor_t		Monitor;
static nexus_bool_t		RemoteKill;

static struct UserScreen 	User1;
static struct UserScreen 	User2;

/*******************/
/*                 */
/* Local Functions */
/*                 */
/*******************/

static int attach_requested(void *arg, char *url, nexus_startpoint_t *sp);
static void allow_attachments(unsigned short *attach_port, char *attach_url);
static nexus_bool_t write_attach_file(char *attach_url);

/****************************/
/*                          */
/* Required Nexus Functions */
/*                          */
/****************************/

/*
 * NexusExit() (required of all Nexus programs)
 */
void NexusExit(void)
{
} /* end NexusExit() */

/*
 * NexusAcquiredAsNode() (required of all Nexus programs)
 */
int NexusAcquiredAsNode(nexus_startpoint_t *startpoint)
{
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);
    return 0;
} /* end NexusAcquiredAsNode() */

/*
 * NexusBoot() (required of all Nexus programs)
 */
int NexusBoot(nexus_startpoint_t *startpoint)
{
    /****************************************/
    /* NexusBoot() must bind its startpoint */
    /* arg to an endpoint in this context   */
    /****************************************/

    /*
    ** before you can bind a startpoint to an endpoint, that endpoint  
    ** must be initialized.  endpoints are initialized only once,     
    ** and once initialized, may be bound to several startpoints.    
    **
    ** initializing an endpoint requires an endpointattr.  again, before you 
    ** can use an endpointattr to initialize an endpoint, the endpointattr
    ** must also be initialized.  endpointattrs are also intialized only
    ** once, and once initialized, may be used to initialize many endpoints. 
    **
    ** So, the sequence of events to bind a startpoint to an endpoint are:
    **
    **    1. initialize an endpointattr 
    **       - done once per endpointattr
    **       - once done, this endpointattr can be used to 
    **         initialize several endpoints
    **    2. initialize endpoint using an initialized enpointattr
    **       - done once per endpoint
    **       - once done, this endpoint can be bound to several startpoints
    **    3. bind startpoint to initialized endpoint
    */

    /* initializing an endpointattr */
    nexus_endpointattr_init(&EpAttr);
    nexus_endpointattr_set_handler_table(&EpAttr, 
				    handlers, 
				    sizeof(handlers)/sizeof(nexus_handler_t));

    /* initializing endpoint ... no particular address */
    nexus_endpoint_init(&GlobalEndpoint, &EpAttr);

    /* binding startpoint to endpoint */
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);

    return NEXUS_SUCCESS;

} /* end NexusBoot() */

/********/
/*      */
/* MAIN */
/*      */
/********/

int main(int argc, char **argv)
{
    int i, ch;
    nexus_buffer_t buffer;
    char attach_url[1024];
    unsigned short attach_port;
    nexus_node_t *nodes;
    int n_nodes;

    /***************************************************/
    /* calls to nexus_init() and nexus_start() are     */
    /* required at the beginning of all Nexus programs */
    /***************************************************/

    nexus_init(&argc, 
                &argv,
               "NEXUS_ARGS", /* conf info env variable          */
               "nx",         /* package designator              */
               NULL,         /* package args init callback func */
               NULL,         /* usage message callback func     */
               NULL,         /* new process params func         */
               NULL,         /* module list                     */
               &nodes,
               &n_nodes);

    /* at this point :                                       */
    /*       nodes[0].starpoint is NULL                      */

    nexus_start();

    /* See if we have 1 node */
    if (n_nodes != 1)
    {
        nexus_stdio_lock();
        fprintf(stderr, "ERROR: Must run with exactly 1 node.\n");
        nexus_stdio_unlock();
        nexus_abort();
    } /* endif */

    /******************/
    /* Initialization */
    /******************/

    /**********************************************************/
    /* Initialize the monitor that will be used for signaling */
    /* that the talk_client has successfully attached to us.  */
    /**********************************************************/

    /* initializing the mutex that controls access to the monitor */
    nexus_mutex_init(&(Monitor.mutex), (nexus_mutexattr_t *) NULL);
    /* initializing the condition variable */
    nexus_cond_init(&(Monitor.cond), (nexus_condattr_t *) NULL);

    /* entering the monitor and clearing the flags */
    nexus_mutex_lock(&(Monitor.mutex));
    Monitor.done = NEXUS_FALSE;
    Monitor.connection_established = NEXUS_FALSE;
    nexus_mutex_unlock(&(Monitor.mutex));

    /* initializing remote kill flag */
    RemoteKill = NEXUS_FALSE;

    /***********************/
    /* allowing attachment */
    /***********************/

    allow_attachments(&attach_port, attach_url);

    /* advertising URL for attachment in URL_FILE */
    if (!write_attach_file(attach_url))
    {
	nexus_stdio_lock();
	fprintf(stderr, "ERROR: Could not write file: %s\n", URL_FILE);
	fprintf(stderr, "File may already exist.\n");
	nexus_stdio_unlock();
	nexus_abort();
    } /* endif */

    InitScreen();
    IntroScreen(attach_url);

    /**************************************************/
    /* wait (using monitor) for talk_client to attach */
    /**************************************************/

    /* the implementation of nexus_cond_wait() and nexus_cond_signal()     */
    /* makes it possible for nexus_cond_wait() to experience a 'false      */
    /* wakeup', i.e., return without having had a nexus_cond_signal()      */
    /* applied to it.                                                      */
    /*                                                                     */
    /* this is why we must wait on a condition variable in the manner      */
    /* we do below, with a loop and a nexus_bool_t done, rather than       */
    /* simply:                                                             */
    /*     nexus_mutex_lock(&(Monitor.mutex));                             */
    /*     nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));             */
    /*     nexus_mutex_unlock(&(Monitor.mutex));                           */

    nexus_mutex_lock(&(Monitor.mutex));
    while (!(Monitor.done))
    {
	nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));
    } /* endwhile */
    nexus_mutex_unlock(&(Monitor.mutex));

    /* at this point, flag in monitor has been set   */
    /* ... the talk_client has successfully attached */

    /******************************************************/
    /* talk session --- ends when either side hits cntl-X */
    /******************************************************/

    StartTalk(&User1, &User2, attach_url);
    
    ch = 0;
    while (ch != CTRL_X && !RemoteKill)
    {
	ch = getchar();
	fflush(stdin);
	if(!RemoteKill)
	{ 
	    /* displaying single character on screen */
	    AddChar(&User1, ch);

	    /* sending single character to talk_client */
	    nexus_buffer_init(&buffer, nexus_sizeof_int(1), 0);
	    nexus_put_int(&buffer, &ch, 1);
	    nexus_send_rsr(&buffer, 
			&TalkClientSP, 
			RECEIVE_CHAR_HANDLER_ID, 
			NEXUS_TRUE, 
			NEXUS_FALSE);
	} /* endif */
    } /* endwhile */

    /************/ 
    /* Shutdown */
    /************/ 

    endwin();
    ThankYou();

    nexus_disallow_attach(attach_port);

    /********************/ 
    /* Initial Clean-up */
    /********************/ 

    if (unlink(URL_FILE))
       nexus_printf("\nError unlinking file\n");

    nexus_startpoint_destroy(&TalkClientSP);

    nexus_mutex_destroy(&(Monitor.mutex));
    nexus_cond_destroy(&(Monitor.cond));

    /*********************/
    /* Standard Clean-up */
    /*********************/

    nexus_endpoint_destroy(&GlobalEndpoint);
    nexus_endpointattr_destroy(&EpAttr);

    /* Free the node list */
    for(i = 0; i < n_nodes; i++)
    {
	nexus_free(nodes[i].name);
	/* OK to call startpoint destroy on nodes[0].startpoint */
	/* even though nodes[0].startpoint known to be NULL     */
	nexus_startpoint_destroy(&(nodes[i].startpoint));
    } /* endfor */
    nexus_free(nodes);
    
    /* Terminate master process */
    nexus_context_destroy(NEXUS_FALSE);

    nexus_printf("main(): ERROR: We should never get here.\n");

    return 0;

} /* end main() */

/*******************/
/*                 */
/* Local Functions */
/*                 */
/*******************/

/*
 * attach_requested()
 *
 * This is the attachment approval function.  When the server started
 * allowing attachments (nexus_allow_attach()), it specified this approval 
 * function.
 *
 * This function gets automatically (asynchronously) called when a
 * client calls nexus_attach() specifying this server's URL.
 *
 * The job of all attachment approval functions is to decide whether to
 * allow the requesting process to "attach" to this one.  If the 
 * attachment is granted, this function must bind the startpoint
 * passsed as the third arg to a local endpoint and return 0.
 * If attachment is denied, this function simply returns a non-zero value.
 *
 * In our example program we have decided to allow at most one connection
 * at a time.  In a more traditional client/server application where the
 * server might service multiple clients simultaneously, the logic for
 * this routine should be modified in the obvious way.
 *
 */
static int attach_requested(void *arg, char *url, nexus_startpoint_t *sp)
{
     int rc;

    nexus_mutex_lock(&(Monitor.mutex));
    if (Monitor.connection_established)
    {
	/* another talk_client has already connected to us */
	rc = 1;
    }
    else
    {
	/* this is the first talk_client */
	rc = 0;
	Monitor.connection_established = NEXUS_TRUE;
    } /* endif */
    nexus_mutex_unlock(&(Monitor.mutex));

     /* binding received startpoint to one of our endpoints */
     if (rc == 0)
     {
	 nexus_startpoint_bind(sp, &GlobalEndpoint);
     } /* endif */

     return rc;

} /* end attach_requested() */

/*
 * allow_attachments()
 *
 * By calling nexus_allow_attach() we allow other Nexus programs
 * to attach to us.  nexus_allow_attach() tells us the hostname (attach_host)
 * and port number (attach_port) that it will start listening for attachments.
 * We use attach_host and attach_port to construct a URL that attach 
 * clients need to use to attach to us.
 *
 * In calling nexus_allow_attach() we also name an approval function, in
 * this case attach_requested(), which is a user-supplied function that
 * gets called every time a client requests an attachment to us.
 *
 */
static void allow_attachments(unsigned short *attach_port, char *attach_url)
{
    int rc;
    char *attach_host;

    /* allow other Nexus programs to attach to us */
    *attach_port = 0;
    rc = nexus_allow_attach(attach_port,      /* port            */
			    &attach_host,     /* host            */
			    attach_requested, /* approval_func() */
			    NULL);
    if (rc != 0)
    {
       /* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
       nexus_stdio_lock();
       fprintf(stderr, "ERROR: nexus_allow_attach() failed: rc=%d\n", rc);
       nexus_stdio_unlock();
       nexus_abort();
    } /* endif */

    sprintf(attach_url, "x-nexus://%s:%hu/", attach_host, *attach_port);

} /* end allow_attachments() */

/*
 * write_attach_file()
 */
static nexus_bool_t write_attach_file(char *attach_url)
{
    FILE *fp;
    nexus_bool_t rc;

   /* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
    nexus_stdio_lock();
    if ((fp = fopen(URL_FILE, "w")) == (FILE *) NULL)
	rc = NEXUS_FALSE;
    else
    {
	rc = NEXUS_TRUE;

	fprintf(fp, "%s\n", attach_url);

	fclose(fp);
    } /* endif */
    nexus_stdio_unlock();

    return rc;

} /* end write_attach_file() */

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

/*
 * attach_handler()
 */
static void attach_handler(nexus_endpoint_t *endpoint, 
			    nexus_buffer_t *buffer, 
			    nexus_bool_t is_non_threaded_handler)
{

    nexus_buffer_t reply_buffer;

    /* extracting talk_client startpoint from buffer */
    nexus_get_startpoint(buffer, &TalkClientSP, 1);

    /* replying to talk_client with empty message */
    /* ACK, invoking reply_handler() there.       */
    nexus_buffer_init(&reply_buffer, 0, 0);
    nexus_send_rsr(&reply_buffer, 
		    &TalkClientSP, 
		    ATTACH_HANDLER_ID, 
		    NEXUS_TRUE, 
		    NEXUS_FALSE);

    /* clean-up */
    nexus_buffer_destroy(buffer);

    /* the implementation of nexus_cond_wait() and nexus_cond_signal()  */
    /* makes it possible for nexus_cond_wait() to experience a 'false   */
    /* wakeup', i.e., return without having had a nexus_cond_signal()   */
    /* applied to it.                                                   */
    /*                                                                  */
    /* this is why we must signal a condition variable in the manner    */
    /* we do below, with a nexus_bool_t done, rather than simply:       */
    /*     nexus_mutex_lock(&(Monitor.mutex));                          */
    /*     nexus_cond_signal(&(Monitor.cond));                          */
    /*     nexus_mutex_unlock(&(Monitor.mutex));                        */

    nexus_mutex_lock(&(Monitor.mutex));
    Monitor.done = NEXUS_TRUE;
    nexus_cond_signal(&(Monitor.cond));
    nexus_mutex_unlock(&(Monitor.mutex));      

} /* end attach_handler() */

/*
 * receive_char_handler()
 */
static void receive_char_handler(nexus_endpoint_t *endpoint, 
				    nexus_buffer_t *buffer,
				    nexus_bool_t is_non_threaded_handler)
{
    int ch;

    /* extracting character from buffer */
    nexus_get_int(buffer, &ch, 1);  

    if (ch == CTRL_X)
    { 
	/* remote user hit ctrl-X */
	RemoteKill = NEXUS_TRUE;
	PartyLeft(&User1); 
    }
    else
	/* writing character to screen */
	AddChar(&User2, ch);

    /* clean-up */
    nexus_buffer_destroy(buffer);

} /* end receive_char_handler() */
