/**************************************************************************
*
* Program:      Single Remote Service Request
*               single_rsr.c
*
* Written at:   High-Performance Computing Laboratory
*               Department of Computer Science
*               Northern Illinois University
*               DeKalb, IL
*
* Author:       Nicholas T. Karonis
*
* Date Written: 2/97
*
* Description:
*
*     This program contains the *minimum* amount of code required to 
*     send a message from one process to another.
* 
*     This program requires a single command line argument, an integer.
*     It places that integer into a Nexus buffer and sends it to an
*     awaiting process.  Sending a message is called initiating a 
*     "remote service request" (rsr).  An rsr invokes a specified function
*     (called a handler routine, or simply handler) on the receiving 
*     process.  In this program, the rsr instructs the receiving process
*     to execute the function rcv_handler(), which is a routine that 
*     is found in this file.
*
*     The receiving process extracts the integer from the buffer and writes
*     that integer to an output file (single_rsr.output).
*
*     Running this program requires one integer command line argument and
*     exactly one other process.
*
* To run this program:
*
*     single_rsr 123 -nx -n 1
*
*         This will start the master and the slave on your current machine.
*
*         This program uses nexus_printf() which sends everything to stdout,
*         so executing the program in this manner will likely display all
*         output (including slave output) to your screen.
*
*         It will send the integer "123".
*
* Output:
*
*         olympus#1:c0:t0:p13691: Slave received 123
*
**************************************************************************/

#include <stdio.h>
#include <nexus.h>

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

static void rcv_handler(nexus_endpoint_t *endpoint, 
                            nexus_buffer_t *buffer,
                            nexus_bool_t is_non_threaded_handler);

/*********************************/
/*                               */
/* Nexus Handler Functions Table */
/*                               */
/*********************************/

#define RCV_HANDLER_ID 0

static nexus_handler_t handlers[] = 
{
    {NEXUS_HANDLER_TYPE_NON_THREADED, rcv_handler}
};

/********************/
/*                  */
/* Global Variables */
/*                  */
/********************/

static nexus_endpointattr_t EpAttr;
static nexus_endpoint_t     GlobalEndpoint;

/****************************/
/*                          */
/* Required Nexus Functions */
/*                          */
/****************************/

/*
 * NexusExit() (required of all Nexus programs)
 */
void NexusExit(void)
{
} /* end NexusExit() */

/*
 * NexusAcquiredAsNode() (required of all Nexus programs)
 */
int NexusAcquiredAsNode(nexus_startpoint_t *startpoint)
{
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);
    return 0;
} /* end NexusAcquiredAsNode() */

/*
 * NexusBoot() (required of all Nexus programs)
 */
int NexusBoot(nexus_startpoint_t *startpoint)
{
    /****************************************/
    /* NexusBoot() must bind its startpoint */
    /* arg to an endpoint in this context   */
    /****************************************/

    /*
    ** before you can bind a startpoint to an endpoint, that endpoint  
    ** must be initialized.  endpoints are initialized only once,     
    ** and once initialized, may be bound to several startpoints.    
    **
    ** initializing an endpoint requires an endpointattr.  again, before you 
    ** can use an endpointattr to initialize an endpoint, the endpointattr
    ** must also be initialized.  endpointattrs are also intialized only
    ** once, and once initialized, may be used to initialize many endpoints. 
    **
    ** So, the sequence of events to bind a startpoint to an endpoint are:
    **
    **    1. initialize an endpointattr 
    **       - done once per endpointattr
    **       - once done, this endpointattr can be used to 
    **         initialize several endpoints
    **    2. initialize endpoint using an initialized enpointattr
    **       - done once per endpoint
    **       - once done, this endpoint can be bound to several startpoints
    **    3. bind startpoint to initialized endpoint
    */

    /* initializing an endpointattr */
    nexus_endpointattr_init(&EpAttr);
    nexus_endpointattr_set_handler_table(&EpAttr, 
				    handlers, 
				    sizeof(handlers)/sizeof(nexus_handler_t));

    /* initializing endpoint ... no particular address */
    nexus_endpoint_init(&GlobalEndpoint, &EpAttr);

    /* binding startpoint to endpoint */
    nexus_startpoint_bind(startpoint, &GlobalEndpoint);

    return NEXUS_SUCCESS;

} /* end NexusBoot() */

/********/
/*      */
/* MAIN */
/*      */
/********/

int main(int argc, char **argv)
{
    nexus_node_t *nodes;
    int n_nodes;
    nexus_buffer_t buffer;
    int i;

    /***************************************************/
    /* calls to nexus_init() and nexus_start() are     */
    /* required at the beginning of all Nexus programs */
    /***************************************************/

    nexus_init(&argc, 
		&argv,
	       "NEXUS_ARGS", /* conf info env variable          */
	       "nx",         /* package designator              */
	       NULL,         /* package args init callback func */
	       NULL,         /* usage message callback func     */
	       NULL,         /* new process params func         */
	       NULL,         /* module list                     */
	       &nodes,
	       &n_nodes);

    /* at this point :                                       */
    /*       nodes[0].starpoint is NULL                      */
    /*       0 < i < n_nodes, nodes[i].starpoint is non-NULL */
    /*                        and points to slave i          */

    nexus_start();

    /* only the master proceeds from here ... all slaves sit    */
    /* around waiting for their handler routines to get called. */

    /* See if we have 2 nodes */
    if (n_nodes != 2)
    {
	/* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
	nexus_stdio_lock();
	fprintf(stderr, "ERROR: Must run with exactly 2 nodes.\n");
	nexus_stdio_unlock();
	nexus_abort();
    } /* endif */

    /* checking command line arg */
    if (argc != 2)
    {
	/* must always nexus_stdio_lock/nexus_stdio_unlock when doing any I/O */
	nexus_stdio_lock();
	fprintf(stderr, "usage: %s <intvalue>\n", argv[0]);
	nexus_stdio_unlock();
	nexus_abort();
    } /* endif */

    i = atoi(argv[1]);

    /**************************/
    /* remote service request */
    /**************************/

    /* initializing the buffer to hold a single integer */
    nexus_buffer_init(&buffer, nexus_sizeof_int(1), 0);

    /* placing single integer into buffer */
    nexus_put_int(&buffer, &i, 1);

    /* sending buffer to other process and invoking rcv_handler() there */
    nexus_send_rsr(&buffer, 
		&(nodes[1].startpoint), 
		RCV_HANDLER_ID, 
		NEXUS_TRUE, 
		NEXUS_FALSE);

    /*********************/
    /* Standard Clean-up */
    /*********************/

    nexus_endpoint_destroy(&GlobalEndpoint);
    nexus_endpointattr_destroy(&EpAttr);

    /* Free the node list */
    for(i = 0; i < n_nodes; i++)
    {
	nexus_free(nodes[i].name);
	/* OK to call startpoint destroy on nodes[0].startpoint */
	/* even though nodes[0].startpoint known to be NULL     */
	nexus_startpoint_destroy(&(nodes[i].startpoint));
    } /* endfor */
    nexus_free(nodes);
    
    /* Terminate master process */
    nexus_context_destroy(NEXUS_FALSE);

    nexus_printf("main(): ERROR: We should never get here.\n");

    return 0;

} /* end main() */

/***************************/
/*                         */
/* Nexus Handler Functions */
/*                         */
/***************************/

/*
 * rcv_handler()
 */
static void rcv_handler(nexus_endpoint_t *endpoint, 
			nexus_buffer_t *buffer,
			nexus_bool_t is_non_threaded_handler)
{
    int i; 
	
    /* extracting int from buffer */
    nexus_get_int(buffer, &i, 1);

    /* printing received value */
    nexus_printf("Slave received %d\n", i);

    /* clean-up */

    /* destroying received buffer */
    nexus_buffer_destroy(buffer);

    /* terminate slave process */
    nexus_context_destroy(is_non_threaded_handler);

} /* end rcv_handler() */
