/*
 * test_dc_small.c
 *
 * Description:
 *	Test of nexus_dc conversion routines for
 *	CRAY <--> IEEE floats in the denormal range.
 *
 * Bugs:
 *	IEEE denormal arithmetic is not fully supported
 *	on some platforms.  This may cause numbers to be
 *	printed incorrectly, or even core dumps.
 *	Fix: enable IEEE denormal support, or find a
 *	different machine.
 *
 * Usage:
 *   no options (normal):
 *	Gets each type and prints it out.  All floats should be
 *	the same when converting from Cray, the last couple floats
 *	will end in 0's when converting to Cray.  The first couple
 *	doubles converting from Cray will be off in the last hex
 *	digit, the last couple doubles converting to Cray will be
 *	off in the last 1-3 hex digits.  All loss of precision is
 *	due to one format having more useable bits in the mantissa
 *	than the other.
 *
 *   any option (debug):
 *	Prints out the local buffer, then divides each line and
 *	prints it out again.  This is for verification of the numbers
 *	used as tests (and only works on machines which support Cray
 *	format or IEEE denormals).  The numbers printed should be the
 *	same on every machine (proving that the sources are all the same).
 */

#include <stdio.h>
#include "nexus_dc.h"

#define NUM_FLTS	6
#define NUM_DBLS	6

/*
 * floats:
 *
 * flt[0] = 0x1.23abc * 2^-126, smallest normal exponent
 * flt[1] =  flt[0] / 8.0
 * flt[2] = -flt[1]
 * flt[3] =  flt[1] / 8.0
 * flt[4] = -flt[3]
 * flt[5] =  flt[3] / 8.0
 */
nexus_byte_t IEEE_FLT_BE[NUM_FLTS * 4] =
{
    0x00,0x91,0xd5,0xe0,
    0x00,0x12,0x3a,0xbc,
    0x80,0x12,0x3a,0xbc,
    0x00,0x02,0x47,0x57,
    0x80,0x02,0x47,0x57,
    0x00,0x00,0x48,0xea
};
nexus_byte_t IEEE_FLT_LE[NUM_FLTS * 4] =
{
    0xe0,0xd5,0x91,0x00,
    0xbc,0x3a,0x12,0x00,
    0xbc,0x3a,0x12,0x80,
    0x57,0x47,0x02,0x00,
    0x57,0x47,0x02,0x80,
    0xea,0x48,0x00,0x00
};
nexus_byte_t CRAY_FLT_BE[NUM_FLTS * 8] =
{
    0x3f,0x83,0x91,0xd5,0xe0,0x00,0x00,0x00,
    0x3f,0x80,0x91,0xd5,0xe0,0x00,0x00,0x00,
    0xbf,0x80,0x91,0xd5,0xe0,0x00,0x00,0x00,
    0x3f,0x7d,0x91,0xd5,0xe0,0x00,0x00,0x00,
    0xbf,0x7d,0x91,0xd5,0xe0,0x00,0x00,0x00,
    0x3f,0x7a,0x91,0xd5,0xe0,0x00,0x00,0x00
};

/*
 * doubles:
 *
 * dbl[0] = 0x1.2ab34cd56ef78 * 2^-1022, smallest normal exponent
 * dbl[1] =  dbl[0] / 32.0
 * dbl[2] = -dbl[1]
 * dbl[3] =  dbl[0] / 32.0
 * dbl[4] = -dbl[3]
 * dbl[5] =  dbl[3] / 32.0
 */
nexus_byte_t IEEE_DBL_BE[NUM_DBLS * 8] =
{
    0x00,0x12,0xab,0x34,0xcd,0x56,0xef,0x78,
    0x00,0x00,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0x80,0x00,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0x00,0x00,0x04,0xaa,0xcd,0x33,0x55,0xbb,
    0x80,0x00,0x04,0xaa,0xcd,0x33,0x55,0xbb,
    0x00,0x00,0x00,0x25,0x56,0x69,0x9a,0xad
};
nexus_byte_t IEEE_DBL_LE[NUM_DBLS * 8] =
{
    0x78,0xef,0x56,0xcd,0x34,0xab,0x12,0x00,
    0x7b,0xb7,0x6a,0xa6,0x59,0x95,0x00,0x00,
    0x7b,0xb7,0x6a,0xa6,0x59,0x95,0x00,0x80,
    0xbb,0x55,0x33,0xcd,0xaa,0x04,0x00,0x00,
    0xbb,0x55,0x33,0xcd,0xaa,0x04,0x00,0x80,
    0xad,0x9a,0x69,0x56,0x25,0x00,0x00,0x00
};
nexus_byte_t CRAY_DBL_BE[NUM_DBLS * 8] =
{
    0x3c,0x03,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0x3b,0xfe,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0xbb,0xfe,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0x3b,0xf9,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0xbb,0xf9,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
    0x3b,0xf4,0x95,0x59,0xa6,0x6a,0xb7,0x7b,
};

/*
 * pointers to the buffers which are the local format
 */
float	*local_flt_buf = NULL;
double	*local_dbl_buf = NULL;

/*
 * buffers to get to
 */
float	get_flt_buf[NUM_FLTS] = { 0.0 };
double	get_dbl_buf[NUM_DBLS] = { 0.0 };


/*
 * print a buffer of floats
 */
void print_flt_buf(char *msg, float *buf)
{
    int i, j, k;
    unsigned char *b = (unsigned char *) buf;
    char s[9];

    printf("%s\n", msg);
    for (i = 0; i < NUM_FLTS; i++)
    {
	switch (NEXUS_DC_FORMAT_LOCAL)
	{
	case NEXUS_DC_FORMAT_64BIT_LE:
	case NEXUS_DC_FORMAT_32BIT_LE:
	    for (j = 3; j > -1; j--)
		sprintf(&(s[j * 2]), "%2x", b[i * 4 + j]);
	    break;
	    
	case NEXUS_DC_FORMAT_64BIT_BE:
	case NEXUS_DC_FORMAT_CRAYT3E:
	case NEXUS_DC_FORMAT_32BIT_BE:
	    for (j = 0; j < 4; j++)
		sprintf(&(s[j * 2]), "%2x", b[i * 4 + j]);
	    break;
	    
	case NEXUS_DC_FORMAT_CRAYC90:
	    for (j = 0; j < 8; j++)
		sprintf(&(s[j * 2]), "%2x", b[i * 8 + j]);
	    break;
	}

	for (k = 0; k < 9; k++)
	{
	    if (s[k] == ' ')
		s[k] = '0';
	}

	printf("\tfloat[%i]  : hex = 0x%s; val = %e\n",
	       i, s, buf[i]);
    }
}

/*
 * print a buffer of doubles
 */
void print_dbl_buf(char *msg, double *buf)
{
    int i, j, k;
    unsigned char *b = (unsigned char *) buf;
    char s[17];

    printf("%s\n", msg);
    for (i = 0; i < NUM_DBLS; i++)
    {
	switch (NEXUS_DC_FORMAT_LOCAL)
	{
	case NEXUS_DC_FORMAT_64BIT_LE:
	case NEXUS_DC_FORMAT_32BIT_LE:
	    for (j = 7; j > -1; j--)
		sprintf(&(s[j * 2]), "%2x", b[i * 8 + j]);
	    break;
	    
	case NEXUS_DC_FORMAT_64BIT_BE:
	case NEXUS_DC_FORMAT_CRAYC90:
	case NEXUS_DC_FORMAT_CRAYT3E:
	case NEXUS_DC_FORMAT_32BIT_BE:
	    for (j = 0; j < 8; j++)
		sprintf(&(s[j * 2]), "%2x", b[i * 8 + j]);
	    break;
	}
	
	for (k = 0; k < 16; k++)
	    if (s[k] == ' ')
		s[k] = '0';
	printf("\tdouble[%i] : hex = 0x%s; val = %e\n",
	       i, s, buf[i]);
    }
}

/*
 * print out the local dc_format and set
 * the buffer pointers for the local format
 */
void get_local_bufs()
{
    switch (NEXUS_DC_FORMAT_LOCAL)
    {
    case NEXUS_DC_FORMAT_32BIT_LE:
	printf("Local dc_format = 32bit_le\n");
	local_flt_buf = (float *) IEEE_FLT_LE;
	local_dbl_buf = (double *) IEEE_DBL_LE;
	break;
    case NEXUS_DC_FORMAT_64BIT_LE:
	printf("Local dc_format = 64bit_le\n");
	local_flt_buf = (float *) IEEE_FLT_LE;
	local_dbl_buf = (double *) IEEE_DBL_LE;
	break;
    case NEXUS_DC_FORMAT_32BIT_BE:
	printf("Local dc_format = 32bit_be\n");
	local_flt_buf = (float *) IEEE_FLT_BE;
	local_dbl_buf = (double *) IEEE_DBL_BE;
	break;
    case NEXUS_DC_FORMAT_64BIT_BE:
	printf("Local dc_format = 64bit_be\n");
	local_flt_buf = (float *) IEEE_FLT_BE;
	local_dbl_buf = (double *) IEEE_DBL_BE;
	break;
    case NEXUS_DC_FORMAT_CRAYT3E:
	printf("Local dc_format = crayt3e\n");
	local_flt_buf = (float *) IEEE_FLT_BE;
	local_dbl_buf = (double *) IEEE_DBL_BE;
	break;
    case NEXUS_DC_FORMAT_CRAYC90:
	printf("Local dc_format = crayc90\n");
	local_flt_buf = (float *) CRAY_FLT_BE;
	local_dbl_buf = (double *) CRAY_DBL_BE;
	break;
    default:
	printf("Local dc_format = unknown!\n");
	break;
    }
    printf("sizeof(float)  = %i\n", sizeof(float));
    printf("sizeof(double) = %i\n", sizeof(double));
}

/*
 * main()
 */
int main(int argc, char **argv)
{
    nexus_byte_t *buffer;
    int		i, j;

    /*
     * init
     */
    get_local_bufs();
    
    /*
     * floats
     */
    if (argc == 1)
    {
	buffer = IEEE_FLT_LE;
	nexus_dc_get_float(&buffer,
			   get_flt_buf,
			   NUM_FLTS,
			   NEXUS_DC_FORMAT_32BIT_LE);
	print_flt_buf("Got IEEE_LE:", get_flt_buf);
	
	buffer = IEEE_FLT_BE;
	nexus_dc_get_float(&buffer,
			   get_flt_buf,
			   NUM_FLTS,
			   NEXUS_DC_FORMAT_32BIT_BE);
	print_flt_buf("Got IEEE_BE:", get_flt_buf);
    
	buffer = CRAY_FLT_BE;
	nexus_dc_get_float(&buffer,
			   get_flt_buf,
			   NUM_FLTS,
			   NEXUS_DC_FORMAT_CRAYC90);
	print_flt_buf("Got CRAY:", get_flt_buf);
    }
    else
    {
	print_flt_buf("Local floats:", local_flt_buf);

	for (i = 0; i < NUM_FLTS; i++)
	    local_flt_buf[i] /= 8.0;
	print_flt_buf("Divided by 8:", local_flt_buf);
    }

    /*
     * doubles
     */
    if (argc == 1)
    {
	buffer = IEEE_DBL_LE;
	nexus_dc_get_double(&buffer,
			    get_dbl_buf,
			    NUM_DBLS,
			    NEXUS_DC_FORMAT_32BIT_LE);
	print_dbl_buf("Got IEEE_LE:", get_dbl_buf);

	buffer = IEEE_DBL_BE;
	nexus_dc_get_double(&buffer,
			    get_dbl_buf,
			    NUM_DBLS,
			    NEXUS_DC_FORMAT_32BIT_BE);
	print_dbl_buf("Got IEEE_BE:", get_dbl_buf);

	buffer = CRAY_DBL_BE;
	nexus_dc_get_double(&buffer,
			    get_dbl_buf,
			    NUM_DBLS,
			    NEXUS_DC_FORMAT_CRAYC90);
	print_dbl_buf("Got CRAY:", get_dbl_buf);
    }
    else
    {
	print_dbl_buf("Local doubles:", local_dbl_buf);
	
	for (i = 0; i < NUM_FLTS; i++)
	    local_dbl_buf[i] /= 32.0;
	print_dbl_buf("Divided by 32:", local_dbl_buf);
    }

    /*
     * cleanup and return
     */
    return (0);
}
